# by Xin Liu
# Email: xliu@mail.missouri.edu

# Load code
source("ivqr_onestep.R")
source("ivqr_see.R")
source("ivqr_gmm.R")
source("gmmq.R")

require('quantreg')

library("MASS") # for mvrnorm()




AVGBS.multipar <- function(Y, D, X.excl, Z.excl, tau, n, h=0, dB, BREPLIC, reportGMM,
                     weights.new,  
                     b.hats.original=data.frame(MM=rep(NA,1),  IV=NA, QR=NA),
                     b.IVQR, DGP, BS.GMMQ.FLAG=TRUE) {


# generate an original dataset
 data.bs <- cbind(Y=Y, D=D, X.excl=X.excl, Z.excl=Z.excl)

Y<-as.matrix(Y)
dD <- ncol(D)
dZ.excl <- ncol(Z.excl)
dX.excl <- ncol(X.excl)
dZ <- dX.excl + dZ.excl
dX <- dX.excl + dD

# Allocate storage variables
b.hats.bs <- data.frame(GMM=rep(NA,BREPLIC),  IV=NA, QR=NA #, AVG=NA
                    )
b.hats.avg.bs <- array(NA,dim=c(13071,BREPLIC))
b.hats7.avg.bs <- b.hats2.avg.bs <- b.hats3.avg.bs <- b.hats4.avg.bs <- b.hats5.avg.bs <- b.hats6.avg.bs <- array(NA,dim=c(13071,BREPLIC))
b.hats.avg.weights.bs <- cbind(b.hats.avg.bs, weights.new)
b.hats.MSE.bs <- cbind(b.hats.avg.bs, weights.new)
b.hats.rRMSE.bs <- cbind(b.hats.avg.bs, weights.new)

b.hats.GMM.bs <- array(NA,dim=c(dB,BREPLIC))
b.hats.IV.bs <- array(NA,dim=c(dB,BREPLIC))
b.hats.QR.bs <- array(NA,dim=c(dB,BREPLIC))


# # begin bootstrap
 cat("Boop=0000") #to update progress
for (ibep in 1:BREPLIC) {
   cat(sprintf("ibep%04d",ibep)) #progress update
  
  # Generate data
  data.sim<-data.bs[sample(x=1:n, size=n, replace=TRUE), ]


  Y.bs <- as.matrix(data.sim[,1],nrow=n)
  D.bs <- as.matrix(data.sim[,2:(dD+1)],nrow=n)
  X.excl.bs <- as.matrix(data.sim[,(dD+2):(dX.excl+dD+1)],nrow=n)
  Z.excl.bs <- as.matrix(data.sim[,(dX.excl+dD+2):(dX.excl+dD+dZ.excl+1)],nrow=n)
  
  Z.bs <- cbind(X.excl.bs, Z.excl.bs) # Z is (intercep slope) # has intercept
  X.bs <- cbind(X.excl.bs, D.bs) # has intercept
  Y.gmmq.bs <- cbind(Y.bs, D.bs)
  
  # 
  
  Lambda.bs <- function(y, x, b){y-x%*%b } # require X and b has the same sequence. If X is (intercept, slope), 
     # b must be (intercept slope). Notice that X=(X.excl, D), but most b(b.gmmq, b.ivqr.gmm, b.onestep) are (slope, intercept), 
     # so need to adjust X or use Lambda.gmmq w/ Y.gmmq, X.excl
  Lambda.derivative.bs <- function(y,x,b) -x # lambda.derivative fn gives the same sequence of X. 
     # If X is (intercept, slope) (as defined here, X=(X.excl, D)), the lambda.derivative gives (intercept, slope); 
     # if wanna labdha.derives generates (slope, intercept), then should use Lambda.derivative.gmmq or use Lambda.derivative (X=cbind(D,X.excl))
  Lambda.gmmq.bs <- function(y,x,b) {y[,1]-cbind(y[,-1],x)%*%b} # b must be (slope, intercept) sequence. need to check this, when calling lambda.gmmq fn. # HAS INTERCEPT
  Lambda.derivative.gmmq.bs <- function(y,x,b) {-cbind(y[,-1],x)} # lambda.derivative.gmmq gives (slope, intercept) sequence. # HAS INTERCEPT
  
  
  
  Itilde <- Itilde.KS17
  Itilde.deriv <- Itilde.deriv.KS17
  
  Gfn <- function(v,h) {      Itilde(v/h)       }
  Gpfn <- function(v,h) {     Itilde.deriv(v/h) }
  
  
  # Compute estimators and store
  # 
  
  
  ##################################################
  if(BS.GMMQ.FLAG){
    gmmq.est.bs <- gmmq(tau=tau, Y=Y.gmmq.bs, X=X.excl.bs, Z.excl=Z.excl.bs, dB=dB, 
                                Lambda=Lambda.gmmq.bs, 
                                Lambda.derivative=Lambda.derivative.gmmq.bs, 
                                h=0, VERBOSE=FALSE, RETURN.Z=FALSE )  
    h.gmmq.bs <- gmmq.est.bs$h # do the gmmq is to get h.gmmq; h.gmmq is used for cov.mtx; 
    # cov.mtx is used for weighting matrix in GMM method to get "aggressive"; i.e. gmmq()-->h.gmmq-->cov.mtx-->weight.mtx in GMM of "aggressive".
    if (is.na(h.gmmq.bs)) {cat("h.gmmq.bs is NA")}
    
    h.small.bs <- h.gmmq.bs
    b.gmmq.bs <- gmmq.est.bs$b
    b.gmmq.bs.intslp<-as.matrix(c(b.gmmq.bs[7], b.gmmq.bs[1:6]),ncol=1)
    
    # store b.conservative
    b.conservative.bs <- b.gmmq.bs
    b.conservative.bs.intslp <- b.gmmq.bs.intslp

    
  }else{
    # MM 
    gmmq.est.bs <- gmmq(tau=tau, Y=Y.gmmq.bs, X=X.excl.bs, Z.excl=Z.excl.bs, dB=dB, 
                                Lambda=Lambda.gmmq.bs, 
                                Lambda.derivative=Lambda.derivative.gmmq.bs, 
                                h=0, VERBOSE=FALSE, RETURN.Z=FALSE )  
    h.gmmq.bs <- gmmq.est.bs$h # do the gmmq is to get h.gmmq; h.gmmq is used for cov.mtx; 
    # cov.mtx is used for weighting matrix in GMM method to get "aggressive"; i.e. gmmq()-->h.gmmq-->cov.mtx-->weight.mtx in GMM of "aggressive".
    if (is.na(h.gmmq.bs)) {cat("h.gmmq.bs is NA")}
    
    h.small.bs <- h.gmmq.bs
    b.gmmq.bs <- gmmq.est.bs$b
    
    L.con.bs <- matrix(Lambda.gmmq(y=Y.gmmq.bs, x=X.excl.bs, b=b.gmmq.bs), ncol=1) # has intercept 
    gni.con.bs <- cbind(Z.excl.bs, X.excl.bs)*repmat((Gfn(-L.con.bs, h.small.bs)-tau),1,dZ) # here gni is (slope, intercept), (Z.excl, X.excl) # has intercetp
    
    g.bar.con.bs <- colMeans(gni.con.bs)                                                   # g.bar.agg.MMQR sequence is (D, Z.excl, X.excl)
    cov.mtx.con.bs <- t(gni.con.bs)%*%gni.con.bs/n-g.bar.con.bs%*%t(g.bar.con.bs) # cov.mat, weight.mtx sequence is (D, Z.excl, 1)
    inv.cov.mtx.con.bs <- solve(cov.mtx.con.bs)
    
    #################### compute  b.initial, for weighting mtx in GMM.CON ################      END ####################  
    
    #################### compute  CON.GMM ##########################################      begin ####################  
    # use inverse of cov.mtx as the weighting matrix to construct the GMM criterion. run GenSA with GMM. get b.
    tmp.con.1step.bs <- ivqr.gmm(Y = Y.bs, X.excl = X.excl.bs, Z.excl =Z.excl.bs, D = D.bs, tau = tau, h = h.small.bs, dB = dB, max.time=5, upper=rep(1000,dB), lower=rep(-1000,dB),
                                         weight.mtx=inv.cov.mtx.con.bs, structure = 'iid', LRV.kernel = 'QS', b.init = b.gmmq.bs ) 
    # ivqr.gmm gives b=(slope, intercept) = on (generalZ.excl, X.excl)=i.e. on (D, Z.excl, X.excl) = on (D, 1) 
    b.con.1step.bs <- tmp.con.1step.bs$b
    #################### compute  CON.GMM ##########################################      end ####################  
    
    # store b.conservative

    b.con.1step.bs.intslp<-as.matrix(c(b.con.1step.bs[7], b.con.1step.bs[1:6]),ncol=1)
    
    b.conservative.bs <- b.con.1step.bs
    b.conservative.bs.intslp <- b.con.1step.bs.intslp
    
    }

  
  # store b.conservative
  b.hats.GMM.bs[,ibep] <-b.conservative.bs.intslp   
  
  if (is.na(b.conservative[1]) | is.na(b.conservative[2])) {cat("b.gmmq.bs is NA")}
  
  
  # 2SLS
  PZ.bs <- Z.bs %*% solve(t(Z.bs)%*%Z.bs) %*% t(Z.bs)
  b.2sls.bs <- solve(t(X.bs)%*%PZ.bs%*%X.bs) %*% t(X.bs)%*%PZ.bs%*%Y.bs
  b.hats.IV.bs[,ibep] <- b.2sls.bs
  
  if (is.na(b.2sls.bs[1]) | is.na(b.2sls.bs[2]) ) { cat("b.2sls.bs is NA")}
  # Z is on (intercept, slope), X is on (intercept, slope), so b.2SLS is on (intercept, slope); we're interested in slope, thus look at b.2sls[2]
  
  # 
  # QR  
  suppressWarnings(b.hats.QR.bs[,ibep] <- rq(Y.bs~D.bs)$coefficients)
  

  for (iweights in 1:13071){
    avg.vector.bs <- b.hats.GMM.bs[,ibep]*repmat(b.hats.avg.weights.bs[iweights, 1+BREPLIC],1,dB) + b.hats.IV.bs[,ibep]*repmat(b.hats.avg.weights.bs[iweights,2+BREPLIC],1,dB)+b.hats.QR.bs[,ibep]*repmat(b.hats.avg.weights.bs[iweights,3+BREPLIC],1,dB)
    b.hats.MSE.bs[iweights,ibep] <- n*t(as.matrix(avg.vector.bs[2:7], dim=c(1,6))-as.matrix(b.IVQR[2:7], dim=c(1,6)))%*%(as.matrix(avg.vector.bs[2:7], dim=c(1,6))-as.matrix(b.IVQR[2:7], dim=c(1,6)))
    b.hats2.avg.bs[iweights, ibep] <- avg.vector.bs[2]
    b.hats3.avg.bs[iweights, ibep] <- avg.vector.bs[3]
    b.hats4.avg.bs[iweights, ibep] <- avg.vector.bs[4]
    b.hats5.avg.bs[iweights, ibep] <- avg.vector.bs[5]
    b.hats6.avg.bs[iweights, ibep] <- avg.vector.bs[6]
    b.hats7.avg.bs[iweights, ibep] <- avg.vector.bs[7]
  }
  

} # END for loop over ibep/BREPLIC

b.hats.MSE.sum <- as.matrix(rowSums(b.hats.MSE.bs[,(1:BREPLIC)], na.rm=TRUE))



b.hats2.avgbs.tr <- t(b.hats2.avg.bs)
b.hats2.avgbs.median.bias <- apply(X = b.hats2.avgbs.tr, MARGIN = 2, FUN = median) - b.IVQR[2]
b.hats2.avgbs.IQR <- apply(X = b.hats2.avgbs.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.hats2.avgbs.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.hats2.avgbs.rRMSE <- sqrt(b.hats2.avgbs.median.bias^2 + (b.hats2.avgbs.IQR/1.349)^2)

b.hats3.avgbs.tr <- t(b.hats3.avg.bs)
b.hats3.avgbs.median.bias <- apply(X = b.hats3.avgbs.tr, MARGIN = 2, FUN = median) - b.IVQR[3]
b.hats3.avgbs.IQR <- apply(X = b.hats3.avgbs.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.hats3.avgbs.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.hats3.avgbs.rRMSE <- sqrt(b.hats3.avgbs.median.bias^2 + (b.hats3.avgbs.IQR/1.349)^2)

b.hats4.avgbs.tr <- t(b.hats4.avg.bs)
b.hats4.avgbs.median.bias <- apply(X = b.hats4.avgbs.tr, MARGIN = 2, FUN = median) - b.IVQR[4]
b.hats4.avgbs.IQR <- apply(X = b.hats4.avgbs.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.hats4.avgbs.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.hats4.avgbs.rRMSE <- sqrt(b.hats4.avgbs.median.bias^2 + (b.hats4.avgbs.IQR/1.349)^2)

b.hats5.avgbs.tr <- t(b.hats5.avg.bs)
b.hats5.avgbs.median.bias <- apply(X = b.hats5.avgbs.tr, MARGIN = 2, FUN = median) - b.IVQR[5]
b.hats5.avgbs.IQR <- apply(X = b.hats5.avgbs.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.hats5.avgbs.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.hats5.avgbs.rRMSE <- sqrt(b.hats5.avgbs.median.bias^2 + (b.hats5.avgbs.IQR/1.349)^2)

b.hats6.avgbs.tr <- t(b.hats6.avg.bs)
b.hats6.avgbs.median.bias <- apply(X = b.hats6.avgbs.tr, MARGIN = 2, FUN = median) - b.IVQR[6]
b.hats6.avgbs.IQR <- apply(X = b.hats6.avgbs.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.hats6.avgbs.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.hats6.avgbs.rRMSE <- sqrt(b.hats6.avgbs.median.bias^2 + (b.hats6.avgbs.IQR/1.349)^2)

b.hats7.avgbs.tr <- t(b.hats7.avg.bs)
b.hats7.avgbs.median.bias <- apply(X = b.hats7.avgbs.tr, MARGIN = 2, FUN = median) - b.IVQR[7]
b.hats7.avgbs.IQR <- apply(X = b.hats7.avgbs.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.hats7.avgbs.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.hats7.avgbs.rRMSE <- sqrt(b.hats7.avgbs.median.bias^2 + (b.hats7.avgbs.IQR/1.349)^2)

b.hats.avg.bs.rRMSE.sqrtofsum <- sqrt(b.hats2.avgbs.rRMSE^2 + b.hats3.avgbs.rRMSE^2 + b.hats4.avgbs.rRMSE^2 + b.hats5.avgbs.rRMSE^2 + b.hats6.avgbs.rRMSE^2 + b.hats7.avgbs.rRMSE^2 )



weightnumber.MSE <- which.min(b.hats.MSE.sum)
weightnumber.RRMSE <- which.min(b.hats.avg.bs.rRMSE.sqrtofsum)


o.weight.AVGBS.MSE <- weights.new[weightnumber.MSE,]
b.AVGBS.MSE <- b.hats.original[, 1]*o.weight.AVGBS.MSE[1]+b.hats.original[, 2]*o.weight.AVGBS.MSE[2]+b.hats.original[, 3]*o.weight.AVGBS.MSE[3]


o.weight.AVGBS.RRMSE <- weights.new[weightnumber.RRMSE,]
b.AVGBS.RRMSE <- b.hats.original[, 1]*o.weight.AVGBS.RRMSE[1]+b.hats.original[, 2]*o.weight.AVGBS.RRMSE[2]+b.hats.original[, 3]*o.weight.AVGBS.RRMSE[3]


return(list(b.AVGBS.MSE=b.AVGBS.MSE,  b.AVGBS.RRMSE=b.AVGBS.RRMSE, o.weight.AVGBS.MSE=o.weight.AVGBS.MSE, o.weight.AVGBS.RRMSE=o.weight.AVGBS.RRMSE))

}






# #EOF
