# by Xin Liu
# Email: xliu@mail.missouri.edu


# Load code
source("ivqr_onestep.R")
source("ivqr_see.R")
source("ivqr_gmm.R")
source("gmmq.R")

require('quantreg')

library("MASS") # for mvrnorm()




AVGBS <- function(Y, D, X.excl, Z.excl, tau, n, h=0, dB, BREPLIC, reportGMM,
                     weights.new,  
                     b.hats.original=data.frame(MM=rep(NA,1),  IV=NA, QR=NA
                                                ),
                     b.IVQR, DGP) {


# generate an original dataset
data.bs <- cbind(Y=Y, D=D, X.excl=X.excl, Z.excl=Z.excl)



# Allocate storage variables
b.hats.bs <- data.frame(MM=rep(NA,BREPLIC), 
                        IV=NA, QR=NA #, AVG=NA
                        )
b.hats.avg.bs <- array(NA,dim=c(13071,BREPLIC))
b.hats.avg.weights.bs <- cbind(b.hats.avg.bs, weights.new)
b.hats.MSE.bs <- cbind(b.hats.avg.bs, weights.new)
b.hats.rRMSE.bs <- cbind(b.hats.avg.bs, weights.new)


# # begin bootstrap
 cat("Boop=0000") #to update progress
for (ibep in 1:BREPLIC) {
   cat(sprintf("ibep%04d",ibep)) #progress update
  
  # Generate data
  data.sim<-data.bs[sample(x=1:n, size=n, replace=TRUE), ]

  
  Y.bs <- as.matrix(data.sim[,1],nrow=n)
  X.excl.bs <- as.matrix(data.sim[,3],nrow=n)
  Z.excl.bs <- as.matrix(data.sim[,4],nrow=n)
  D.bs <- as.matrix(data.sim[,2],nrow=n)
  Z.bs <- cbind(X.excl.bs, Z.excl.bs) # Z is (intercep slope)
  X.bs <- cbind(X.excl.bs, D.bs)
  Y.gmmq.bs <- cbind(Y.bs, D.bs)
  # 
  
  Lambda.bs <- function(y, x, b){y-x%*%b } # require X and b has the same sequence. If X is (intercept, slope), 
     # b must be (intercept slope). Notice that X=(X.excl, D), but most b(b.gmmq, b.ivqr.gmm, b.onestep) are (slope, intercept), 
     # so need to adjust X or use Lambda.gmmq w/ Y.gmmq, X.excl
  Lambda.derivative.bs <- function(y,x,b) -x # lambda.derivative fn gives the same sequence of X. 
     # If X is (intercept, slope) (as defined here, X=(X.excl, D)), the lambda.derivative gives (intercept, slope); 
     # if wanna labdha.derives generates (slope, intercept), then should use Lambda.derivative.gmmq or use Lambda.derivative (X=cbind(D,X.excl))
  Lambda.gmmq.bs <- function(y,x,b) {y[,1]-cbind(y[,-1],x)%*%b} # b must be (slope, intercept) sequence. need to check this, when calling lambda.gmmq fn.
  Lambda.derivative.gmmq.bs <- function(y,x,b) {-cbind(y[,-1],x)} # lambda.derivative.gmmq gives (slope, intercept) sequence.
  
  
  
  Itilde <- Itilde.KS17
  Itilde.deriv <- Itilde.deriv.KS17
  
  Gfn <- function(v,h) {      Itilde(v/h)       }
  Gpfn <- function(v,h) {     Itilde.deriv(v/h) }
  
  
  # Compute estimators and store
  # 
  # MM 
  gmmq.est.bs <- gmmq(tau=tau, Y=Y.gmmq.bs, X=X.excl.bs, Z.excl=Z.excl.bs, dB=dB, 
                   Lambda=Lambda.gmmq.bs, 
                   Lambda.derivative=Lambda.derivative.gmmq.bs, 
                   h=0, VERBOSE=FALSE, RETURN.Z=FALSE )  
  h.gmmq.bs <- gmmq.est.bs$h # do the gmmq is to get h.gmmq; h.gmmq is used for cov.mtx; 
   # cov.mtx is used for weighting matrix in GMM method to get "aggressive"; i.e. gmmq()-->h.gmmq-->cov.mtx-->weight.mtx in GMM of "aggressive".
  
  if (is.na(h.gmmq.bs)) {
    cat("h.gmmq.bs is NA")
  }
  
  h.small.bs <- h.gmmq.bs
  
  b.conservative.bs <- b.gmmq.bs <- gmmq.est.bs$b # b.gmmq=(slope, intercept) (interested in slope)

  b.hats.bs$MM[ibep] <-b.conservative.bs[1]  
  
  if (is.na(b.conservative[1]) | is.na(b.conservative[2])) {
    cat("b.gmmq.bs is NA")
  }
  
  
  # 2SLS
  PZ.bs <- Z.bs %*% solve(t(Z.bs)%*%Z.bs) %*% t(Z.bs)
  b.2sls.bs <- solve(t(X.bs)%*%PZ.bs%*%X.bs) %*% t(X.bs)%*%PZ.bs%*%Y.bs
  b.hats.bs$IV[ibep] <- b.2sls.bs[2]
  
  if (is.na(b.2sls.bs[1]) | is.na(b.2sls.bs[2]) ) { cat("b.2sls.bs is NA")
    }
  # Z is on (intercept, slope), X is on (intercept, slope), so b.2SLS is on (intercept, slope); we're interested in slope, thus look at b.2sls[2]
  
  # 
  # QR  
  suppressWarnings(b.hats.bs$QR[ibep] <- rq(formula=Y~X,tau=tau,data=data.frame(Y=Y.bs, X=D.bs))$coefficients[2])
  
  if (is.na(b.hats.bs$QR[ibep][1]) ) {cat("rq is NA")}
  
  #
  for (iweights in 1:13071){
    b.hats.avg.weights.bs[iweights,ibep] <- b.hats.bs$MM[ibep]*b.hats.avg.weights.bs[iweights,1+BREPLIC] + b.hats.bs$IV[ibep]*b.hats.avg.weights.bs[iweights,2+BREPLIC]+b.hats.bs$QR[ibep]*b.hats.avg.weights.bs[iweights,3+BREPLIC]
    b.hats.MSE.bs[iweights,ibep] <- (b.hats.avg.weights.bs[iweights,ibep]-b.IVQR)^2 
        }
  

} # END for loop over ibep/BREPLIC


b.hats.avgbs.tr <- t(b.hats.avg.weights.bs[,1:BREPLIC])
b.hats.avgbs.median.bias <- apply(X = b.hats.avgbs.tr, MARGIN = 2, FUN = median) - b.IVQR
b.hats.avgbs.IQR <- apply(X = b.hats.avgbs.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.hats.avgbs.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.hats.avgbs.rRMSE <- sqrt(b.hats.avgbs.median.bias^2 + (b.hats.avgbs.IQR/1.349)^2)


b.hats.avgbs.mean <- colMeans(b.hats.avgbs.tr)
b.hats.avgbs.bias <- b.hats.avgbs.mean - b.IVQR
b.hats.avgbs.sd <- apply(X = b.hats.avgbs.tr, MARGIN = 2, FUN = sd)
b.hats.avgbs.RMSE <- sqrt(b.hats.avgbs.bias^2 + b.hats.avgbs.sd^2)



b.hats.MSE.sum <- as.matrix(rowSums(b.hats.MSE.bs[,(1:BREPLIC)], na.rm=TRUE))


weightnumber.MSE <- which.min(b.hats.MSE.sum)
weightnumber.RMSE <- which.min(b.hats.avgbs.RMSE)
weightnumber.RRMSE <- which.min(b.hats.avgbs.rRMSE)


o.weight.AVGBS.MSE <- weights.new[weightnumber.MSE,]
b.AVGBS.MSE <- b.hats.original[1]*o.weight.AVGBS.MSE[1]+b.hats.original[2]*o.weight.AVGBS.MSE[2]+b.hats.original[3]*o.weight.AVGBS.MSE[3]

o.weight.AVGBS.RMSE <- weights.new[weightnumber.RMSE,]
b.AVGBS.RMSE <- b.hats.original[1]*o.weight.AVGBS.RMSE[1]+b.hats.original[2]*o.weight.AVGBS.RMSE[2]+b.hats.original[3]*o.weight.AVGBS.RMSE[3]

o.weight.AVGBS.RRMSE <- weights.new[weightnumber.RRMSE,]
b.AVGBS.RRMSE <- b.hats.original[1]*o.weight.AVGBS.RRMSE[1]+b.hats.original[2]*o.weight.AVGBS.RRMSE[2]+b.hats.original[3]*o.weight.AVGBS.RRMSE[3]


# b.hats.orginal [1,2,3]=[MM, IV, QR]

return(list(b.AVGBS.MSE=b.AVGBS.MSE, b.AVGBS.RMSE=b.AVGBS.RMSE, b.AVGBS.RRMSE=b.AVGBS.RRMSE, 
            o.weight.AVGBS.MSE=o.weight.AVGBS.MSE, o.weight.AVGBS.RMSE=o.weight.AVGBS.RMSE, 
            o.weight.AVGBS.RRMSE=o.weight.AVGBS.RRMSE))

}



# 
# #EOF


