# simulation model 1 code for ``averaging estimation for instrumental variables quantile regression"
# by Xin Liu
# xl6f6@mail.missouri.edu or xliu@mail.missouri.edu


rm(list=ls())
st0 <- Sys.time()

SAVE.FLAG <- TRUE # save to file?
NREPLIC <- 3 # number of replications/draws per simulation
BREPLIC <- 2
reportGMM <-FALSE
PDF.FLAG <- TRUE

PARALLEL.FLAG <- TRUE

if (NREPLIC==1000){
  BS.FLAG <- FALSE
}else{
  BS.FLAG <- TRUE
}

 BS.FLAG <- TRUE ###

HETER.PIC.FLAG <- FALSE
ENDOG.PIC.FLAG <- FALSE
FATAIL.PIC.FLAG <- FALSE

LEGROOM <- 1.25

# Load code
source("ivqr_onestep.R")
source("ivqr_see.R")
source("ivqr_gmm_avg.R")
source("gmmq.R")

source("AVGBS_fn_new.R")

source("ivqr_iv_avg.R")


require('quantreg')

library("GenSA")
library("MASS") # for mvrnorm()


library("ks")
library("kdensity")


library(parallel)



require("parallel")
require("foreach")
require("doRNG")
require("doParallel")



# Set parameters
n <- 1000 #sample size
NS <- 1000

tau <- 0.4 #quantile index
TAUs <-0.4


H <- 0.0001 #bandwidth
Ups<-matrix(c(1,0,0,0),ncol = 2,nrow = 2)


dB <-2


#NEW NEW 14 DGP sequence 
para.space <- matrix(c(0,    0,   0,
                       0.15, 0,   0,
                       0.3,  0,   0,
                       0.5,  0,   0,
                       0,    0.4, 0,
                       0.1,  0.4, 0,
                       0.5,  0.3, 0,
                       0.5,  0.5, 0,
                       0.1,  0.9, 0,
                       0,    0,   3,
                       0,    0,   1,
                       0.5,  0,   5,
                       0.5,  0,   3,
                       0.5,  0,   1), ncol=3, byrow = TRUE)




if(ENDOG.PIC.FLAG){
  para.space <- matrix(c(0,    0,   0,
                         0.1,  0,   0,
                         0.2,  0,   0,
                         0.3,  0,   0,
                         0.4,  0,   0,
                         0.5,  0,   0,
                         0.6,  0,   0,
                         0.7,  0,   0,
                         0.8,  0,   0,
                         0.9,  0,   0,
                         1,    0,   0,
                         0,    0.5,   0,
                         0.1,  0.5,   0,
                         0.2,  0.5,   0,
                         0.3,  0.5,   0,
                         0.4,  0.5,   0,
                         0.5,  0.5,   0,
                         0.6,  0.5,   0,
                         0.7,  0.5,   0,
                         0.8,  0.5,   0,
                         0.9,  0.5,   0,
                         1,    0.5,   0,
                         0,    1,   0,
                         0.1,  1,   0,
                         0.2,  1,   0,
                         0.3,  1,   0,
                         0.4,  1,   0,
                         0.5,  1,   0,
                         0.6,  1,   0,
                         0.7,  1,   0,
                         0.8,  1,   0,
                         0.9,  1,   0,
                         1,    1,   0
  ), ncol=3, byrow = TRUE)
  
}

if (HETER.PIC.FLAG){
  para.space <- matrix(c(0, 0,       0,
                         0, 0.1,     0,
                         0, 0.2,     0,
                         0, 0.3,     0,
                         0, 0.4,     0,
                         0, 0.5,     0,
                         0, 0.6,     0,
                         0, 0.7,     0,
                         0, 0.8,     0,
                         0, 0.9,     0,
                         0, 1,       0,
                         0.2, 0,       0,
                         0.2, 0.1,     0,
                         0.2, 0.2,     0,
                         0.2, 0.3,     0,
                         0.2, 0.4,     0,
                         0.2, 0.5,     0,
                         0.2, 0.6,     0,
                         0.2, 0.7,     0,
                         0.2, 0.8,     0,
                         0.2, 0.9,     0,
                         0.2, 1,       0,
                         0.5, 0,       0,
                         0.5, 0.1,     0,
                         0.5, 0.2,     0,
                         0.5, 0.3,     0,
                         0.5, 0.4,     0,
                         0.5, 0.5,     0,
                         0.5, 0.6,     0,
                         0.5, 0.7,     0,
                         0.5, 0.8,     0,
                         0.5, 0.9,     0,
                         0.5, 1,       0
  ), ncol=3, byrow = TRUE)
}

if (FATAIL.PIC.FLAG){
  para.space <- matrix(c(0, 0,       1,
                         0, 0,     1.5,
                         0, 0,     2,
                         0, 0,     2.5,
                         0, 0,     3,
                         0, 0,     3.5,
                         0, 0,     4,
                         0, 0,     4.5,
                         0, 0,     5,
                         0, 0,     6,
                         0, 0,       7,
                         0.2, 0,       1,
                         0.2, 0,     1.5,
                         0.2, 0,     2,
                         0.2, 0,     2.5,
                         0.2, 0,     3,
                         0.2, 0,     3.5,
                         0.2, 0,     4,
                         0.2, 0,     4.5,
                         0.2, 0,     5,
                         0.2, 0,     6,
                         0.2, 0,       7,
                         0.5, 0,       1,
                         0.5, 0,     1.5,
                         0.5, 0,     2,
                         0.5, 0,     2.5,
                         0.5, 0,     3,
                         0.5, 0,     3.5,
                         0.5, 0,     4,
                         0.5, 0,     4.5,
                         0.5, 0,     5,
                         0.5, 0,     6,
                         0.5, 0,       7  ), ncol=3, byrow = TRUE)
}


if ( ENDOG.PIC.FLAG && HETER.PIC.FLAG){
  cat(paste0("ERROR, COULDN'T BOTH SET endog and hetero level"))
  para.space <- matrix(c(0, 0,       0), ncol=3, byrow = TRUE)
}



# Set output file
OUTFILE <- ""
if (SAVE.FLAG) {
  OUTFILE <- paste0(sprintf("JPTA_sims_NREP%d__BREP%d_tau%.2f_n%d_DGP%d", NREPLIC, BREPLIC, tau, n, nrow(para.space)),format(Sys.time(),"%Y_%m_%d"),".txt")
}




############################################################ 13071 sets of weights ################
w1.new <- w2.new <- w3.new <- share1 <- share2 <- array(NA,dim=c(13071,1))

for (i in 1:10000){
    w1.new[i] <- (floor((i-1)/100))/100
    w2.new[i] <- (1-w1.new[i])*(i-100*(floor(i/100)))/100
    w3.new[i] <- (1-w1.new[i])*(1-(i-100*(floor(i/100)))/100)
}
#######

t <- 10000
for (i in 85:99){
  for (j in 1:99){
        t<- t+1
        w1.new[t] <- i/100+j/10000
        w2.new[t] <- 1-i/100-j/10000
        w3.new[t] <- 0
    
  }
}
# 15*99=1485 counts
# t= 10001 to 11485

for (i in 85:99){
  for (j in 1:99){
    t<- t+1
    w1.new[t] <- i/100+j/10000
    w2.new[t] <- 0
    w3.new[t] <- 1-i/100-j/10000
    
  }
}
# 1485 counts
# t= 11486 to 12970

for (i in 0:100){
  t<- t+1
  w1.new[t] <- i/100
  w2.new[t] <- 1-i/100
  w3.new[t] <- 0
}
# 101
# t=12971 to 13071


weights.new <- cbind(w1.new, w2.new, w3.new)
############################################################ 13071 sets of weights ################






if (PARALLEL.FLAG){

# Calculate the number of cores
num_cores <- detectCores() 

# Initiate cluster
cl <- makeCluster(num_cores)

num_cores <- tryCatch({cl <- makeCluster(num_cores); registerDoParallel(cl); on.exit(stopCluster(cl),add=TRUE); num_cores}, error=function(Err){warning(sprintf("Error creating %d clusters",num_cores)); 1})
set.seed(112358) #for replication

}else{
   set.seed(112358) #for replication
}

start.time <- Sys.time() #to see elapsed time


b.hats.rRMSE.par <- b.hats.RMSE.par<-  cbind(data.frame(MM=rep(NA,1), 
                                                        SEE=NA,
                                                        IV=NA, QR=NA, 
                                                        AVG.MMIV=NA, AGG.MMIV=NA, AVG.MMQR=NA, AGG.MMQR=NA,
                                                        AVG.BS.MSE=NA, AVG.BS.RMSE=NA, AVG.BS.RRMSE=NA
                                                        )
                                             )
b.hats.rRMSE.par.rel <- b.hats.RMSE.par.rel <-data.frame(MM=rep(NA,1), relSEE=NA, relIV=NA, relQR=NA, relAVG.MMIV=NA, relAGG.MMIV=NA, relAVG.MMQR=NA,  relAGG.MMQR=NA, 
                                                         relAVG.BS.MSE=NA, relAVG.BS.RMSE=NA, relAVG.BS.RRMSE=NA
                                                         )


rel.rRMSE <- rel.RMSE <- cbind(data.frame(MM=rep(NA,nrow(para.space)), 
                              SEE=NA,
                              IV=NA, QR=NA, 
                              AVG.MMIV=NA, AGG.MMIV=NA, AVG.MMQR=NA, AGG.MMQR=NA,
                              AVG.BS.MSE=NA, AVG.BS.RMSE=NA, AVG.BS.RRMSE=NA,
                              endog=NA, heter=NA, fattail=NA)
)



MMIV.weight.new <- array(NA,dim=c(nrow(para.space),NREPLIC))
MMQR.weight.new <- array(NA,dim=c(nrow(para.space),NREPLIC))

MMQR.weight.all <- array(NA,dim=c(nrow(para.space),NREPLIC+3))
MMIV.weight.all <- array(NA,dim=c(nrow(para.space),NREPLIC+3))

BS.W1.weight <- array(NA,dim=c(nrow(para.space),NREPLIC))
BS.W2.weight <- array(NA,dim=c(nrow(para.space),NREPLIC))
BS.W3.weight <- array(NA,dim=c(nrow(para.space),NREPLIC))

w.avg.MMIV.num <- w.avg.MMIV.den.sqrbias <- array(NA,dim=c(nrow(para.space),NREPLIC))


 for (ipar in 1:nrow(para.space)) {   #non-parallel version, if trying to parallel #DGPs
  # foreach (ipar=1:nrow(para.space), .combine=rbind, .inorder=TRUE) {   #parallel, SEEMS not successful
    
     par <- para.space[ipar, ]
      c <- para.space[ipar, 1]
     c2 <- para.space[ipar, 2]
     c3 <- para.space[ipar, 3]
     
# random functions drawing (Y,X,Z.excl)
rfn.JTPA1 <- function(n,beta.fn) {
  if (missing(n)) stop("n missing from rfn()")
  Z.excl <- 0 + as.numeric(runif(n)>0.5)
  U <- runif(n)
  D <- 0 + (Z.excl*as.numeric(runif(n)<0.5+c*(U-0.5)))

  Y1 <- rowSums(t(beta.fn(U))*cbind(1,D))
  return(list(Y=cbind(Y1,D),X=array(1,dim=c(n,1)),Z.excl=matrix(Z.excl,nrow=n),U=U))
}

if (c3==0) {
  betafn <- function(tau)rbind(60+qchisq(tau,3),    c2*100*tau^4)
} else {
  betafn <- function(tau)rbind(60+qt(tau,c3), 0)
}



DGP <- list(DGPid="1",
            structure='iid',
            h.init.fn=function(tau,n) 0.0001, 
            rfn=rfn.JTPA1,ns=NS,taus=TAUs,dB=2, 
            beta.fn=betafn
)

# Output the parameter values
cat(sprintf(paste0("ipar=%d, NREPLIC=%d, n=%d, tau=%g, H=%g"),
            ipar, NREPLIC, n, tau, H), 
    file=OUTFILE, sep="\n",append=TRUE)

# Allocate storage variables
b.hats <- data.frame(MM=rep(NA,NREPLIC), 
                     SEE=NA,
                     IV=NA, QR=NA, 
                     AVG.MMIV=NA, AGG.MMIV=NA, AVG.MMQR=NA, AGG.MMQR=NA, 
                     AVG.BS.MSE=NA, AVG.BS.RMSE=NA, AVG.BS.RRMSE=NA
                     )

b.hats.avg <- array(NA,dim=c(13071,NREPLIC))
b.hats.avg.weights <- cbind(b.hats.avg, weights.new)
w.avg.MMIV <- w.avg.MMQR <- data.frame(weight=rep(NA,NREPLIC))

cons.sig11 <- cons.sig12 <- cons.sig22 <- MMQR.agg.sig11 <- MMQR.agg.sig12 <- MMQR.agg.sig22 <- data.frame(weight=rep(NA,NREPLIC))
MMIV.agg.sig11 <- MMIV.agg.sig12 <- MMIV.agg.sig22 <- data.frame(weight=rep(NA,NREPLIC))

AVGBS.MSE.weights <- AVGBS.RMSE.weights <- AVGBS.RRMSE.weights <- data.frame(w1.new=rep(NA,NREPLIC), w2.new=NA, w3.new=NA)





cat(sprintf("ipar%04d\n",ipar)) #progress update
Sys.time()-start.time 

  # for (irep in 1:NREPLIC) {  #non-parallel version, uncomment, if trying to parallel replications.
    bp.raw <- foreach (irep=1:NREPLIC, .combine=cbind, .inorder=TRUE) %dorng% {   #non-parallel, comment 

      
      # comment in nonparallel; uncommented in parallel (begin)
      require('quantreg')
      library("MASS") # for mvrnorm()
      library("ks")
      library("kdensity")
      library("GenSA")
      # comment in nonparallel; uncommented in parallel (end)
      
      

  # sink("log.txt", append=TRUE)
  cat(sprintf("irep%04d\n",irep)) #progress update
  

  DGP <- list(DGPid="1",
              structure='iid',
              h.init.fn=function(tau,n) 0.0001, 
              rfn=rfn.JTPA1,ns=NS,taus=TAUs,dB=2, 
              beta.fn=betafn
  )
  

  
  Z.excl <- 0 + as.numeric(runif(n)>0.5)

  U <- runif(n)

  D <- 0 + (Z.excl*as.numeric(runif(n)<0.5+c*(U-0.5)))

  if (c3==0) {
    beta.fn <- function(tau)rbind(60+qchisq(tau,3),    c2*100*tau^4)
  } else {
    beta.fn <- function(tau)rbind(60+qt(tau,c3), 0)
  }
  

  
  Y1 <- rowSums(t(beta.fn(U))*cbind(1,D))

  dat <- list(Y=cbind(Y1,D),X=array(1,dim=c(n,1)),Z.excl=matrix(Z.excl,nrow=n),U=U)



  Y <- as.matrix(dat$Y[1:n,1],nrow=n)
  X.excl <- as.matrix(dat$X,nrow=n)
  Z.excl <- as.matrix(dat$Z.excl,nrow=n)
  demean.Z.excl<- Z.excl-colMeans(Z.excl)
  D <- as.matrix(dat$Y[1:n,2],nrow=n)
  Z <- cbind(X.excl, Z.excl)
  X <- cbind(X.excl, D)
  
  dZ <- ncol(Z) #number of instruments
  dX <- ncol(X) #number of regressors
  dD <- ncol(D)
  Y.gmmq <- cbind(Y,D)
  

  n11 <- sum(D)
  n00 <- sum((1-D)*(1-Z.excl))
  n01 <- sum((1-D)*Z.excl)
  
  p11 <- n11/n
  p00 <- n00/n
  p01 <- n01/n
  

  
  # Compute estimators and store
  # 
  # # Kaplan and Sun (2017), IVQR-SEE with "optimal" bandwidth (no h_in specified)
  tmp.SEE <- ivqr_see(p = tau, Y = Y, X = X, Z = Z)


  H.SEE <- H <- tmp.SEE$h # for SEE,  b.SEE has the same sequence with X and Z. i.e. if X and Z are (intercept, slope), then b.SEE is (intercept,slope);if X and Z are (slope, intercept), then b.SEE is (slope, intercept)
  # here b/c X and Z are (intercept, slope), then b.SEE is (intercept, slope). so we're interested in b.SEE[2], the slope.
  H.LARGE <- H.SEE*n^(-2/35)


  b.SEE <- tmp.SEE$b
  b.hats$SEE[irep] <- b.SEE[2]
  
  
  # 2SLS
  PZ <- Z %*% solve(t(Z)%*%Z) %*% t(Z)
  b.2sls <- solve(t(X)%*%PZ%*%X) %*% t(X)%*%PZ%*%Y
  b.hats$IV[irep] <- b.2sls[2]
  # Z is on (intercept, slope), X is on (intercept, slope), so b.2SLS is on (intercept, slope); we're interested in slope, thus look at b.2sls[2]

  # 
  # QR  
  suppressWarnings(b.qr <- rq(formula=Y~X,tau=tau,data=data.frame(Y=dat$Y[1:n,1],X=dat$Y[1:n,2]))$coefficients[2])

  suppressWarnings(b.hats$QR[irep] <- b.qr)


  # 

  Lambda <- function(y, x, b){y-x%*%b } # require X and b has the same sequence. If X is (intercept, slope), b must be (intercept slope). Notice that X=(X.excl, D), but most b(b.gmmq, b.ivqr.gmm, b.onestep) are (slope, intercept), so need to adjust X or use Lambda.gmmq w/ Y.gmmq, X.excl
  Lambda.derivative <- function(y,x,b) -x # lambda.derivative fn gives the same sequence of X. If X is (intercept, slope) (as defined here, X=(X.excl, D)), the lambda.derivative gives (intercept, slope); if wanna labdha.derives generates (slope, intercept), then should use Lambda.derivative.gmmq or use Lambda.derivative (X=cbind(D,X.excl))
  Lambda.gmmq <- function(y,x,b) {y[,1]-cbind(y[,-1],x)%*%b} # b must be (slope, intercept) sequence. need to check this, when calling lambda.gmmq fn.
  Lambda.derivative.gmmq <- function(y,x,b) {-cbind(y[,-1],x)} # lambda.derivative.gmmq gives (slope, intercept) sequence.
  
  
  Itilde <- Itilde.KS17
  Itilde.deriv <- Itilde.deriv.KS17
  
  Gfn <- function(v,h) {      Itilde(v/h)       }
  Gpfn <- function(v,h) {     Itilde.deriv(v/h) }
  

  
  
  ####################################################################################################

 
  ####################################################################################################
  ## 1 = IVQR    "conservative" . only do the "conservative" once. "Conservative" is unsmoothed IVQR. (i.e. unsmoothed 2step GMM)
  # MM: gmmq(h=0) is exactID case, & smoothed case with h=0.
  # use unsmoothed or smoothed(h=0): just use "conservative"=gmmq(h=0) for convenience for now.  
  gmmq.est <- gmmq(tau=tau, Y=Y.gmmq, X=X.excl, Z.excl=Z.excl, dB=dB, 
                   Lambda=Lambda.gmmq, 
                   Lambda.derivative=Lambda.derivative.gmmq, 
                   h=0, VERBOSE=FALSE, RETURN.Z=FALSE )  
  h.gmmq <- gmmq.est$h # do the gmmq is to get h.gmmq; h.gmmq is used for cov.mtx; cov.mtx is used for weighting matrix in GMM method to get "aggressive"; i.e. gmmq()-->h.gmmq-->cov.mtx-->weight.mtx in GMM of "aggressive".

  h.small <- h.gmmq
  
  b.conservative <- b.gmmq <- gmmq.est$b # b.gmmq=(slope, intercept) (interested in slope)
  b.hats$MM[irep] <-b.conservative[1]  
  

  # b.gmmq is "conservative" estimates in this example(i.e. b.conservative=b.gmmq). 
  # h.gmmq is used for cov.mtx
  # h=0 is used for "conservative" and "aggressive" estimation. (i.e. use h=0 to get b.conservatie and b.aggressive.)
  


  Uhat <- Y- cbind(D,X.excl)%*%b.conservative
  
  U11<-Uhat[which(D==1)]
  U01<-Uhat[which((1-D)*Z.excl==1)]
  U00<-Uhat[which((1-D)*(1-Z.excl)==1)]


  d11 <- density(U11)
  d01 <- density(U01)
  d00 <- density(U00)
  

  D11 <- kdensity(U11, kernel = "gaussian")
  D01 <- kdensity(U01, kernel = "gaussian")
  D00 <- kdensity(U00, kernel = "gaussian")
  

  f00 <- D00(0)
  f01 <- D01(0)
  f11 <- D11(0)
  
  num.Kato <- 4.5*(f00*p00+2*f01*p01+4*f11*p11)

  fhatU11<- kdde(x=U11, deriv.order=2) 
  fhatU01<- kdde(x=U01, deriv.order=2) 
  fhatU00<- kdde(x=U00, deriv.order=2) 

  f2_11 <- predict(fhatU11,x=0)
  f2_01 <- predict(fhatU01,x=0)
  f2_00 <- predict(fhatU01,x=0)
  
  
  den.Kato <- (f2_00*p00+2*f2_01*p01+4*f2_11*p11)^2
  
  h.Kato <- (num.Kato/den.Kato/n)^(1/5)
  

  
  num.Kato.aggMMQR <- 4.5*(f00*p00+2*f01*p01+6*f11*p11)
  den.Kato.aggMMQR <- (f2_00*p00+2*f2_01*p01+6*f2_11*p11)^2
  h.Kato.aggMMQR <- (num.Kato.aggMMQR/den.Kato.aggMMQR/n)^(1/5)
  
  
  L <- matrix(Lambda.gmmq(y=Y.gmmq,x=X.excl,b=b.conservative), ncol=1)
  Lp <- Lambda.derivative.gmmq(y=Y.gmmq, x=X.excl, b=b.conservative) # Lp is in (slope, intercept) sequence, consistent with jac. 
  
  
  gni <- cbind(Z.excl, X.excl)*repmat((Gfn(-L,h.small)-tau),1,dZ) # here gni is (slope, intercept), (Z.excl, X.excl)
  g.bar <- colMeans(gni) # g.bar is in (slope, intercept) sequence.
  
  h.use.con.sig <- h.Kato
  jac <- t(cbind(Z.excl, X.excl))%*%(Lp*repmat(Gpfn(-L,h.use.con.sig),1,dB))/(-h.use.con.sig)/n # add divide n # jac is (slope, intercept), (Z.excl, X.excl)
  
  
  Omega.bar <- t(gni)%*%gni/n-g.bar%*%t(g.bar) # Omega.bar is in (slope, intercept) sequence.

  sig.conservative <- solve(t(jac)%*%solve(Omega.bar)%*%jac)   # here sig.conservative is (slope, intercept) matrix. (Z.excl, X.excl)
  
  # Weighting.mtx (in GMM criterion) is inverse of Omega.
  # Omega is 1/n sum(gg')-(g.bar)(g.bar)'
  # Sigma in weight.formula is inverse of ( t(jac) %*% weighting.mtx %*% jac ).
  
  ####################################################################################################
  # Averaging TYPE #1, (IVQR, QR)
  # STAGE 2 = (IVQR,QR)    "aggressive"
  # estimate cov.mtx using b.conservative and full GenSA moments.
  L.agg.MMQR <- matrix(Lambda.gmmq(y=Y.gmmq, x=X.excl, b=b.conservative), ncol=1)  
  
  gni.agg.MMQR <- cbind(D, Z.excl, X.excl)*repmat((Gfn(-L.agg.MMQR, h.small)-tau),1,dZ+ncol(D))  # gni.agg.MMQR sequence is (D, Z.excl, X.excl) 
  g.bar.agg.MMQR <- colMeans(gni.agg.MMQR)                                                   # g.bar.agg.MMQR sequence is (D, Z.excl, X.excl) 
  cov.mtx.agg.MMQR <- t(gni.agg.MMQR)%*%gni.agg.MMQR/n-g.bar.agg.MMQR%*%t(g.bar.agg.MMQR) # cov.mat, weight.mtx sequence is (D, Z.excl, 1)
  inv.cov.mtx.agg.MMQR <- solve(cov.mtx.agg.MMQR)
  
  # use inverse of cov.mtx as the weighting matrix to construct the GMM criterion. run GenSA with GMM. get b
  tmp.MMQR.1step <- ivqr.gmm(Y = Y, X.excl = X.excl, Z.excl =cbind(D, Z.excl), D = D, tau = tau, h = h.small, dB = dB, max.time=2, upper=rep(100000,dB), lower=rep(-100000,dB),
                      weight.mtx=inv.cov.mtx.agg.MMQR, structure = 'iid', LRV.kernel = 'QS', b.init=b.conservative ) 
  b.MMQR.1step <- tmp.MMQR.1step$b

  

  
  # store b.aggressive
  b.agg.MMQR <- b.MMQR.1step 
  b.hats$AGG.MMQR[irep] <-b.agg.MMQR[1]  
  


  
  
  ########################################################################################################
  ##################### new way to compute sigma.aggressive ###### evaluated at b.conservative. 
  # use b.conservative to calculate sigma.aggressive.
  L.agg.MMQR.sig <- matrix(Lambda.gmmq(y=Y.gmmq, x=X.excl, b=b.conservative), ncol=1)
  Lp.agg.MMQR.sig <- Lambda.derivative.gmmq(y=Y.gmmq, x=X.excl, b=b.conservative)  # note. b. is (slope, intercept), thus need to use Lambda.derivative.gmmq instead of Lambda.derivative, bc Lambda.derivative.gmmq give X in (D, X.excl) order; Lambda.derivative gives X in (X.excl, D) order.

  gni.agg.MMQR.sig <- cbind(D, Z.excl, X.excl)*repmat((Gfn(-L.agg.MMQR.sig, h.small)-tau),1,dZ+ncol(D)) #gni2.sig is in (D, Z.excl, X.excl) order
  g.bar.agg.MMQR.sig <- colMeans(gni.agg.MMQR.sig)
  
  
  
  h.use.MMQR.sig <- h.Kato
  
  jac.agg.MMQR.sig <- rbind(t(D)%*%(Lp.agg.MMQR.sig*repmat(Gpfn(-L.agg.MMQR.sig, h.use.MMQR.sig),1,dB))/(-h.use.MMQR.sig)/n, t(cbind(Z.excl, X.excl))%*%(Lp.agg.MMQR.sig*repmat(Gpfn(-L.agg.MMQR.sig, h.use.MMQR.sig),1,dB))/(-h.use.MMQR.sig)/n) 
  
  Omega.bar.agg.MMQR.sig <- t(gni.agg.MMQR.sig)%*%gni.agg.MMQR.sig/n-g.bar.agg.MMQR.sig%*%t(g.bar.agg.MMQR.sig) # iid

  sig.agg.MMQR <- tryCatch(solve(t(jac.agg.MMQR.sig)%*%solve(Omega.bar.agg.MMQR.sig)%*%jac.agg.MMQR.sig),
                           error=function(w)NA)
  if (is.na(sig.agg.MMQR[1])) next
  ##################### new way to compute sigma.aggressive ###### evaluated at b.conservative. 
  ########################################################################################################
  
  
  
  
  # averaging
  #Ups = matrix(c(1,0,0,0),2,2), b/c b1,b2 use b.gmmq. b.gmmq[1]=slope

  TR1 <- Trace(Ups%*%(sig.conservative-sig.agg.MMQR))
  if (TR1 <0) {
    w.avg.MMQR[irep,] <- MMQRweight <- 0
  }else{
    w.avg.MMQR[irep,] <- MMQRweight <- max(0, min(1, TR1/(n*t(b.agg.MMQR-b.conservative)%*%Ups%*%(b.agg.MMQR-b.conservative)+TR1)))
  }
  
  b.AVG.MMQR <- repmat((1-MMQRweight),dB,1)*b.conservative+repmat(MMQRweight,dB,1)*b.agg.MMQR
  b.hats$AVG.MMQR[irep] <- b.AVG.MMQR[1]
  
  
  
  
  cons.sig11[irep,] <- sig.conservative[1,1]
  cons.sig12[irep,] <- sig.conservative[1,2]
  cons.sig22[irep,] <- sig.conservative[2,2]
  MMQR.agg.sig11[irep,] <- sig.agg.MMQR[1,1]
  MMQR.agg.sig12[irep,] <- sig.agg.MMQR[1,2]
  MMQR.agg.sig22[irep,] <- sig.agg.MMQR[2,2]

  
  
  ####################################################################################################
  # Averaging TYPE #2 (IVQR, IV)    "conservative"
  # Conservative part is the same as before.

  ######################################################################################
  # 22: compute "aggressive" using GenSA. get b2 and sig2.      "aggressive"
 

  # estimate cov.mtx using b.conservative and full GenSA moments.
  L.agg.MMIV <- matrix(Lambda.gmmq(y=Y.gmmq, x=X.excl, b=b.conservative), ncol=1)  
  gni.agg.MMIV <- cbind(demean.Z.excl*L.agg.MMIV, cbind(Z.excl, X.excl)*repmat((Gfn(-L.agg.MMIV, h.small)-tau),1,dZ)) # gni sequence is (demean.Z.excl, Z.excl, X.excl) 
  g.bar.agg.MMIV <- colMeans(gni.agg.MMIV) 
  cov.mtx.agg.MMIV <- t(gni.agg.MMIV)%*%gni.agg.MMIV/n-g.bar.agg.MMIV%*%t(g.bar.agg.MMIV) # cov.mat, weight.mtx sequence is (demean.Z.excl, Z.excl, X.excl) 
  inv.cov.mtx.agg.MMIV <- solve(cov.mtx.agg.MMIV)
  
  # use inverse of cov.mtx as the weighting matrix to construct the GMM criterion. run GenSA with GMM. get b
  tmp.MMIV.1step <- ivqr.iv.aggressive(Y = Y, X.excl = X.excl, Z.excl = Z.excl, D = D, tau = tau, h = h.small, dB = dB, max.time=2, upper=rep(100000,dB), lower=rep(-100000,dB),
                              weight.mtx=inv.cov.mtx.agg.MMIV, structure = 'iid', LRV.kernel = 'QS', b.init=b.conservative  ) 

  b.MMIV.1step <- tmp.MMIV.1step$b
  
  
  # store b.aggressive
  b.agg.MMIV <- b.MMIV.1step 
  b.hats$AGG.MMIV[irep] <-b.agg.MMIV[1]  
  

    
  # ########################################################################################################
  # ##################### NEW way to compute sigma.aggressive ###### evaluated at b.conservative which is correct 
  # use b.conservative to calculate sigma2. i.e. sigma.aggressive.
  L.agg.MMIV.sig <- matrix(Lambda.gmmq(y=Y.gmmq,x=X.excl,b=b.conservative), ncol=1)
  Lp.agg.MMIV.sig <- Lambda.derivative.gmmq(y=Y.gmmq, x=X.excl, b=b.conservative)  # note. b. is (slope, intercept), thus need to use Lambda.derivative.gmmq instead of Lambda.derivative, bc Lambda.derivative.gmmq give X in (D, X.excl) order; Lambda.derivative gives X in (X.excl, D) order.
  # Lp2.sig is in (D, X.excl) order.
  
  gni.agg.MMIV.sig <- cbind(demean.Z.excl*L.agg.MMIV.sig, cbind(Z.excl, X.excl)*repmat((Gfn(-L.agg.MMIV.sig, h.small)-tau),1,dZ))     #gni is in (demean.Z.excl, Z.excl, X.excl) order
  g.bar.agg.MMIV.sig <- colMeans(gni.agg.MMIV.sig)
  
  h.use.MMIV.sig <- h.Kato
  jac.agg.MMIV.sig <- rbind(t(demean.Z.excl)%*%(Lp.agg.MMIV.sig)/n, t(cbind(Z.excl, X.excl))%*%(Lp.agg.MMIV.sig*repmat(Gpfn(-L.agg.MMIV.sig, h.use.MMIV.sig),1,dB))/(-h.use.MMIV.sig)/n) 
  
  Omega.bar.agg.MMIV.sig <- t(gni.agg.MMIV.sig)%*%gni.agg.MMIV.sig/n-g.bar.agg.MMIV.sig%*%t(g.bar.agg.MMIV.sig) # iid

  sig.agg.MMIV <- tryCatch(solve(t(jac.agg.MMIV.sig)%*%solve(Omega.bar.agg.MMIV.sig)%*%jac.agg.MMIV.sig),
                           error=function(w)NA)
  if (is.na(sig.agg.MMIV[1])) next
  # ##################### NEW way to compute sigma.aggressive ###### evaluated at b.conservative which is correct 
  # ########################################################################################################
  
  
  
  # "averaging"
  

  
  TR2 <- Trace(Ups%*%(sig.conservative-sig.agg.MMIV))
  
  if (TR2<0){
    w.avg.MMIV[irep, ] <- MMIVweight <- 0
    w.avg.MMIV.num[ipar, irep] <- "neg"
    w.avg.MMIV.den.sqrbias[ipar, irep] <- "den neg"
  }else{
    w.avg.MMIV[irep,] <- MMIVweight <- max(0, min(1, TR2/(n*t(b.agg.MMIV-b.conservative)%*%Ups%*%(b.agg.MMIV-b.conservative)+TR2)))
    w.avg.MMIV.num[ipar, irep] <- TR2
    w.avg.MMIV.den.sqrbias[ipar, irep] <- n*t(b.agg.MMIV-b.conservative)%*%Ups%*%(b.agg.MMIV-b.conservative)
  }
  b.AVG.MMIV <- repmat((1-MMIVweight),dB,1)*b.conservative+repmat(MMIVweight,dB,1)*b.agg.MMIV
  b.hats$AVG.MMIV[irep] <- b.AVG.MMIV[1]
  
  
  

  MMIV.agg.sig11[irep,] <- sig.agg.MMIV[1,1]
  MMIV.agg.sig12[irep,] <- sig.agg.MMIV[1,2]
  MMIV.agg.sig22[irep,] <- sig.agg.MMIV[2,2]
  
  
  
  
  

  if (BS.FLAG){
  # bootstrap
  # AVG.BS  :(for each simulation sample, run bootstrap and pick the optimal weight based on miniminizing RMSE/rRMSE.)
  b.hats.original <- cbind(b.hats$MM[irep], b.hats$IV[irep], b.hats$QR[irep] )
  b.IVQR <- b.hats$MM[irep]

  tmp.BS <- AVGBS(Y = Y, D = D, X.excl = X.excl, Z.excl = Z.excl, tau = tau, n=n, h = h.small, dB = dB, BREPLIC=BREPLIC, reportGMM=reportGMM, weights.new=weights.new,
                  b.hats.original=b.hats.original, b.IVQR=b.IVQR, DGP=DGP)

  b.hats$AVG.BS.MSE[irep] <- tmp.BS$b.AVGBS.MSE[1]
  b.hats$AVG.BS.RMSE[irep] <- tmp.BS$b.AVGBS.RMSE[1]
  b.hats$AVG.BS.RRMSE[irep] <- tmp.BS$b.AVGBS.RRMSE[1]


  # optimal weights used by AVG.BS
  AVGBS.MSE.weights[irep,] <- tmp.BS$o.weight.AVGBS.MSE
  AVGBS.RMSE.weights[irep,] <- tmp.BS$o.weight.AVGBS.RMSE
  AVGBS.RRMSE.weights[irep,] <- tmp.BS$o.weight.AVGBS.RRMSE
  
  BS.W1.weight[ipar, irep] <- tmp.BS$o.weight.AVGBS.RRMSE[1]
  BS.W2.weight[ipar, irep] <- tmp.BS$o.weight.AVGBS.RRMSE[2]
  BS.W3.weight[ipar, irep] <- tmp.BS$o.weight.AVGBS.RRMSE[3]
  
  }
  
  SLI.full<-rbind(unlist(t(b.hats[ipar,])), MMIVweight, MMQRweight, BS.W1.weight[ipar, irep], BS.W2.weight[ipar, irep], BS.W3.weight[ipar, irep], TR2, n*t(b.agg.MMIV-b.conservative)%*%Ups%*%(b.agg.MMIV-b.conservative))
  SLI.old<-rbind(b.conservative[1], b.SEE[2], b.2sls[2], rq(formula=Y~X,tau=tau,data=data.frame(Y=dat$Y[1:n,1],X=dat$Y[1:n,2]))$coefficients[2], b.AVG.MMIV[1], b.agg.MMIV[1], b.AVG.MMQR[1], b.agg.MMQR[1], c(MMIVweight), c(MMQRweight), Trace(Ups%*%(sig.conservative-sig.agg.MMIV)), n*t(b.agg.MMIV-b.conservative)%*%Ups%*%(b.agg.MMIV-b.conservative))
  
  if (BS.FLAG){
    SLI<-rbind(c(b.conservative[1]), c(b.SEE[2]), c(b.2sls[2]), c(rq(formula=Y~X,tau=tau,data=data.frame(Y=dat$Y[1:n,1],X=dat$Y[1:n,2]))$coefficients[2]), c(b.AVG.MMIV[1]), c(b.agg.MMIV[1]), c(b.AVG.MMQR[1]), c(b.agg.MMQR[1]), tmp.BS$b.AVGBS.MSE[1], tmp.BS$b.AVGBS.RMSE[1], tmp.BS$b.AVGBS.RRMSE[1], BS.W1.weight[ipar, irep], BS.W2.weight[ipar, irep], BS.W3.weight[ipar, irep], c(MMIVweight), c(MMQRweight), c(Trace(Ups%*%(sig.conservative-sig.agg.MMIV))), c(n*t(b.agg.MMIV-b.conservative)%*%Ups%*%(b.agg.MMIV-b.conservative)))
  }else{
    SLI<-rbind(c(b.conservative[1]), c(b.SEE[2]), c(b.2sls[2]), c(rq(formula=Y~X,tau=tau,data=data.frame(Y=dat$Y[1:n,1],X=dat$Y[1:n,2]))$coefficients[2]), c(b.AVG.MMIV[1]), c(b.agg.MMIV[1]), c(b.AVG.MMQR[1]), c(b.agg.MMQR[1]), "NA", "NA", "NA", "NA", "NA", "NA", c(MMIVweight), c(MMQRweight), c(Trace(Ups%*%(sig.conservative-sig.agg.MMIV))), c(n*t(b.agg.MMIV-b.conservative)%*%Ups%*%(b.agg.MMIV-b.conservative)))
  }


  matrix(SLI,ncol=1)

  
} # END for loop over irep/NREPLIC # also the end of foreach loop
  

################################ for each DGP, finish its 1000 (or large NREPLIC) replications ############################### now begin to compute this DGP's results.
###########################################################################################################################################################
###########################################################################################################################################################
###########################################################################################################################################################

    
b.hats$MM <- t(t(as.numeric(bp.raw[1, ])))  # if non-parallel, comment
b.hats$SEE <- t(t(as.numeric(bp.raw[2, ])))  # if non-parallel, comment
b.hats$IV <- t(t(as.numeric(bp.raw[3, ])))  # if non-parallel, comment
b.hats$QR <- t(t(as.numeric(bp.raw[4, ]))) # if non-parallel, comment
b.hats$AVG.MMIV <- t(t(as.numeric(bp.raw[5, ])))  # if non-parallel, comment
b.hats$AGG.MMIV <- t(t(as.numeric(bp.raw[6, ])))  # if non-parallel, comment
b.hats$AVG.MMQR <- t(t(as.numeric(bp.raw[7, ])))  # if non-parallel, comment
b.hats$AGG.MMQR <- t(t(as.numeric(bp.raw[8, ])))  # if non-parallel, comment

if (BS.FLAG){
  b.hats$AVG.BS.MSE <- t(t(as.numeric(bp.raw[9, ])))  # if non-parallel, comment
  b.hats$AVG.BS.RMSE <- t(t(as.numeric(bp.raw[10, ])))  # if non-parallel, comment
  b.hats$AVG.BS.RRMSE <- t(t(as.numeric(bp.raw[11, ])))  # if non-parallel, comment
}else{
  b.hats$AVG.BS.MSE <- array(NA,dim=c(NREPLIC,1))  # if non-parallel, comment
  b.hats$AVG.BS.RMSE <- array(NA,dim=c(NREPLIC,1))  # if non-parallel, comment
  b.hats$AVG.BS.RRMSE <- array(NA,dim=c(NREPLIC,1))  # if non-parallel, comment
}


sum(is.na(b.hats$AVG.MMQR))
sum(is.na(b.hats$AVG.MMIV))

# DGP

# Compute results and save to OUTFILE
beta0s <- DGP$beta.fn(DGP$taus)

B.TRUE <- beta0s[2]
# beta0s and B.TRUE are unrelated with irep replications. 



b.hats.mean <- colMeans(b.hats, na.rm=TRUE)
b.hats.bias <- b.hats.mean - B.TRUE

b.hats.sd <- apply(X = b.hats, MARGIN = 2, FUN = sd, na.rm=TRUE)
b.hats.median.bias <- apply(X = b.hats, MARGIN = 2, FUN = median, na.rm=TRUE) - B.TRUE

b.hats.RMSE <- sqrt(b.hats.bias^2 + b.hats.sd^2)
b.hats.IQR <- apply(X = b.hats, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.hats, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.hats.rRMSE <- sqrt(b.hats.median.bias^2 + (b.hats.IQR/1.349)^2)



a<-t(as.vector(b.hats.rRMSE)) 

b.hats.rRMSE.par[ipar, ] <- a 
b.hats.RMSE.par[ipar, ] <- t(as.matrix(b.hats.RMSE)) 


# to report relative results: #relIV=NA, relQR=NA, relAVGGMM1.MMQR=NA, relAVGGMM2.MMIV=NA
b.hats.rRMSE.par.rel[ipar, 'MM' ] <- b.hats.rRMSE.par[ipar, 'MM'] 
b.hats.rRMSE.par.rel[ipar, 'relSEE' ] <- b.hats.rRMSE.par[ipar, 'SEE'] /b.hats.rRMSE.par[ipar, 'MM'] 
b.hats.rRMSE.par.rel[ipar, 'relAVG.BS.MSE' ] <- b.hats.rRMSE.par[ipar, 'AVG.BS.MSE'] /b.hats.rRMSE.par[ipar, 'MM'] 
b.hats.rRMSE.par.rel[ipar, 'relAVG.BS.RMSE' ] <- b.hats.rRMSE.par[ipar, 'AVG.BS.RMSE'] /b.hats.rRMSE.par[ipar, 'MM'] 
b.hats.rRMSE.par.rel[ipar, 'relAVG.BS.RRMSE' ] <- b.hats.rRMSE.par[ipar, 'AVG.BS.RRMSE'] /b.hats.rRMSE.par[ipar, 'MM'] 

b.hats.rRMSE.par.rel[ipar, 'relIV' ] <- b.hats.rRMSE.par[ipar, 'IV'] /b.hats.rRMSE.par[ipar, 'MM'] 
b.hats.rRMSE.par.rel[ipar, 'relQR' ] <- b.hats.rRMSE.par[ipar, 'QR'] /b.hats.rRMSE.par[ipar, 'MM'] 
b.hats.rRMSE.par.rel[ipar, 'relAVG.MMQR' ] <- b.hats.rRMSE.par[ipar, 'AVG.MMQR'] /b.hats.rRMSE.par[ipar, 'MM'] 
b.hats.rRMSE.par.rel[ipar, 'relAVG.MMIV' ] <- b.hats.rRMSE.par[ipar, 'AVG.MMIV'] /b.hats.rRMSE.par[ipar, 'MM'] 

b.hats.rRMSE.par.rel[ipar, 'relAGG.MMQR' ] <- b.hats.rRMSE.par[ipar, 'AGG.MMQR'] /b.hats.rRMSE.par[ipar, 'MM'] 
b.hats.rRMSE.par.rel[ipar, 'relAGG.MMIV' ] <- b.hats.rRMSE.par[ipar, 'AGG.MMIV'] /b.hats.rRMSE.par[ipar, 'MM'] 



b.hats.RMSE.par.rel[ipar, 'MM' ] <- b.hats.RMSE.par[ipar, 'MM'] 
b.hats.RMSE.par.rel[ipar, 'relSEE' ] <- b.hats.RMSE.par[ipar, 'SEE'] /b.hats.RMSE.par[ipar, 'MM'] 
b.hats.RMSE.par.rel[ipar, 'relAVG.BS.MSE' ] <- b.hats.RMSE.par[ipar, 'AVG.BS.MSE'] /b.hats.RMSE.par[ipar, 'MM'] 
b.hats.RMSE.par.rel[ipar, 'relAVG.BS.RMSE' ] <- b.hats.RMSE.par[ipar, 'AVG.BS.RMSE'] /b.hats.RMSE.par[ipar, 'MM'] 
b.hats.RMSE.par.rel[ipar, 'relAVG.BS.RRMSE' ] <- b.hats.RMSE.par[ipar, 'AVG.BS.RRMSE'] /b.hats.RMSE.par[ipar, 'MM'] 

b.hats.RMSE.par.rel[ipar, 'relIV' ] <- b.hats.RMSE.par[ipar, 'IV'] /b.hats.RMSE.par[ipar, 'MM'] 
b.hats.RMSE.par.rel[ipar, 'relQR' ] <- b.hats.RMSE.par[ipar, 'QR'] /b.hats.RMSE.par[ipar, 'MM'] 
b.hats.RMSE.par.rel[ipar, 'relAVG.MMQR' ] <- b.hats.RMSE.par[ipar, 'AVG.MMQR'] /b.hats.RMSE.par[ipar, 'MM'] 
b.hats.RMSE.par.rel[ipar, 'relAVG.MMIV' ] <- b.hats.RMSE.par[ipar, 'AVG.MMIV'] /b.hats.RMSE.par[ipar, 'MM'] 

b.hats.RMSE.par.rel[ipar, 'relAGG.MMIV' ] <- b.hats.RMSE.par[ipar, 'AGG.MMIV'] /b.hats.RMSE.par[ipar, 'MM'] 
b.hats.RMSE.par.rel[ipar, 'relAGG.MMQR' ] <- b.hats.RMSE.par[ipar, 'AGG.MMQR'] /b.hats.RMSE.par[ipar, 'MM'] 


w.avg.MMIV[,1] <- t(t(as.numeric(bp.raw[15,])))  # if non-parallel, comment
w.avg.MMQR[,1] <- t(t(as.numeric(bp.raw[16,])))  # if non-parallel, comment


MMIV.weight.test <- as.matrix(w.avg.MMIV[,1], nrow=1, ncol=NREPLIC)
MMIV.weight.new[ipar, ] <- as.numeric(t(as.numeric(MMIV.weight.test)))

MMQR.weight.test <- as.matrix(w.avg.MMQR[,1], nrow=1, ncol=NREPLIC)
MMQR.weight.new[ipar, ] <- t(MMQR.weight.test)



#####################################################################################################################################################
### output the b.hats.rRMSE.rel table
rel.rRMSE[ipar,] <- cbind(b.hats.rRMSE.par.rel[ipar,], c, c2, c3)
rel.RMSE[ipar,] <- cbind(b.hats.RMSE.par.rel[ipar,], c, c2, c3)

write.csv(rel.rRMSE, file = "b.hats.rRMSE.par.rel.new.csv")
write.csv(rel.RMSE, file = "b.hats.RMSE.par.rel.new.csv")

write.csv(rel.rRMSE, file = paste0(sprintf("b.hats.rRMSE.par.rel.new_NREP%d_BREP%d_tau%.2f_n%d_DGP%d_", NREPLIC, BREPLIC, tau, n, nrow(para.space)), format(Sys.time(),"%Y_%m_%d"),".csv") )
write.csv(rel.RMSE, file = paste0(sprintf("b.hats.RMSE.par.rel.new_NREP%d_BREP%d_tau%.2f_n%d_DGP%d_",NREPLIC, BREPLIC, tau, n, nrow(para.space)), format(Sys.time(),"%Y_%m_%d"),".csv"))
####################################################################################################################################################


####################################################################################################################################################
### output the weights table.
MMQR.weight.all[ipar,] <- cbind(t(MMQR.weight.test), c, c2, c3)
write.csv(MMQR.weight.all, file = paste0(sprintf("MMQRweight_NREP%d_tau%.2f_n%d_DGP%d_", NREPLIC, tau, n, nrow(para.space)), format(Sys.time(),"%Y_%m_%d"),".csv") )

MMIV.weight.all[ipar,] <- cbind(t(MMIV.weight.test), c, c2, c3)
write.csv(MMIV.weight.all, file = paste0(sprintf("MMIVweight_NREP%d_tau%.2f_n%d_DGP%d_", NREPLIC, tau, n, nrow(para.space)), format(Sys.time(),"%Y_%m_%d"),".csv") )

#####################################################################################################

if(BS.FLAG){
  BS.W1.weight[ipar,] <- t(as.numeric(bp.raw[12,]))  # if non-parallel, comment
  BS.W2.weight[ipar,] <- t(as.numeric(bp.raw[13,]))  # if non-parallel, comment
  BS.W3.weight[ipar,] <- t(as.numeric(bp.raw[14,]))  # if non-parallel, comment
  
}else{
  BS.W1.weight[ipar,] <- array(0,dim=c(1,NREPLIC))  # if non-parallel, comment
  BS.W2.weight[ipar,] <- array(0,dim=c(1,NREPLIC))  # if non-parallel, comment
  BS.W3.weight[ipar,] <- array(0,dim=c(1,NREPLIC))  # if non-parallel, comment
}



write.csv(cbind(BS.W1.weight, para.space) , file = paste0(sprintf("BSW1_optimalweight_NREP%d_tau%.2f_n%d_DGP%d_", NREPLIC, tau, n, nrow(para.space)), format(Sys.time(),"%Y_%m_%d"),".csv") )
write.csv(cbind(BS.W2.weight, para.space) , file = paste0(sprintf("BSW2_optimalweight_NREP%d_tau%.2f_n%d_DGP%d_", NREPLIC, tau, n, nrow(para.space)), format(Sys.time(),"%Y_%m_%d"),".csv") )
write.csv(cbind(BS.W3.weight, para.space) , file = paste0(sprintf("BSW3_optimalweight_NREP%d_tau%.2f_n%d_DGP%d_", NREPLIC, tau, n, nrow(para.space)), format(Sys.time(),"%Y_%m_%d"),".csv") )

w.avg.MMIV.num[ipar, ] <- t(as.numeric(bp.raw[17,]))
w.avg.MMIV.den.sqrbias[ipar,] <- t(as.numeric(bp.raw[18,]))

write.csv(cbind(w.avg.MMIV.num, para.space) , file = paste0(sprintf("MMIV_NUM_NREP%d_tau%.2f_n%d_DGP%d_", NREPLIC, tau, n, nrow(para.space)), format(Sys.time(),"%Y_%m_%d"),".csv") )
write.csv(cbind(w.avg.MMIV.den.sqrbias, para.space) , file = paste0(sprintf("MMIV_DENsqrbias_optimalweight_NREP%d_tau%.2f_n%d_DGP%d_", NREPLIC, tau, n, nrow(para.space)), format(Sys.time(),"%Y_%m_%d"),".csv") )


####################################################################################################################################################

} # end of ipar loop for ipar


################################ NOW finished all ipar=#DPGs and irep=#NERPLIC computing of b.hats, b.hats.rRMSE.par.rel.new, weights.##############################
####################################################################################################################################################################
####################################################################################################################################################################


# report matrix tables:b.hats.rRMSE.Endo.rel, b.hats.RMSE.Endo.rel, b.hats.rRMSE.het.rel, b.hats.RMSE.het.rel
write.csv(b.hats.rRMSE.par.rel, file = "b.hats.rRMSE.par.rel.csv")
write.csv(b.hats.RMSE.par.rel, file = "b.hats.RMSE.par.rel.csv")



############################################################################################################################
# finally report results into .txt file.
############################################################################################################################
## JTPA (ALL ESTIMATES) 14 DGP TABLE ####################  BEGIN  ################### 
cat("\\begin{table}[htbp]
    \\centering", file=OUTFILE, sep='', append=TRUE)

cat(sprintf("\\caption{\\label{tab:est:M1:tau=%s} Finite sample relative robust RMSE in JTPA-based simulation model M1 at %s.}
            ",
            tau, ifelse(tau==0.5, sprintf("%s quantile (median)",tau), sprintf("%s quantile",tau)) ), file=OUTFILE, sep='', append=TRUE)

cat("\\sisetup{round-precision=2,round-mode=places}
    \\begin{threeparttable}
    \\begin{tabular}{clllccclcclccc}
    \\toprule
    &          &          & & \\multicolumn{2}{c}{$\\mathrm{IVQR.2SLS}$} &               & \\multicolumn{2}{c}{$\\mathrm{IVQR.QR} $} &               &  &     \\\\
    \\cmidrule{5-6} \\cmidrule{8-9}
    $\\mathrm{DGP }$  &  $\\mathrm{Endog}$ &  $\\mathrm{Hetero}$ &  $\\mathrm{Fatail}$ &  $\\mathrm{AVG} $ & $\\mathrm{AGG}$       & $\\mathrm{IV}$ & $\\mathrm{AVG}$ & $\\mathrm{AGG}$         & $\\mathrm{QR}$ & $\\mathrm{BS} $   & $\\mathrm{SEE}$ \\\\
    \\midrule\n", file=OUTFILE, sep='', append=TRUE)


for (ipar in 1:nrow(para.space)) {
  endog <- para.space[ipar,1]
  hetero <- para.space[ipar,2]
  fattail <- para.space[ipar,3]
  if (endog==0){
    endogstr <- "No"
  }else if (endog<=0.3 & endog>0) {
    endogstr <- "Some"
  } else if (endog>0.3) {
    endogstr <- "Much"
  } 
  
  if (hetero==0){
    heterostr <- "No"
  }else if (hetero==0.1 | hetero==0.2) {
    heterostr <- "Little"
  } else if (hetero>=0.3 & hetero<=0.6) {
    heterostr <- "Some"
  } else if (hetero>0.6) {
    heterostr <- "Much"
  }
  
  if (fattail==1){
    fattailstr <- "Much"
  }else if (fattail<5 && fattail>1) {
    fattailstr <- "Some"
  } else if (fattail>=5) {
    fattailstr <- "Little"
  } else if(fattail==0){
    fattailstr <- "NA"
  }
  

  cat(sprintf("%-10s & %6s & %6s & %6s &   %s  & %s  & %s  &  %s & %s & %s   \\\\\n",
              ipar, endogstr, heterostr, fattailstr,
              ifelse(b.hats.rRMSE.par.rel[ipar,'relAVG.MMIV']<1, sprintf("$\\num[math-rm=\\mathbf]{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMIV']), sprintf("$\\num{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMIV'])),
              paste0(sprintf("$\\num{%4.6f}$",c(b.hats.rRMSE.par.rel[ipar,'relAGG.MMIV'],b.hats.rRMSE.par.rel[ipar,'relIV'])),collapse=' & '),
              ifelse(b.hats.rRMSE.par.rel[ipar,'relAVG.MMQR']<1, sprintf("$\\num[math-rm=\\mathbf]{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMQR']), sprintf("$\\num{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMQR'])),
              paste0(sprintf("$\\num{%4.6f}$",c(b.hats.rRMSE.par.rel[ipar,'relAGG.MMQR'],b.hats.rRMSE.par.rel[ipar,'relQR'])),collapse=' & '),
              ifelse(b.hats.rRMSE.par.rel[ipar,'relAVG.BS.RRMSE']<1, sprintf("$\\num[math-rm=\\mathbf]{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.BS.RRMSE']), sprintf("$\\num{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.BS.RRMSE'])),
              sprintf("$\\num{%4.6f}$",b.hats.rRMSE.par.rel[ipar,'relSEE'])
  ), file=OUTFILE, sep='', append=TRUE)

} # end of loop for IVQR-2SLS latex table


cat(sprintf("\\bottomrule
            \\end{tabular}
            %%
            \\begin{tablenotes}
            \\item $\\num{%d}$ replications. $\\num{%d}$ bootstraps. Sample size is %d.\\\\
            \\end{tablenotes}
            \\end{threeparttable}
            \\end{table}\n", NREPLIC, BREPLIC,n), file=OUTFILE, sep='', append=TRUE)

cat(sprintf("%% total time elapsed: %s\n\n\n",format(Sys.time()-start.time)),
    file=OUTFILE, sep='', append=TRUE)

# JTPA (ALL ESTIMATES) 14 DGP TABLE ####################  END  ################### 
###########################################################################################################################################################




############################################################################################################################
## JTPA (ALL ESTIMATES) 14 DGP TABLE ### ADD abs rRMSE of B.CON ####################  BEGIN  ################### 
cat("\\begin{table}[htbp]
    \\centering", file=OUTFILE, sep='', append=TRUE)

cat(sprintf("\\caption{\\label{tab:est:M1:tau=%s} Finite sample relative robust RMSE in JTPA-based simulation model M1 at %s.}
            ",
            tau, ifelse(tau==0.5, sprintf("%s quantile (median)",tau), sprintf("%s quantile",tau)) ), file=OUTFILE, sep='', append=TRUE)

cat("\\sisetup{round-precision=2,round-mode=places}
    \\begin{threeparttable}
    \\begin{tabular}{clllccclcclccc}
    \\toprule
    &          &          & & \\multicolumn{2}{c}{$\\mathrm{IVQR.2SLS}$} &               & \\multicolumn{2}{c}{$\\mathrm{IVQR.QR} $} &               &  &     \\\\
    \\cmidrule{5-6} \\cmidrule{8-9}
    $\\mathrm{DGP }$  &  $\\mathrm{Endog}$ &  $\\mathrm{Hetero}$ &  $\\mathrm{Fatail}$ &  $\\mathrm{AVG} $ & $\\mathrm{AGG}$       & $\\mathrm{IV}$ & $\\mathrm{AVG}$ & $\\mathrm{AGG}$         & $\\mathrm{QR}$ & $\\mathrm{BS} $   & $\\mathrm{CON}$ \\\\
    \\midrule\n", file=OUTFILE, sep='', append=TRUE)


for (ipar in 1:nrow(para.space)) {
  endog <- para.space[ipar,1]
  hetero <- para.space[ipar,2]
  fattail <- para.space[ipar,3]
  if (endog==0){
    endogstr <- "No"
  }else if (endog<=0.3 & endog>0) {
    endogstr <- "Some"
  } else if (endog>0.3) {
    endogstr <- "Much"
  } 
  
  if (hetero==0){
    heterostr <- "No"
  }else if (hetero==0.1 | hetero==0.2) {
    heterostr <- "Little"
  } else if (hetero>=0.3 & hetero<=0.6) {
    heterostr <- "Some"
  } else if (hetero>0.6) {
    heterostr <- "Much"
  }
  
  if (fattail==1){
    fattailstr <- "Much"
  }else if (fattail<5 && fattail>1) {
    fattailstr <- "Some"
  } else if (fattail>=5) {
    fattailstr <- "Little"
  } else if(fattail==0){
    fattailstr <- "NA"
  }
  

  cat(sprintf("%-10s & %6s & %6s & %6s &  %s  & %s  & %s  &  %s & %s & %s   \\\\\n",
              ipar, endogstr, heterostr, fattailstr,
              ifelse(b.hats.rRMSE.par.rel[ipar,'relAVG.MMIV']<1, sprintf("$\\num[math-rm=\\mathbf]{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMIV']), sprintf("$\\num{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMIV'])),
              paste0(sprintf("$\\num{%4.6f}$",c(b.hats.rRMSE.par.rel[ipar,'relAGG.MMIV'],b.hats.rRMSE.par.rel[ipar,'relIV'])),collapse=' & '),
              ifelse(b.hats.rRMSE.par.rel[ipar,'relAVG.MMQR']<1, sprintf("$\\num[math-rm=\\mathbf]{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMQR']), sprintf("$\\num{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMQR'])),
              paste0(sprintf("$\\num{%4.6f}$",c(b.hats.rRMSE.par.rel[ipar,'relAGG.MMQR'],b.hats.rRMSE.par.rel[ipar,'relQR'])),collapse=' & '),
              ifelse(b.hats.rRMSE.par.rel[ipar,'relAVG.BS.RRMSE']<1, sprintf("$\\num[math-rm=\\mathbf]{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.BS.RRMSE']), sprintf("$\\num{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.BS.RRMSE'])),
              sprintf("$\\num{%4.6f}$",b.hats.rRMSE.par.rel[ipar,'MM'])
  ), file=OUTFILE, sep='', append=TRUE)
  
} # end of loop for IVQR-2SLS latex table


cat(sprintf("\\bottomrule
            \\end{tabular}
            %%
            \\begin{tablenotes}
            \\item $\\num{%d}$ replications. $\\num{%d}$ bootstraps. Sample size is %d.\\\\
            \\end{tablenotes}
            \\end{threeparttable}
            \\end{table}\n", NREPLIC, BREPLIC,n), file=OUTFILE, sep='', append=TRUE)

cat(sprintf("%% total time elapsed: %s\n\n\n",format(Sys.time()-start.time)),
    file=OUTFILE, sep='', append=TRUE)

# JTPA (ALL ESTIMATES) 14 DGP TABLE ### ADD abs rRMSE of B.CON  ####################  END  ################### 
###########################################################################################################################################################



############################################################################################################################
## JTPA (ALL ESTIMATES) 14 DGP TABLE ### FOR IVQR-2SLS compare tau table B.CON ####################  BEGIN  ################### 
cat("\\begin{table}[htbp]
    \\centering", file=OUTFILE, sep='', append=TRUE)

cat(sprintf("\\caption{\\label{tab:est:M1:2SLS:compareTau:tau=%s} Finite sample relative robust RMSE in JTPA-based simulation model M1 at %s.}
            ",
            tau, ifelse(tau==0.5, sprintf("%s quantile (median)",tau), sprintf("%s quantile",tau)) ), file=OUTFILE, sep='', append=TRUE)

cat(sprintf("\\sisetup{round-precision=2,round-mode=places}
    \\begin{threeparttable}
    \\begin{tabular}{cccclcc}
    \\toprule
    &                     \\multicolumn{4}{c}{$\\tau=%s$}  \\\\
    \\cmidrule{2-5} 
    &           \\multicolumn{2}{c}{$\\mathrm{IVQR.2SLS}$} &       &            \\\\
    \\cmidrule{2-3} 
    $\\mathrm{DGP }$ &  $\\mathrm{AVG} $ & $\\mathrm{AGG}$       & $\\mathrm{IV}$    & $\\mathrm{CON}$ \\\\
    \\midrule\n",
            tau), file=OUTFILE, sep='', append=TRUE)


for (ipar in 1:nrow(para.space)) {
  endog <- para.space[ipar,1]
  hetero <- para.space[ipar,2]
  fattail <- para.space[ipar,3]


  cat(sprintf("%-10s  &  %s  & %s   & %s   \\\\\n",
              ipar, 
              ifelse(b.hats.rRMSE.par.rel[ipar,'relAVG.MMIV']<1, sprintf("$\\num[math-rm=\\mathbf]{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMIV']), sprintf("$\\num{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMIV'])),
              paste0(sprintf("$\\num{%4.6f}$",c(b.hats.rRMSE.par.rel[ipar,'relAGG.MMIV'],b.hats.rRMSE.par.rel[ipar,'relIV'])),collapse=' & '),
              sprintf("$\\num{%4.6f}$",b.hats.rRMSE.par.rel[ipar,'MM'])
  ), file=OUTFILE, sep='', append=TRUE)
  
} # end of loop for IVQR-2SLS latex table


cat(sprintf("\\bottomrule
            \\end{tabular}
            %%
            \\begin{tablenotes}
            \\item $\\num{%d}$ replications. $\\num{%d}$ bootstraps. Sample size is %d.\\\\
            \\end{tablenotes}
            \\end{threeparttable}
            \\end{table}\n", NREPLIC, BREPLIC,n), file=OUTFILE, sep='', append=TRUE)

cat(sprintf("%% total time elapsed: %s\n\n\n",format(Sys.time()-start.time)),
    file=OUTFILE, sep='', append=TRUE)

# JTPA (ALL ESTIMATES) 14 DGP TABLE ###  FOR IVQR-2SLS compare tau table B.CON  ####################  END  ################### 
###########################################################################################################################################################

############################################################################################################################
## JTPA (ALL ESTIMATES) 14 DGP TABLE ### FOR IVQR-QR compare tau table B.CON ####################  BEGIN  ################### 
cat("\\begin{table}[htbp]
    \\centering", file=OUTFILE, sep='', append=TRUE)

cat(sprintf("\\caption{\\label{tab:est:M1:QR:compareTau:tau=%s} Finite sample relative robust RMSE in JTPA-based simulation model M1 at %s.}
            ",
            tau, ifelse(tau==0.5, sprintf("%s quantile (median)",tau), sprintf("%s quantile",tau)) ), file=OUTFILE, sep='', append=TRUE)

cat(sprintf("\\sisetup{round-precision=2,round-mode=places}
            \\begin{threeparttable}
            \\begin{tabular}{cccclcc}
            \\toprule
            &                     \\multicolumn{4}{c}{$\\tau=%s$}  \\\\
            \\cmidrule{2-5} 
            &           \\multicolumn{2}{c}{$\\mathrm{IVQR.QR}$} &       &            \\\\
            \\cmidrule{2-3} 
            $\\mathrm{DGP }$ &  $\\mathrm{AVG} $ & $\\mathrm{AGG}$       & $\\mathrm{QR}$    & $\\mathrm{CON}$ \\\\
            \\midrule\n",
            tau), file=OUTFILE, sep='', append=TRUE)


for (ipar in 1:nrow(para.space)) {
  endog <- para.space[ipar,1]
  hetero <- para.space[ipar,2]
  fattail <- para.space[ipar,3]
  
  cat(sprintf("%-10s  &  %s  & %s   & %s   \\\\\n",
              ipar, 
              ifelse(b.hats.rRMSE.par.rel[ipar,'relAVG.MMQR']<1, sprintf("$\\num[math-rm=\\mathbf]{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMQR']), sprintf("$\\num{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMQR'])),
              paste0(sprintf("$\\num{%4.6f}$",c(b.hats.rRMSE.par.rel[ipar,'relAGG.MMQR'],b.hats.rRMSE.par.rel[ipar,'relQR'])),collapse=' & '),
              sprintf("$\\num{%4.6f}$",b.hats.rRMSE.par.rel[ipar,'MM'])
  ), file=OUTFILE, sep='', append=TRUE)
  
} # end of loop for IVQR-QR latex table


cat(sprintf("\\bottomrule
            \\end{tabular}
            %%
            \\begin{tablenotes}
            \\item $\\num{%d}$ replications. $\\num{%d}$ bootstraps. Sample size is %d.\\\\
            \\end{tablenotes}
            \\end{threeparttable}
            \\end{table}\n", NREPLIC, BREPLIC,n), file=OUTFILE, sep='', append=TRUE)

cat(sprintf("%% total time elapsed: %s\n\n\n",format(Sys.time()-start.time)),
    file=OUTFILE, sep='', append=TRUE)

# JTPA (ALL ESTIMATES) 14 DGP TABLE ###  FOR IVQR-QR compare tau table B.CON  ####################  END  ################### 
###########################################################################################################################################################

###########################################################################################################################################################
# JTPA (ALL ESTIMATES) 14 DGP TABLE ###  FOR BS compare tau table B.CON  ####################  BEGIN  ################### 
if(BS.FLAG){
  cat("\\begin{table}[htbp]
    \\centering", file=OUTFILE, sep='', append=TRUE)
  
  cat(sprintf("\\caption{\\label{tab:est:M1:BS:compareTau:tau=%s} Finite sample relative robust RMSE in JTPA-based simulation model M1 at %s.}
              ",
              tau, ifelse(tau==0.5, sprintf("%s quantile (median)",tau), sprintf("%s quantile",tau)) ), file=OUTFILE, sep='', append=TRUE)
  
  cat(sprintf("\\sisetup{round-precision=2,round-mode=places}
              \\begin{threeparttable}
              \\begin{tabular}{cccc}
              \\toprule
              &                     \\multicolumn{2}{c}{$\\tau=%s$}  \\\\
              \\cmidrule{2-3} 
              $\\mathrm{DGP }$     & $\\mathrm{BS}$    & $\\mathrm{CON}$ \\\\
              \\midrule\n",
            tau), file=OUTFILE, sep='', append=TRUE)


for (ipar in 1:nrow(para.space)) {
  endog <- para.space[ipar,1]
  hetero <- para.space[ipar,2]
  fattail <- para.space[ipar,3]
  

  cat(sprintf("%-10s  &  %s  & %s      \\\\\n",
              ipar, 
              ifelse(b.hats.rRMSE.par.rel[ipar,'relAVG.BS.RRMSE']<1, sprintf("$\\num[math-rm=\\mathbf]{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.BS.RRMSE']), sprintf("$\\num{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.BS.RRMSE'])),
              sprintf("$\\num{%4.6f}$",b.hats.rRMSE.par.rel[ipar,'MM'])
  ), file=OUTFILE, sep='', append=TRUE)
  
} # end of loop for IVQR-QR latex table


cat(sprintf("\\bottomrule
            \\end{tabular}
            %%
            \\begin{tablenotes}
            \\item $\\num{%d}$ replications. $\\num{%d}$ bootstraps. Sample size is %d.\\\\
            \\end{tablenotes}
            \\end{threeparttable}
            \\end{table}\n", NREPLIC, BREPLIC,n), file=OUTFILE, sep='', append=TRUE)

cat(sprintf("%% total time elapsed: %s\n\n\n",format(Sys.time()-start.time)),
    file=OUTFILE, sep='', append=TRUE)
}# end of if(BS.FLAG) 
# JTPA (ALL ESTIMATES) 14 DGP TABLE ###  FOR BS compare tau table B.CON  ####################  END  ################### 
###########################################################################################################################################################




#### LATEX table for endogPic or HeteroPic or FattailPic   ######## begin ############################################################################
###########################################################################################################################################################
if(ENDOG.PIC.FLAG | HETER.PIC.FLAG | FATAIL.PIC.FLAG){
  for(ibtex in 1:3){
    st<-(ibtex-1)*11+1
    ed<-ibtex*11
  ############################################################################################################################
  ## JTPA (ALL ESTIMATES) 14 DGP TABLE ### ADD abs rRMSE of B.CON ####################  BEGIN  ################### 
  cat("\\begin{table}[htbp]
      \\centering", file=OUTFILE, sep='', append=TRUE)
  
  if(ENDOG.PIC.FLAG){
    cat(sprintf("\\caption{\\label{tab:est:M1EndogPic_at%sHetero:tau=%s} Finite sample relative robust RMSE in JTPA-based simulation model M1 at %s.}",
                para.space[st,2], tau, ifelse(tau==0.5, sprintf("%s quantile (median)",tau), sprintf("%s quantile",tau)) ), file=OUTFILE, sep='', append=TRUE)
  }else if(HETER.PIC.FLAG){
    cat(sprintf("\\caption{\\label{tab:est:M1HeterPic_at%sEndog:tau=%s} Finite sample relative robust RMSE in JTPA-based simulation model M1 at %s.}",
                para.space[st,1], tau, ifelse(tau==0.5, sprintf("%s quantile (median)",tau), sprintf("%s quantile",tau)) ), file=OUTFILE, sep='', append=TRUE)
  }else if(FATAIL.PIC.FLAG){
    cat(sprintf("\\caption{\\label{tab:est:M1FatailPic_at%sEndog:tau=%s} Finite sample relative robust RMSE in JTPA-based simulation model M1 at %s.}",
                para.space[st,1], tau, ifelse(tau==0.5, sprintf("%s quantile (median)",tau), sprintf("%s quantile",tau)) ), file=OUTFILE, sep='', append=TRUE)
  }

  cat("\\sisetup{round-precision=2,round-mode=places}
      \\begin{threeparttable}
      \\begin{tabular}{clllccclcclccc}
      \\toprule
      &          &          & & \\multicolumn{2}{c}{$\\mathrm{IVQR.2SLS}$} &               & \\multicolumn{2}{c}{$\\mathrm{IVQR.QR} $} &               &  &     \\\\
      \\cmidrule{5-6} \\cmidrule{8-9}
      $\\mathrm{DGP }$  &  $\\mathrm{Endog}$ &  $\\mathrm{Hetero}$ &  $\\mathrm{Fatail}$ &  $\\mathrm{AVG} $ & $\\mathrm{AGG}$       & $\\mathrm{IV}$ & $\\mathrm{AVG}$ & $\\mathrm{AGG}$         & $\\mathrm{QR}$ & $\\mathrm{BS} $   & $\\mathrm{CON}$ \\\\
      \\midrule\n", file=OUTFILE, sep='', append=TRUE)
  

      
      for (ii in 1:11){
      ipar <- st+ii-1
    endog <- para.space[ipar,1]
    hetero <- para.space[ipar,2]
    fattail <- para.space[ipar,3]
    

    cat(sprintf("%-10s & %s & %s & %s &  %s  & %s  & %s  &  %s & %s & %s   \\\\\n",
                ii, endog, hetero, fattail,
                ifelse(b.hats.rRMSE.par.rel[ipar,'relAVG.MMIV']<1, sprintf("$\\num[math-rm=\\mathbf]{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMIV']), sprintf("$\\num{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMIV'])),
                paste0(sprintf("$\\num{%4.6f}$",c(b.hats.rRMSE.par.rel[ipar,'relAGG.MMIV'],b.hats.rRMSE.par.rel[ipar,'relIV'])),collapse=' & '),
                ifelse(b.hats.rRMSE.par.rel[ipar,'relAVG.MMQR']<1, sprintf("$\\num[math-rm=\\mathbf]{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMQR']), sprintf("$\\num{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMQR'])),
                paste0(sprintf("$\\num{%4.6f}$",c(b.hats.rRMSE.par.rel[ipar,'relAGG.MMQR'],b.hats.rRMSE.par.rel[ipar,'relQR'])),collapse=' & '),
                ifelse(b.hats.rRMSE.par.rel[ipar,'relAVG.BS.RRMSE']<1, sprintf("$\\num[math-rm=\\mathbf]{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.BS.RRMSE']), sprintf("$\\num{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.BS.RRMSE'])),
                sprintf("$\\num{%4.6f}$",b.hats.rRMSE.par.rel[ipar,'MM'])
    ), file=OUTFILE, sep='', append=TRUE)
      } # end of ii loop

  
  cat(sprintf("\\bottomrule
            \\end{tabular}
            %%
            \\begin{tablenotes}
            \\item $\\num{%d}$ replications. $\\num{%d}$ bootstraps. Sample size is %d.\\\\
            \\end{tablenotes}
            \\end{threeparttable}
            \\end{table}\n", NREPLIC, BREPLIC,n), file=OUTFILE, sep='', append=TRUE)
  
  cat(sprintf("%% total time elapsed: %s\n\n\n",format(Sys.time()-start.time)),
      file=OUTFILE, sep='', append=TRUE)
  
  # JTPA (ALL ESTIMATES) 14 DGP TABLE ### ADD abs rRMSE of B.CON  ####################  END  ################### 
  ###########################################################################################################################################################
  } # end of ibtex
  
} # end of if (endogflag, heterflag, fattailflag)
###########################################################################################################################################################
#### LATEX table for endogPic or HeteroPic or FattailPic   ######## end ############################################################################




for(ibtab in 1:3){
  st<-(ibtab-1)*11+1
  ed<-ibtab*11

  

###########################################################################################################################################################
# line for increasing endogeneity level, performance of rel.rRMSE(b.hats)
if (ENDOG.PIC.FLAG){
  
if (SAVE.FLAG) pdf(file=sprintf("JTPA_EndogPic_at%sHetero_AVGMMIV_DGPid%s_NREP%d_BREP%d_tau%s_%s.pdf", para.space[st, 2], nrow(para.space), NREPLIC, BREPLIC, tau, format(Sys.time(),"%Y_%m_%d")),
                   pointsize=12, width=7, height=7)
par(family="serif",mar=c(5.1, 6, 6, 3))
plot(para.space[st:ed, 1], b.hats.rRMSE.par.rel[st:ed,5],type="n",
     xlab="c1 value (endogeneity level)",ylab="(relative) Robust RMSE",
     xlim=range(0,1), 
     ylim=c(0,LEGROOM*max(unlist(b.hats.rRMSE.par.rel[st:ed,5] ))), mgp=c(3,1,0),
     cex.main=2, cex.lab=2, cex.axis=2)

LWD <- 4
lines(para.space[st:ed, 1], b.hats.rRMSE.par.rel[st:ed, 5], col=1, lty=2, lwd=LWD) #ip,tau,n
lines(para.space[st:ed, 1], array(1, ed-st+1), col=8, lty=1, lwd=LWD) #ip,tau,n

legend('bottomright', 
       c(as.expression(substitute(avg.IVQR.2SLS)), as.expression(substitute(IVQR)) ),
       lwd=5,col=c(1,8,8),lty=c(2,1,1),inset=0,bty='n',cex=1.8, x.intersp=0.5, y.intersp=1.2)
if (SAVE.FLAG) dev.off()



if (SAVE.FLAG) pdf(file=sprintf("JTPA_EndogPic_at%sHetero_AVGMMQR_DGPid%s_NREP%d_BREP%d_tau%s_%s.pdf", para.space[st, 2], nrow(para.space), NREPLIC, BREPLIC, tau,format(Sys.time(),"%Y_%m_%d")),
                   pointsize=12, width=7, height=7)
par(family="serif",mar=c(5.1, 6, 6, 3))

plot(para.space[st:ed, 1], b.hats.rRMSE.par.rel[st:ed,7],type="n",
     xlab="c1 value (endogeneity level)",ylab="(relative) Robust RMSE",
     xlim=range(0,1),
     ylim=c(0,LEGROOM*max(unlist(b.hats.rRMSE.par.rel[st:ed,7] ))), mgp=c(3,1,0),
     cex.main=2, cex.lab=2, cex.axis=2)
LWD <- 4
lines(para.space[st:ed, 1], b.hats.rRMSE.par.rel[st:ed, 7], col=1, lty=2, lwd=LWD) #ip,tau,n
lines(para.space[st:ed, 1], array(1, ed-st+1), col=8, lty=1, lwd=LWD) #ip,tau,n

# topright
legend('bottomright',
       c(as.expression(substitute(avg.IVQR.QR)), as.expression(substitute(IVQR)) ),
       lwd=5,col=c(1,8,8),lty=c(2,1,1),inset=0,bty='n',cex=1.8, x.intersp=0.5, y.intersp=1.2)


if (SAVE.FLAG) dev.off()

if (SAVE.FLAG) pdf(file=sprintf("JTPA_EndogPic_at%sHetero_3Estimator_DGPid%s_NREP%d_BREP%d_tau%s_%s.pdf", para.space[st, 2], nrow(para.space), NREPLIC, BREPLIC, tau,format(Sys.time(),"%Y_%m_%d")),
                   pointsize=12, width=7, height=7)
par(family="serif",mar=c(5.1, 6, 6, 3))

plot(para.space[, 1], b.hats.rRMSE.par.rel[,5],type="n",
     xlab="c1 value (endogeneity level)",ylab="(realtive) Robust RMSE",
     xlim=range(0,1),
     ylim=c(0,LEGROOM*max(unlist(b.hats.rRMSE.par.rel[st:ed,5] ),unlist(b.hats.rRMSE.par.rel[,7] ))), mgp=c(3,1,0),
     cex.main=2, cex.lab=2, cex.axis=2)
LWD <- 4
lines(para.space[st:ed, 1], b.hats.rRMSE.par.rel[st:ed, 5], col=1, lty=3, lwd=LWD) #ip,tau,n
lines(para.space[st:ed, 1], b.hats.rRMSE.par.rel[st:ed, 7], col="springgreen", lty=1, lwd=LWD) #ip,tau,n
if(BS.FLAG){
  lines(para.space[st:ed, 1], b.hats.rRMSE.par.rel[st:ed, 11], col="chocolate1", lty=2, lwd=LWD) #ip,tau,n
}

lines(para.space[st:ed, 1], array(1, ed-st+1), col=8, lty=1, lwd=LWD) #ip,tau,n

# topright
legend('bottomright',
       c(as.expression(substitute(avg.IVQR.2SLS)), as.expression(substitute(avg.IVQR.QR)), as.expression(substitute(BS)), as.expression(substitute(IVQR)) ),
       lwd=5,col=c(1,"springgreen","chocolate1",8),lty=c(3,1,2,1),inset=0,bty='n',cex=1.8, x.intersp=0.5, y.intersp=1.2)


if (SAVE.FLAG) dev.off()



} # end of if(ENDOG.PIC.FLAG)
###########################################################################################################################################################
}# END OF ibtab loop




###########################################################################################################################################################
# line for increasing heterogeneity level, performance of rel.rRMSE(b.hats)
for(ibtab in 1:3){
  st<-(ibtab-1)*11+1
  ed<-ibtab*11
  if (HETER.PIC.FLAG){
  
  if (SAVE.FLAG) pdf(file=sprintf("JTPA_HeteroPic_at%sEndog_AVGMMIV_DGPid%s_NREP%d_BREP%d_tau%s_%s.pdf", para.space[st, 1], nrow(para.space), NREPLIC, BREPLIC, tau,format(Sys.time(),"%Y_%m_%d")),
                     pointsize=12, width=7, height=7)
  par(family="serif",mar=c(5.1, 6, 6, 3))

  plot(para.space[st:ed, 2], b.hats.rRMSE.par.rel[st:ed,5],type="n",
       xlab="c2 value (heterogeneity level)",ylab="(relative) Robust RMSE",
       xlim=range(0,1), 
       ylim=c(0,LEGROOM*max(unlist(b.hats.rRMSE.par.rel[st:ed,5] ))), mgp=c(3,1,0),
       cex.main=2, cex.lab=2, cex.axis=2)
  
  LWD <- 4
  lines(para.space[st:ed, 2], b.hats.rRMSE.par.rel[st:ed, 5], col=1, lty=2, lwd=LWD) #ip,tau,n
  lines(para.space[st:ed, 2], array(1, ed-st+1), col=8, lty=1, lwd=LWD) #ip,tau,n
  
  legend('bottomright', 
         c(as.expression(substitute(avg.IVQR.2SLS)), as.expression(substitute(IVQR)) ),
         lwd=5,col=c(1,8,8),lty=c(2,1,1),inset=0,bty='n',cex=1.8, x.intersp=0.5, y.intersp=1.2)
  
  if (SAVE.FLAG) dev.off()
  
  
  
  if (SAVE.FLAG) pdf(file=sprintf("JTPA_HeteroPic_at%sEndog_AVGMMQR_DGPid%s_NREP%d_BREP%d_tau%s_%s.pdf", para.space[st, 1], nrow(para.space), NREPLIC, BREPLIC, tau,format(Sys.time(),"%Y_%m_%d")),
                     pointsize=12, width=7, height=7)
  par(family="serif",mar=c(5.1, 6, 6, 3))
  
  plot(para.space[st:ed, 2], b.hats.rRMSE.par.rel[st:ed,7],type="n",
       xlab="c2 value (heterogeneity level)",ylab="(relative) Robust RMSE",
       xlim=range(0,1), 
       ylim=c(0,LEGROOM*max(unlist(b.hats.rRMSE.par.rel[st:ed,7] ))), mgp=c(3,1,0),
       cex.main=2, cex.lab=2, cex.axis=2)
  LWD <- 4
  lines(para.space[st:ed, 2], b.hats.rRMSE.par.rel[st:ed, 7], col=1, lty=2, lwd=LWD) #ip,tau,n
  lines(para.space[st:ed, 2], array(1, ed-st+1), col=8, lty=1, lwd=LWD) #ip,tau,n
  
  # topright
  legend('bottomright', 
         c(as.expression(substitute(avg.IVQR.QR)), as.expression(substitute(IVQR)) ),
         lwd=5,col=c(1,8,8),lty=c(2,1,1),inset=0,bty='n',cex=1.8, x.intersp=0.5, y.intersp=1.2)
  
  
  if (SAVE.FLAG) dev.off()
  
  if (SAVE.FLAG) pdf(file=sprintf("JTPA_HeteroPic_at%sEndog_3Estimator_DGPid%s_NREP%d_BREP%d_tau%s_%s.pdf", para.space[st, 1], nrow(para.space), NREPLIC, BREPLIC, tau,format(Sys.time(),"%Y_%m_%d")),
                     pointsize=12, width=7, height=7)
  par(family="serif",mar=c(5.1, 6, 6, 3))
  
  plot(para.space[st:ed, 2], b.hats.rRMSE.par.rel[st:ed,5],type="n",
       xlab="c2 value (heterogeneity level)",ylab="(realtive) Robust RMSE",
       xlim=range(0,1), 
       ylim=c(0,LEGROOM*max(unlist(b.hats.rRMSE.par.rel[st:ed,5] ),unlist(b.hats.rRMSE.par.rel[st:ed,7] ))), mgp=c(3,1,0),
       cex.main=2, cex.lab=2, cex.axis=2)
  LWD <- 4
  lines(para.space[st:ed, 2], b.hats.rRMSE.par.rel[st:ed, 5], col=1, lty=3, lwd=LWD) #ip,tau,n
  lines(para.space[st:ed, 2], b.hats.rRMSE.par.rel[st:ed, 7], col="springgreen", lty=1, lwd=LWD) #ip,tau,n
  if(BS.FLAG){
    lines(para.space[st:ed, 2], b.hats.rRMSE.par.rel[st:ed, 11], col="chocolate1", lty=2, lwd=LWD) #ip,tau,n
  }
  
  lines(para.space[st:ed, 2], array(1, ed-st+1), col=8, lty=1, lwd=LWD) #ip,tau,n
  
  #topright
  legend('bottomright', 
         c(as.expression(substitute(avg.IVQR.2SLS)), as.expression(substitute(avg.IVQR.QR)), as.expression(substitute(BS)), as.expression(substitute(IVQR)) ),
         lwd=5,col=c(1,"springgreen","chocolate1",8),lty=c(3,1,2,1),inset=0,bty='n',cex=1.8, x.intersp=0.5, y.intersp=1.2)
  if (SAVE.FLAG) dev.off()
  
} # end of if
###########################################################################################################################################################
} # END OF ibtab loop



###########################################################################################################################################################
# line for Decreasing fat-tail level, performance of rel.rRMSE(b.hats)
for(ibtab in 1:3){
  st<-(ibtab-1)*11+1
  ed<-ibtab*11
  if (FATAIL.PIC.FLAG){
  
  if (SAVE.FLAG) pdf(file=sprintf("JTPA_FatailPic_at%sEndog_AVGMMIV_DGPid%s_NREP%d_BREP%d_tau%s_%s.pdf", para.space[st, 1], nrow(para.space), NREPLIC, BREPLIC, tau,format(Sys.time(),"%Y_%m_%d")),
                     pointsize=12, width=7, height=7)
  par(family="serif",mar=c(5.1, 6, 6, 3))
  
  plot(para.space[st:ed, 3], b.hats.rRMSE.par.rel[st:ed,5],type="n",
       xlab="c3 value (Fat tail level (decreasing))",ylab="(relative) Robust RMSE",
       xlim=range(min(unlist(para.space[st:ed, 3] )),max(unlist(para.space[st:ed, 3] ))), 
       ylim=c(0,LEGROOM*max(max(unlist(b.hats.rRMSE.par.rel[st:ed,5] ),1))), mgp=c(3,1,0),
       cex.main=2, cex.lab=2, cex.axis=2)
  
  LWD <- 4
  lines(para.space[st:ed, 3], b.hats.rRMSE.par.rel[st:ed, 5], col=1, lty=2, lwd=LWD) #ip,tau,n
  lines(para.space[st:ed, 3], array(1, ed-st+1), col=8, lty=1, lwd=LWD) #ip,tau,n
  
  legend('bottomright', 
         c(as.expression(substitute(avg.IVQR.2SLS)), as.expression(substitute(IVQR)) ),
         lwd=5,col=c(1,8,8),lty=c(2,1,1),inset=0,bty='n',cex=1.8, x.intersp=0.5, y.intersp=1.2)
 
  if (SAVE.FLAG) dev.off()
  
  
  
  if (SAVE.FLAG) pdf(file=sprintf("JTPA_FatailPic_at%sEndog_AVGMMQR_DGPid%s_NREP%d_BREP%d_tau%s_%s.pdf", para.space[st, 1], nrow(para.space), NREPLIC, BREPLIC, tau,format(Sys.time(),"%Y_%m_%d")),
                     pointsize=12, width=7, height=7)
  par(family="serif",mar=c(5.1, 6, 6, 3))
  
  plot(para.space[st:ed, 3], b.hats.rRMSE.par.rel[st:ed,7],type="n",
       xlab="c3 value (Fat tail level (decreasing))",ylab="(relative) Robust RMSE",
       xlim=range(min(unlist(para.space[st:ed, 3] )),max(unlist(para.space[st:ed, 3] ))), 
       ylim=c(0,LEGROOM*max(max(unlist(b.hats.rRMSE.par.rel[st:ed,7] ),1))), mgp=c(3,1,0),
       cex.main=2, cex.lab=2, cex.axis=2)
  LWD <- 4
  lines(para.space[st:ed, 3], b.hats.rRMSE.par.rel[st:ed, 7], col=1, lty=2, lwd=LWD) #ip,tau,n
  lines(para.space[st:ed, 3], array(1, ed-st+1), col=8, lty=1, lwd=LWD) #ip,tau,n
  
  # topright
  legend('bottomright', 
         c(as.expression(substitute(avg.IVQR.QR)), as.expression(substitute(IVQR)) ),
         lwd=5,col=c(1,8,8),lty=c(2,1,1),inset=0,bty='n',cex=1.8, x.intersp=0.5, y.intersp=1.2)

  
  if (SAVE.FLAG) dev.off()
  
  if (SAVE.FLAG) pdf(file=sprintf("JTPA_FatailPic_at%sEndog_3Estimator_DGPid%s_NREP%d_BREP%d_tau%s_%s.pdf", para.space[st, 1], nrow(para.space), NREPLIC, BREPLIC, tau,format(Sys.time(),"%Y_%m_%d")),
                     pointsize=12, width=7, height=7)
  par(family="serif",mar=c(5.1, 6, 6, 3))
  
  plot(para.space[st:ed, 3], b.hats.rRMSE.par.rel[st:ed,5],type="n",
       xlab="c3 value (Fat tail level (decreasing))",ylab="(realtive) Robust RMSE",
       xlim=range(min(unlist(para.space[st:ed, 3] )),max(unlist(para.space[st:ed, 3] ))), 
       ylim=c(0,LEGROOM*max(1,unlist(b.hats.rRMSE.par.rel[st:ed,5] ),unlist(b.hats.rRMSE.par.rel[st:ed,7] ))), mgp=c(3,1,0),
       cex.main=2, cex.lab=2, cex.axis=2)
  LWD <- 4
  lines(para.space[st:ed, 3], b.hats.rRMSE.par.rel[st:ed, 5], col=1, lty=3, lwd=LWD) #ip,tau,n
  lines(para.space[st:ed, 3], b.hats.rRMSE.par.rel[st:ed, 7], col="springgreen", lty=1, lwd=LWD) #ip,tau,n
  if(BS.FLAG){
    lines(para.space[st:ed, 3], b.hats.rRMSE.par.rel[st:ed, 11], col="chocolate1", lty=2, lwd=LWD) #ip,tau,n
  }
  
  lines(para.space[st:ed, 3], array(1, ed-st+1), col=8, lty=1, lwd=LWD) #ip,tau,n
  
  #topright
  legend('bottomright', 
         c(as.expression(substitute(avg.IVQR.2SLS)), as.expression(substitute(avg.IVQR.QR)), as.expression(substitute(BS)), as.expression(substitute(IVQR)) ),
         lwd=5,col=c(1,"springgreen","chocolate1",8),lty=c(3,1,2,1),inset=0,bty='n',cex=1.8, x.intersp=0.5, y.intersp=1.2)
  if (SAVE.FLAG) dev.off()
  
}
###########################################################################################################################################################
} # END OF ibtab loop





st1 <- Sys.time()
st1-st0

############################### if using parallel, then close parallel in the end
if (PARALLEL.FLAG){
if (num_cores>1) stopCluster(cl)
}
#############################################################################################


#EOF