# by Xin Liu
# xl6f6@mail.missouri.edu or xliu@mail.missouri.edu

# Simulated model 3: extension simulated model 2 to varying heterogeneity.

# finished all graphs and table for the 6 table and 60 DGPs.


rm(list=ls())
st0 <- Sys.time()


# load code and packages
source("gmmq.R")
source("ivqr_see.R")
source("ivqr_onestep.R")

source("ivqr_gmm_avg.R")
source("AVGBS_fn_multipar.R")
source("ivqr_iv_avg.R")


require('quantreg')
library("GenSA")
library("MASS") # for mvrnorm()

library("ks")
library("kdensity")
####
library(parallel)

require("parallel")
require("foreach")
require("doRNG")
require("doParallel")


# initial value
n <- 1000 #sample size
ns <- c(1000, 500, 250, 100, 50)
tau <- 0.5 #quantile index
NREPLIC <- 200
BREPLIC <- 50
S1S2 <- 2 # S1S2 <- 2 # S1S2 is to choose CLS'S sim model (S1) or (S2)
Hetero.FLAG <- TRUE
BS.FLAG <- TRUE


dB <- 7
PARALLEL.FLAG <- TRUE
BS.GMMQ.FLAG <-TRUE

PDF.FLAG <- TRUE
SAVE.FLAG <- TRUE


LEGROOM <- 1.25

# Set output file
OUTFILE <- paste0(sprintf("CLS_sims_VaryHetero_NREP%d_BREP%d_tau%.2f_n%d_S1S2%d", NREPLIC, BREPLIC, tau, n, S1S2),format(Sys.time(),"%Y_%m_%d"),".txt")




para.space <- matrix(c(
  0,     0,
  0,     0.1,
  0,     0.2,
  0,     0.3,
  0,     0.4,
  0,     0.5,
  0,     0.6,
  0,     0.7,
  0,     0.8,
  0,     0.9,
  0,     1,
  0.2,     0,
  0.2,     0.1,
  0.2,     0.2,
  0.2,     0.3,
  0.2,     0.4,
  0.2,     0.5,
  0.2,     0.6,
  0.2,     0.7,
  0.2,     0.8,
  0.2,     0.9,
  0.2,     1,
  0.4,     0,
  0.4,     0.1,
  0.4,     0.2,
  0.4,     0.3,
  0.4,     0.4,
  0.4,     0.5,
  0.4,     0.6,
  0.4,     0.7,
  0.4,     0.8,
  0.4,     0.9,
  0.4,     1,
  0,     0,
  0.05,  0,
  0.1,   0,
  0.15,  0,
  0.2,   0,
  0.25,  0,
  0.3,   0,
  0.35,  0,
  0.4,   0,
  0,     0.5,
  0.05,  0.5,
  0.1,   0.5,
  0.15,  0.5,
  0.2,   0.5,
  0.25,  0.5,
  0.3,   0.5,
  0.35,  0.5,
  0.4,   0.5,
  0,     1,
  0.05,  1,
  0.1,   1,
  0.15,  1,
  0.2,   1,
  0.25,  1,
  0.3,   1,
  0.35,  1,
  0.4,   1  ), ncol=2, byrow = TRUE)

############################################################ 13071 sets of weights ################
w1.new <- w2.new <- w3.new <- share1 <- share2 <- array(NA,dim=c(13071,1))

for (i in 1:10000){
  w1.new[i] <- (floor((i-1)/100))/100
  w2.new[i] <- (1-w1.new[i])*(i-100*(floor(i/100)))/100
  w3.new[i] <- (1-w1.new[i])*(1-(i-100*(floor(i/100)))/100)
}
#######
t <- 10000
for (i in 85:99){
  for (j in 1:99){
    t<- t+1
    w1.new[t] <- i/100+j/10000
    w2.new[t] <- 1-i/100-j/10000
    w3.new[t] <- 0
  }
}
# 15*99=1485 counts
# t= 10001 to 11485

for (i in 85:99){
  for (j in 1:99){
    t<- t+1
    w1.new[t] <- i/100+j/10000
    w2.new[t] <- 0
    w3.new[t] <- 1-i/100-j/10000
  }
}
# 1485 counts
# t= 11486 to 12970

for (i in 0:100){
  t<- t+1
  w1.new[t] <- i/100
  w2.new[t] <- 1-i/100
  w3.new[t] <- 0
}
# 101
# t=12971 to 13071

weights.new <- cbind(w1.new, w2.new, w3.new)
############################################################ 13071 sets of weights ################



if (PARALLEL.FLAG){
  
  # Calculate the number of cores
  num_cores <- detectCores() 
  
  # Initiate cluster
  cl <- makeCluster(num_cores)
  
  
  num_cores <- tryCatch({cl <- makeCluster(num_cores); registerDoParallel(cl); on.exit(stopCluster(cl),add=TRUE); num_cores}, error=function(Err){warning(sprintf("Error creating %d clusters",num_cores)); 1})
  set.seed(112358) #for replication
}else{
  set.seed(112358) #for replication
}

start.time <- Sys.time() #to see elapsed time



# generating data set:



b.hats.rRMSE.par.rel <- b.hats.MSE.par.rel <- data.frame(CON=rep(NA,nrow(para.space)), relSEE=NA, relIV=NA, relQR=NA, relAVG.MMIV=NA, relAGG.MMIV=NA, relAVG.MMQR=NA,  relAGG.MMQR=NA, 
                                                                                relAVG.BS.MSE=NA, 
                                                         relAVG.BS.RRMSE=NA, relGMMQ=NA)
rel.rRMSE <- rel.MSE <- cbind(data.frame(CON=rep(NA,nrow(para.space)), 
                                          SEE=NA,
                                          IV=NA, QR=NA, 
                                          AVG.MMIV=NA, AGG.MMIV=NA, AVG.MMQR=NA, AGG.MMQR=NA,
                                          AVG.BS.MSE=NA, 
                                          AVG.BS.RRMSE=NA, GMMQ=NA,
                                          endog=NA, heter=NA) )


# store the MSE details (ipar, irep) of the 11 estimates
b.CON.MSE <- b.SEE.MSE <- b.IV.MSE <- b.QR.MSE <- b.AVG.MMIV.MSE <- b.AGG.MMIV.MSE <- b.AVG.MMQR.MSE <- b.AGG.MMQR.MSE <- b.AVG.BS.MSE.MSE <- b.AVG.BS.RRMSE.MSE <- b.GMMQ.MSE <- array(NA,dim=c(nrow(para.space),NREPLIC))
# store the MSE (averaged over NREPLIC) of the 11 estimates
b.CON.MSE.AVG <- b.SEE.MSE.AVG <- b.IV.MSE.AVG <- b.QR.MSE.AVG <- b.AVG.MMIV.MSE.AVG <- b.AGG.MMIV.MSE.AVG <- b.AVG.MMQR.MSE.AVG <- b.AGG.MMQR.MSE.AVG <- b.AVG.BS.MSE.MSE.AVG <- b.AVG.BS.RRMSE.MSE.AVG <- b.GMMQ.MSE.AVG <- array(NA,dim=c(nrow(para.space),1))

# store the rRMSE (averaged over NREPLIC) of the 11 estimates
b.CON.rRMSE <- b.SEE.rRMSE <- b.IV.rRMSE <- b.QR.rRMSE <- b.AVG.MMIV.rRMSE <- b.AGG.MMIV.rRMSE <- b.AVG.MMQR.rRMSE <- b.AGG.MMQR.rRMSE <- b.AVG.BS.MSE.rRMSE <- b.AVG.BS.RRMSE.rRMSE <- b.GMMQ.rRMSE <- array(NA,dim=c(nrow(para.space),1))

# store the RMSE (averaged over NREPLIC) of the 11 estimates
b.CON.RMSE <- b.SEE.RMSE <- b.IV.RMSE <- b.QR.RMSE <- b.AVG.MMIV.RMSE <- b.AGG.MMIV.RMSE <- b.AVG.MMQR.RMSE <- b.AGG.MMQR.RMSE <- b.AVG.BS.MSE.RMSE <- b.AVG.BS.RRMSE.RMSE <- b.GMMQ.RMSE <- array(NA,dim=c(nrow(para.space),1))

cat(format(Sys.time())); cat('\n00000')

for (ipar in 1:nrow(para.space)){ # loop for c<-c(0:9/10)
c0 <- para.space[ipar, 1]
hetero <- para.space[ipar, 2]

if (!Hetero.FLAG){
  B.TRUE <-2.5
}else{
  B.TRUE <- hetero*1*tau^4
}

# multivariate normal covarianc matrix:
Sigma.upp <- matrix(cbind(diag(18), array(0,dim=c(18,7))), 18, 25)
Sigma.middle <- matrix(cbind(array(0,dim=c(6,18)), diag(6), array(c0, dim=c(6,1))), 6, 25)
Simga.bottom <- matrix(cbind(array(0,dim=c(1,18)), array(c0, dim=c(1,6)), 1), 1, 25)

Sigma.all <- rbind(Sigma.upp, Sigma.middle, Simga.bottom) # Sigma.all is the 25*25 covariance matrix for vector (Z1-Z18, eps1-eps6, ustar)




# Output the parameter values
cat(sprintf(paste0("iPAR=%d, NREPLIC=%d, n=%d, tau=%g"), ipar,  NREPLIC, n, tau), 
    file=OUTFILE, sep="\n",append=TRUE)

# Allocate storage variables
b.hats <- data.frame(MM=rep(NA,NREPLIC), 
                     SEE=NA,
                     IV=NA, QR=NA, 
                     AVG.MMIV=NA, AGG.MMIV=NA, AVG.MMQR=NA, AGG.MMQR=NA, 
                     AVG.BS.MSE=NA,
                     AVG.BS.RRMSE=NA, GMMQ=NA)
# store the 11 estimates (per replication)
b.CON <- b.SEE <- b.IV <- b.QR <- b.AVG.MMIV <- b.AGG.MMIV <- b.AVG.MMQR <- b.AGG.MMQR <- b.AVG.BS.MSE <- b.AVG.BS.RRMSE <- b.GMMQ <- array(NA,dim=c(dB,NREPLIC))

b.hats.avg <- array(NA,dim=c(13071,NREPLIC))
b.hats.avg.weights <- cbind(b.hats.avg, weights.new)
w.avg.MMIV <- w.avg.MMQR <- data.frame(weight=rep(NA,NREPLIC))

 cat(sprintf("ipar=%03d\n",ipar)) #progress update





  # for (irep in 1:NREPLIC) {  #non-parallel version, uncomment, if trying to parallel replications.
  bp.raw <- foreach (irep=1:NREPLIC, .combine=cbind, .inorder=TRUE) %dorng% {   #non-parallel, comment 

    # comment in nonparallel; uncommented in parallel (begin)
    require('quantreg')
    library("MASS") # for mvrnorm()
    library("ks")
    library("kdensity")
    library("GenSA")
    # comment in nonparallel; uncommented in parallel (end)
   
     # sink("log.txt", append=TRUE)
    
    if(BS.FLAG){
    cat(sprintf("irep=%04d\n",irep))
  }else{
    cat(sprintf("irep=%04d",irep))
  }
    

  # generate data
  dat <- mvrnorm(n = n, mu = rep(0, 25), Sigma = Sigma.all)
  Z.con <- dat[, 1:12] # Z.con is exog.
  ustar <- dat[, 25]
  
  if (S1S2!=1){
    u <- ustar
    u <- u-qnorm(p=tau)
  }else{
    u <- qchisq(pnorm(ustar), df=4) - qchisq(tau, df=4)
    
  }

  # using Z, generate X. (Epsilon(contained in X) and u are correlated. so X and u are correlated. X is endog)
  X1 <- (dat[, 1]+dat[, 7])/2+dat[, 13]+dat[, 19]
  X2 <- (dat[, 2]+dat[, 8])/2+dat[, 14]+dat[, 20]
  X3 <- (dat[, 3]+dat[, 9])/2+dat[, 15]+dat[, 21]
  X4 <- (dat[, 4]+dat[, 10])/2+dat[, 16]+dat[, 22]
  X5 <- (dat[, 5]+dat[, 11])/2+dat[, 17]+dat[, 23]
  X6 <- (dat[, 6]+dat[, 12])/2+dat[, 18]+dat[, 24]
  
  Z1 <- dat[, 1]; Z2 <- dat[, 2]; Z3 <- dat[, 3]; Z4 <- dat[, 4]; Z5 <- dat[, 5]; Z6 <- dat[, 6]
  Z7 <- dat[, 7]; Z8 <- dat[, 8]; Z9 <- dat[, 9]; Z10 <- dat[, 10]; Z11 <- dat[, 11]; Z12 <- dat[, 12]
  EPS1 <- dat[, 19]; EPS2 <- dat[, 20]; EPS3 <- dat[, 21]; EPS4 <- dat[, 22]; EPS5 <- dat[, 23]; EPS6 <- dat[, 24]

    if (!Hetero.FLAG){
    theta <-2.5
  }else{
    X1 <- (Z1+Z7)+EPS1
    X2 <- (Z2+Z8)+EPS2
    X3 <- (Z3+Z9)+EPS3
    X4 <- (Z4+Z10)+EPS4
    X5 <- (Z5+Z11)+EPS5
    X6 <- (Z6+Z12)+EPS6
    
    X1 <- ((X1+3.1*sqrt(3))>=0)*(X1+3.1*sqrt(3))
    X2 <- ((X2+3.1*sqrt(3))>=0)*(X2+3.1*sqrt(3))
    X3 <- ((X3+3.1*sqrt(3))>=0)*(X3+3.1*sqrt(3))
    X4 <- ((X4+3.1*sqrt(3))>=0)*(X4+3.1*sqrt(3))
    X5 <- ((X5+3.1*sqrt(3))>=0)*(X5+3.1*sqrt(3))
    X6 <- ((X6+3.1*sqrt(3))>=0)*(X6+3.1*sqrt(3))

    # NEW
    theta <- hetero*1*pnorm(u)^4
    if (S1S2==1){
      theta <- hetero*1*pchisq(u,df=4)^4
    }

  }
  
  X.excl <- array(1,dim = c(n,1))
  
  # using X, generate Y
  Y <- (X1+X2+X3+X4+X5+X6)*theta+u+X.excl*1


  D <- cbind(X1, X2, X3, X4, X5, X6)
  Z.excl <- cbind(Z1, Z2, Z3, Z4, Z5, Z6, Z7, Z8, Z9, Z10, Z11, Z12)
  demean.Z.excl<- Z.excl-colMeans(Z.excl)

  Z <- cbind(X.excl, Z.excl)
  X <- cbind(X.excl, D)
  
  dZ <- ncol(Z) #number of instruments
  dX <- ncol(X) #number of regressors
  dD <- ncol(D)
  dB <- ncol(X)
  Y.gmmq <- cbind(Y,D)

  
  
  # Compute estimators and store
  # # Kaplan and Sun (2017), IVQR-SEE with "optimal" bandwidth (no h_in specified)
  tmp.SEE <- ivqr_see(p = tau, Y = Y, X = X, Z = Z)
  H.SEE <- H <- tmp.SEE$h # for SEE,  b.SEE has the same sequence with X and Z. i.e. if X and Z are (intercept, slope), then b.SEE is (intercept,slope);if X and Z are (slope, intercept), then b.SEE is (slope, intercept)
  # here b/c X and Z are (intercept, slope), then b.SEE is (intercept, slope). so we're interested in b.SEE[2], the slope.
  H.LARGE <- H.SEE*n^(-2/35)
  b.SEE[,irep] <- tmp.SEE$b


  # 2SLS
  PZ <- Z %*% solve(t(Z)%*%Z) %*% t(Z)
  b.2sls <- solve(t(X)%*%PZ%*%X) %*% t(X)%*%PZ%*%Y
  b.IV[,irep] <-b.2sls
  # Z is on (intercept, slope), X is on (intercept, slope), so b.2SLS is on (intercept, slope); we're interested in slope, thus look at b.2sls[2]
  

  # QR  
  suppressWarnings(b.qr <- rq(formula=Y~D,tau=tau)$coefficients)
  suppressWarnings(b.QR[,irep] <- b.qr)
  
  # SETUP for CON.GMM and AGG.GMM
  Lambda <- function(y, x, b){y-x%*%b } # require X and b has the same sequence. If X is (intercept, slope), b must be (intercept slope). Notice that X=(X.excl, D), but most b(b.gmmq, b.ivqr.gmm, b.onestep) are (slope, intercept), so need to adjust X or use Lambda.gmmq w/ Y.gmmq, X.excl
  Lambda.derivative <- function(y,x,b) -x # lambda.derivative fn gives the same sequence of X. If X is (intercept, slope) (as defined here, X=(X.excl, D)), the lambda.derivative gives (intercept, slope); if wanna labdha.derives generates (slope, intercept), then should use Lambda.derivative.gmmq or use Lambda.derivative (X=cbind(D,X.excl))
  Lambda.gmmq <- function(y,x,b) {y[,1]-cbind(y[,-1],x)%*%b} # b must be (slope, intercept) sequence. need to check this, when calling lambda.gmmq fn.
  Lambda.derivative.gmmq <- function(y,x,b) {-cbind(y[,-1],x)} # lambda.derivative.gmmq gives (slope, intercept) sequence.

  Itilde <- Itilde.KS17
  Itilde.deriv <- Itilde.deriv.KS17
  
  Gfn <- function(v,h) {      Itilde(v/h)       }
  Gpfn <- function(v,h) {     Itilde.deriv(v/h) }
  

  #################### compute  b.initial, for weighting mtx in GMM.CON ################      begin ####################  
  gmmq.est <- gmmq(tau=tau, Y=Y.gmmq, X=X.excl, Z.excl=Z.excl, dB=dB, 
                   Lambda=Lambda.gmmq, 
                   Lambda.derivative=Lambda.derivative.gmmq, 
                   h=0, VERBOSE=FALSE, RETURN.Z=FALSE )  # b.gmmq=(slope, intercept) (interested in slope)

  h.gmmq <- gmmq.est$h # do the gmmq is to get h.gmmq; h.gmmq is used for cov.mtx; cov.mtx is used for weighting matrix in GMM method to get "aggressive"; i.e. gmmq()-->h.gmmq-->cov.mtx-->weight.mtx in GMM of "aggressive".
  h.small <- h.gmmq
  b.gmmq <- gmmq.est$b # b.gmmq=(slope, intercept) (interested in slope)
  b.GMMQ[1,irep] <-b.gmmq[7]
  b.GMMQ[2:7,irep] <-b.gmmq[1:6]

  b.gmmq.intslp<-as.matrix(c(b.gmmq[7], b.gmmq[1:6]),ncol=1)
 
  # to compute 2step-GMM

  L.con <- matrix(Lambda.gmmq(y=Y.gmmq,x=X.excl,b=b.gmmq), ncol=1) # has intercept 
  gni.con <- cbind(Z.excl, X.excl)*repmat((Gfn(-L.con,h.small)-tau),1,dZ) # here gni is (slope, intercept), (Z.excl, X.excl) # has intercetp
  
  g.bar.con <- colMeans(gni.con)                                                   # g.bar.agg.MMQR sequence is (D, Z.excl, X.excl)
  cov.mtx.con <- t(gni.con)%*%gni.con/n-g.bar.con%*%t(g.bar.con) # cov.mat, weight.mtx sequence is (D, Z.excl, 1)
  inv.cov.mtx.con <- solve(cov.mtx.con)
  
  #################### compute  b.initial, for weighting mtx in GMM.CON ################      END ####################
  
  #################### compute  CON.GMM ##########################################      begin ####################
  # use inverse of cov.mtx as the weighting matrix to construct the GMM criterion. run GenSA with GMM. get b.1step

  tmp.con.1step <- ivqr.gmm(Y = Y, X.excl = X.excl, Z.excl =Z.excl, D = D, tau = tau, h = h.small, dB = dB, max.time=5, upper=rep(100,dB), lower=rep(-100,dB),
                                    weight.mtx=inv.cov.mtx.con, structure = 'iid', LRV.kernel = 'QS', b.init = b.gmmq)
  # ivqr.gmm gives b=(slope, intercept) = on (generalZ.excl, X.excl)=i.e. on (D, Z.excl, X.excl) = on (D, 1) =around (3,62)
  b.con.1step <- tmp.con.1step$b
  #################### compute  CON.GMM ##########################################      end ####################
  

  b.con.1step.intslp<-as.matrix(c(b.con.1step[7], b.con.1step[1:6]),ncol=1)
  
  # store b.conservative
  b.conservative <- b.con.1step 
  b.conservative.intslp <- b.con.1step.intslp
  b.CON[1,irep] <-b.con.1step[7]
  b.CON[2:7,irep] <-b.con.1step[1:6]
  
  #################### compute  b.GMM.CON     #######################     END ####################
  
  # b.conservative <- b.gmmq
  Uhat <- Y- cbind(X.excl,D)%*%b.conservative.intslp
  Uhat.sd <- apply(X = Uhat, MARGIN = 2, FUN = sd)
  
  alpha.tau.old.N01wrongform <- dnorm(qnorm(tau))*(1-qnorm(tau))^2
  alpha.tau.N01 <- dnorm(qnorm(tau))*(1-qnorm(tau)^2)^2
  alpha.tau.general <- dnorm(qnorm(tau))*(1-qnorm(tau)^2)^2/(Uhat.sd^5)
  
  num.Kato <- 4.5*sum(1*Z1^2 + 1*Z2^2 + 1*Z3^2 + 1*Z4^2 + 1*Z5^2 + 1*Z6^2 + 1*Z7^2 + 1*Z8^2 + 1*Z9^2 + 1*Z10^2 + 1*Z11^2 + 1*Z12^2
                      + X1^2*Z1^2 + X1^2*Z2^2 + X1^2*Z3^2 + X1^2*Z4^2 + X1^2*Z5^2 + X1^2*Z6^2 + X1^2*Z7^2 + X1^2*Z8^2 + X1^2*Z9^2 + X1^2*Z10^2 + X1^2*Z11^2 + X1^2*Z12^2
                      + X2^2*Z1^2 + X2^2*Z2^2 + X2^2*Z3^2 + X2^2*Z4^2 + X2^2*Z5^2 + X2^2*Z6^2 + X2^2*Z7^2 + X2^2*Z8^2 + X2^2*Z9^2 + X2^2*Z10^2 + X2^2*Z11^2 + X2^2*Z12^2
                      + X3^2*Z1^2 + X3^2*Z2^2 + X3^2*Z3^2 + X3^2*Z4^2 + X3^2*Z5^2 + X3^2*Z6^2 + X3^2*Z7^2 + X3^2*Z8^2 + X3^2*Z9^2 + X3^2*Z10^2 + X3^2*Z11^2 + X3^2*Z12^2
                      + X4^2*Z1^2 + X4^2*Z2^2 + X4^2*Z3^2 + X4^2*Z4^2 + X4^2*Z5^2 + X4^2*Z6^2 + X4^2*Z7^2 + X4^2*Z8^2 + X4^2*Z9^2 + X4^2*Z10^2 + X4^2*Z11^2 + X4^2*Z12^2
                      + X5^2*Z1^2 + X5^2*Z2^2 + X5^2*Z3^2 + X5^2*Z4^2 + X5^2*Z5^2 + X5^2*Z6^2 + X5^2*Z7^2 + X5^2*Z8^2 + X5^2*Z9^2 + X5^2*Z10^2 + X5^2*Z11^2 + X5^2*Z12^2
                      + X6^2*Z1^2 + X6^2*Z2^2 + X6^2*Z3^2 + X6^2*Z4^2 + X6^2*Z5^2 + X6^2*Z6^2 + X6^2*Z7^2 + X6^2*Z8^2 + X6^2*Z9^2 + X6^2*Z10^2 + X6^2*Z11^2 + X6^2*Z12^2)/n
  
  den.Kato <- ((sum(X*Z1)/n)^2 + (sum(1*Z2)/n)^2 + (sum(1*Z3)/n)^2 + (sum(1*Z4)/n)^2 + (sum(1*Z5)/n)^2 + (sum(1*Z6)/n)^2 + (sum(1*Z7)/n)^2 + (sum(1*Z8)/n)^2 + (sum(1*Z9)/n)^2 + (sum(1*Z10)/n)^2 + (sum(1*Z11)/n)^2 + (sum(1*Z12)/n)^2
               + (sum(X1*Z1)/n)^2 + (sum(X1*Z2)/n)^2 + (sum(X1*Z3)/n)^2 + (sum(X1*Z4)/n)^2 + (sum(X1*Z5)/n)^2 + (sum(X1*Z6)/n)^2 + (sum(X1*Z7)/n)^2 + (sum(X1*Z8)/n)^2 + (sum(X1*Z9)/n)^2 + (sum(X1*Z10)/n)^2 + (sum(X1*Z11)/n)^2 + (sum(X1*Z12)/n)^2
               + (sum(X2*Z1)/n)^2 + (sum(X2*Z2)/n)^2 + (sum(X2*Z3)/n)^2 + (sum(X2*Z4)/n)^2 + (sum(X2*Z5)/n)^2 + (sum(X2*Z6)/n)^2 + (sum(X2*Z7)/n)^2 + (sum(X2*Z8)/n)^2 + (sum(X2*Z9)/n)^2 + (sum(X2*Z10)/n)^2 + (sum(X2*Z11)/n)^2 + (sum(X2*Z12)/n)^2
               + (sum(X3*Z1)/n)^2 + (sum(X3*Z2)/n)^2 + (sum(X3*Z3)/n)^2 + (sum(X3*Z4)/n)^2 + (sum(X3*Z5)/n)^2 + (sum(X3*Z6)/n)^2 + (sum(X3*Z7)/n)^2 + (sum(X3*Z8)/n)^2 + (sum(X3*Z9)/n)^2 + (sum(X3*Z10)/n)^2 + (sum(X3*Z11)/n)^2 + (sum(X3*Z12)/n)^2
               + (sum(X4*Z1)/n)^2 + (sum(X4*Z2)/n)^2 + (sum(X4*Z3)/n)^2 + (sum(X4*Z4)/n)^2 + (sum(X4*Z5)/n)^2 + (sum(X4*Z6)/n)^2 + (sum(X4*Z7)/n)^2 + (sum(X4*Z8)/n)^2 + (sum(X4*Z9)/n)^2 + (sum(X4*Z10)/n)^2 + (sum(X4*Z11)/n)^2 + (sum(X4*Z12)/n)^2
               + (sum(X5*Z1)/n)^2 + (sum(X5*Z2)/n)^2 + (sum(X5*Z3)/n)^2 + (sum(X5*Z4)/n)^2 + (sum(X5*Z5)/n)^2 + (sum(X5*Z6)/n)^2 + (sum(X5*Z7)/n)^2 + (sum(X5*Z8)/n)^2 + (sum(X5*Z9)/n)^2 + (sum(X5*Z10)/n)^2 + (sum(X5*Z11)/n)^2 + (sum(X5*Z12)/n)^2
               + (sum(X6*Z1)/n)^2 + (sum(X6*Z2)/n)^2 + (sum(X6*Z3)/n)^2 + (sum(X6*Z4)/n)^2 + (sum(X6*Z5)/n)^2 + (sum(X6*Z6)/n)^2 + (sum(X6*Z7)/n)^2 + (sum(X6*Z8)/n)^2 + (sum(X6*Z9)/n)^2 + (sum(X6*Z10)/n)^2 + (sum(X6*Z11)/n)^2 + (sum(X6*Z12)/n)^2 )
  
  h.Kato.old.N01wrongform <- (num.Kato/(alpha.tau.old.N01wrongform*den.Kato)/n)^{1/5}
  h.Kato.N01 <- (num.Kato/(alpha.tau.N01*den.Kato)/n)^{1/5}
  h.Kato <- (num.Kato/(alpha.tau.general*den.Kato)/n)^{1/5}
  
  

  #################### compute  sig.conservative ####################      begin ####################  

  L <- matrix(Lambda.gmmq(y=Y.gmmq,x=X.excl,b=b.conservative), ncol=1) # HAS INTERCEPT
  Lp <- Lambda.derivative.gmmq(y=Y.gmmq, x=X.excl, b=b.conservative) # Lp is in (slope, intercept) sequence, consistent with jac. # should check all Lp uses Lambda.derivative.gmmq, instead of Lambda.derivative. # HAS INTERCEPT
  gni <- cbind(Z.excl, X.excl)*repmat((Gfn(-L,h.small)-tau),1,dZ) # here gni is (slope, intercept), (Z.excl, X.excl) # HAS INTERCEPT
  
  g.bar <- colMeans(gni) # g.bar is in (slope, intercept) sequence.
  
  h.use.con.sig <- h.Kato
  jac <- t(cbind(Z.excl, X.excl))%*%(Lp*repmat(Gpfn(-L,h.use.con.sig),1,dB))/(-h.use.con.sig)/n # add divide n # jac is (slope, intercept), (Z.excl, X.excl) # HAS INTERCEPT
  

  
  Omega.bar <- t(gni)%*%gni/n-g.bar%*%t(g.bar) # Omega.bar is in (slope, intercept) sequence.

  sig.conservative <- solve(t(jac)%*%solve(Omega.bar)%*%jac)   # here sig.conservative is (slope, intercept) matrix. (Z.excl, X.excl)
  #################### compute  sig.conservative ####################      end ####################  
  

  #################################################################################################### BEGIN OF MMQR
  # Averaging TYPE #1, (IVQR, QR)
  # STAGE 2 = (IVQR,QR)    "aggressive"
  # estimate cov.mtx using b.conservative and full GenSA moments.

  ################### compute  b.AGG.MMQR     #######################     begin ####################
  # Preparation for the weigting matrix in  GMM ################     begin ####################

  L.agg.MMQR <- matrix(Lambda.gmmq(y=Y.gmmq, x=X.excl, b=b.conservative), ncol=1)  # HAS INTERCEPT
  gni.agg.MMQR <- cbind(D, Z.excl, X.excl)*repmat((Gfn(-L.agg.MMQR, h.small)-tau),1,dZ+ncol(D))  # gni.agg.MMQR sequence is (D, Z.excl, X.excl) # HAS INTERCEPT
  g.bar.agg.MMQR <- colMeans(gni.agg.MMQR)                                                   # g.bar.agg.MMQR sequence is (D, Z.excl, X.excl)
  cov.mtx.agg.MMQR <- t(gni.agg.MMQR)%*%gni.agg.MMQR/n-g.bar.agg.MMQR%*%t(g.bar.agg.MMQR) # cov.mat, weight.mtx sequence is (D, Z.excl, 1)
  inv.cov.mtx.agg.MMQR <- solve(cov.mtx.agg.MMQR)
  # Preparation for the weigting matrix in  GMM ################     END ####################
  
  
  ######### Compute  GMM:  AVG.GMM.MMQR #######################     begin ####################

  tmp.MMQR.1step <- ivqr.gmm(Y = Y, X.excl = X.excl,  Z.excl =cbind(D, Z.excl), D = D, tau = tau, h = h.small, dB = dB, max.time=5, upper=rep(100,dB), lower=rep(-100,dB),
                                     weight.mtx=inv.cov.mtx.agg.MMQR, structure = 'iid', LRV.kernel = 'QS', b.init=b.conservative ) # HAS INTERCEPT
  b.MMQR.1step <- tmp.MMQR.1step$b
  ######### Compute  GMM:  AVG.GMM.MMQR #######################     END ####################
  


  
  # store b.aggressive
  b.agg.MMQR <- b.MMQR.1step 
  b.MMQR.1step.intslp<-as.matrix(c(b.MMQR.1step[7], b.MMQR.1step[1:6]),ncol=1)

  b.AGG.MMQR[1,irep] <-b.MMQR.1step[7]
  b.AGG.MMQR[2:7,irep] <-b.MMQR.1step[1:6]
  
  #################### compute  b.AGG.MMQR     #######################     END ####################  
  
  #################### compute Sigma.aggressive.MMQR #################   begin ####################  
  # use b.conservative to calculate sigma.aggressive.

  L.agg.MMQR.sig <- matrix(Lambda.gmmq(y=Y.gmmq, x=X.excl, b=b.conservative), ncol=1) # HAS INTERCEPT
  Lp.agg.MMQR.sig <- Lambda.derivative.gmmq(y=Y.gmmq, x=X.excl, b=b.conservative)  # note. b.2step is (slope, intercept), thus need to use Lambda.derivative.gmmq instead of Lambda.derivative, bc Lambda.derivative.gmmq give X in (D, X.excl) order; Lambda.derivative gives X in (X.excl, D) order. # HAS INTERCEPT
  gni.agg.MMQR.sig <- cbind(D, Z.excl, X.excl)*repmat((Gfn(-L.agg.MMQR.sig, h.small)-tau),1,dZ+ncol(D)) #gni2.sig is in (D, Z.excl, X.excl) order # HAS INTERCEPT
  g.bar.agg.MMQR.sig <- colMeans(gni.agg.MMQR.sig)
  
  
  
  h.use.MMQR.sig <- h.Kato
  
  jac.agg.MMQR.sig <- rbind(t(D)%*%(Lp.agg.MMQR.sig*repmat(Gpfn(-L.agg.MMQR.sig, h.use.MMQR.sig),1,dB))/(-h.use.MMQR.sig)/n, t(cbind(Z.excl, X.excl))%*%(Lp.agg.MMQR.sig*repmat(Gpfn(-L.agg.MMQR.sig, h.use.MMQR.sig),1,dB))/(-h.use.MMQR.sig)/n) # HAS INTERCEPT
  
  Omega.bar.agg.MMQR.sig <- t(gni.agg.MMQR.sig)%*%gni.agg.MMQR.sig/n-g.bar.agg.MMQR.sig%*%t(g.bar.agg.MMQR.sig) # iid

  sig.agg.MMQR <- tryCatch(solve(t(jac.agg.MMQR.sig)%*%solve(Omega.bar.agg.MMQR.sig)%*%jac.agg.MMQR.sig),
                           error=function(w)NA)
  if (is.na(sig.agg.MMQR[1])) next
  #################### compute Sigma.aggressive.MMQR #################   end  ####################  


  # averaging
  #################### compute averaging weights MMQR.weight #########   begin ####################  
  tt<-c(0,1,1,1,1,1,1)
  Ups <- diag(tt)
  TR1 <- Trace(Ups%*%(sig.conservative-sig.agg.MMQR))
  if (TR1 <0) {
    w.avg.MMQR[irep,] <- MMQRweight <- 0
  }else{
    w.avg.MMQR[irep,] <- MMQRweight <- max(0, min(1, TR1/(n*t(b.agg.MMQR-b.conservative)%*%Ups%*%(b.agg.MMQR-b.conservative)+TR1)))
  }
  
  # both b.agg.MMQR and b.conservative are slope, intercept
  # thus, b.avgMMQR is slope, intercept
  
  b.avgMMQR <- repmat((1-MMQRweight),dB,1)*b.conservative+repmat(MMQRweight,dB,1)*b.agg.MMQR

  b.avgMMQR.intslp<-as.matrix(c(b.avgMMQR[7], b.avgMMQR[1:6]),ncol=1)
  
  b.AVG.MMQR[1,irep] <-b.avgMMQR[7]
  b.AVG.MMQR[2:7,irep] <-b.avgMMQR[1:6]
  
  #################### compute averaging weights MMQR.weight #########   end ####################  
  
  ######################################################################################################## end of MMQR

  
  #################################################################################################### begin of MMIV
  # Averaging TYPE #2 (IVQR, IV)    "conservative"
  # Conservative part is the same as before.
  
  ######################################################################################
  # 22: compute "aggressive" using GenSA. get b2 and sig2.      "aggressive"
  
  # preparation weighting matrix for  GMM: AGG.GMM.MMIV
  # estimate cov.mtx using b.conservative and full GenSA moments.
  L.agg.MMIV <- matrix(Lambda.gmmq(y=Y.gmmq, x=X.excl, b=b.conservative), ncol=1)  # HAS INTERCEPT
  gni.agg.MMIV <- cbind(demean.Z.excl*repmat(L.agg.MMIV,1,ncol(Z.excl)), cbind(Z.excl, X.excl)*repmat((Gfn(-L.agg.MMIV, h.small)-tau),1,dZ)) # gni sequence is (demean.Z.excl, Z.excl, X.excl) # HAS INTERCEPT

  g.bar.agg.MMIV <- colMeans(gni.agg.MMIV) 
  cov.mtx.agg.MMIV <- t(gni.agg.MMIV)%*%gni.agg.MMIV/n-g.bar.agg.MMIV%*%t(g.bar.agg.MMIV) # cov.mat, weight.mtx sequence is (demean.Z.excl, Z.excl, X.excl) 
  inv.cov.mtx.agg.MMIV <- solve(cov.mtx.agg.MMIV)
  
  ######### Compute  GMM:  AVG.GMM.MMIV #######################     begin ####################
  # use inverse of cov.mtx as the weighting matrix to construct the GMM criterion. run GenSA with GMM. get b
  tmp.MMIV.1step <- ivqr.iv.aggressive(Y = Y, X.excl = X.excl, Z.excl = Z.excl, D = D, tau = tau, h = h.small, dB = dB, max.time=5, upper=rep(100,dB), lower=rep(-100,dB),
                                       weight.mtx=inv.cov.mtx.agg.MMIV, structure = 'iid', LRV.kernel = 'QS', b.init=b.conservative  ) 
  
  b.MMIV.1step <- tmp.MMIV.1step$b #(slope intercept)
  ######### Compute  GMM:  AVG.GMM.MMIV #######################     END ####################
  
  
  # 
  # store b.aggressive
  b.MMIV.1step.intslp<-as.matrix(c(b.MMIV.1step[7], b.MMIV.1step[1:6]),ncol=1)
  
  b.AGG.MMIV[1,irep] <-b.MMIV.1step[7]
  b.AGG.MMIV[2:7,irep] <-b.MMIV.1step[1:6]
  

  # ##################### compute sigma.aggressive ###### evaluated at b.conservative  
  # use b.conservative to calculate sigma2. i.e. sigma.aggressive.

  L.agg.MMIV.sig <- matrix(Lambda.gmmq(y=Y.gmmq,x=X.excl,b=b.conservative), ncol=1) # HAS INTERCETP
  Lp.agg.MMIV.sig <- Lambda.derivative.gmmq(y=Y.gmmq, x=X.excl, b=b.conservative)  # note. b. is (slope, intercept), thus need to use Lambda.derivative.gmmq instead of Lambda.derivative, bc Lambda.derivative.gmmq give X in (D, X.excl) order; Lambda.derivative gives X in (X.excl, D) order.
  gni.agg.MMIV.sig <- cbind(demean.Z.excl*repmat(L.agg.MMIV.sig, 1, ncol(Z.excl)), cbind(Z.excl, X.excl)*repmat((Gfn(-L.agg.MMIV.sig, h.small)-tau),1,dZ))     #gni is in (demean.Z.excl, Z.excl, X.excl) order  #HAS INTERCETP
  g.bar.agg.MMIV.sig <- colMeans(gni.agg.MMIV.sig)
  
  h.use.MMIV.sig <- h.Kato
  jac.agg.MMIV.sig <- rbind(t(demean.Z.excl)%*%(Lp.agg.MMIV.sig)/n, t(cbind(Z.excl, X.excl))%*%(Lp.agg.MMIV.sig*repmat(Gpfn(-L.agg.MMIV.sig, h.use.MMIV.sig),1,dB))/(-h.use.MMIV.sig)/n) #has intercept
  
  Omega.bar.agg.MMIV.sig <- t(gni.agg.MMIV.sig)%*%gni.agg.MMIV.sig/n-g.bar.agg.MMIV.sig%*%t(g.bar.agg.MMIV.sig) # iid

  sig.agg.MMIV <- tryCatch(solve(t(jac.agg.MMIV.sig)%*%solve(Omega.bar.agg.MMIV.sig)%*%jac.agg.MMIV.sig),
                           error=function(w)NA)
  if (is.na(sig.agg.MMIV[1])) next
  # ##################### compute sigma.aggressive ###### evaluated at b.conservative 


  # "averaging"
  
  TR2 <- Trace(Ups%*%(sig.conservative-sig.agg.MMIV))
  
  if (TR2<0){
    w.avg.MMIV[irep, ] <- MMIVweight <- 0
  }else{
    w.avg.MMIV[irep,] <- MMIVweight <- max(0, min(1, TR2/(n*t(b.MMIV.1step-b.conservative)%*%Ups%*%(b.MMIV.1step-b.conservative)+TR2)))
  }

  b.avgMMIV<- repmat((1-MMIVweight),dB,1)*b.conservative+repmat(MMIVweight,dB,1)*b.MMIV.1step

  b.avgMMIV.intslp<-as.matrix(c(b.avgMMIV[7], b.avgMMIV[1:6]),ncol=1)
  
  b.AVG.MMIV[1,irep] <-b.avgMMIV[7]
  b.AVG.MMIV[2:7,irep] <-b.avgMMIV[1:6]
  
  
  
  ######################################################################################################## end of MMIV
  
  
  
  

  
  ################# stop here
  #########not run bootstrap for the nreplic=1000
  
  if (BS.FLAG){
    # bootstrap
    # AVG.BS  :(for each simulation sample, run bootstrap and pick the optimal weight based on miniminizing RMSE/rRMSE.)
    b.hats.original <- cbind(b.CON[,irep], b.IV[,irep], b.QR[,irep] ) #(INTERCEPT, SLOPE)
    b.IVQR <- b.CON[,irep]

    tmp.BS <- AVGBS.multipar(Y = Y, D = D, X.excl=X.excl, Z.excl = Z.excl, tau = tau, n=n, h = h.small, dB = dB, BREPLIC=BREPLIC, reportGMM=reportGMM, weights.new=weights.new,
                    b.hats.original=b.hats.original, b.IVQR=b.IVQR, DGP=DGP, BS.GMMQ.FLAG=BS.GMMQ.FLAG)
    
    b.AVG.BS.MSE[,irep] <- b.bs.mse <- tmp.BS$b.AVGBS.MSE
    b.AVG.BS.RRMSE[,irep] <- b.bs.rrmse <- tmp.BS$b.AVGBS.RRMSE
    
  } # end of computing b.BS results

  
  # MSE computed inside the irep loop. 
  b.CON.MSE[ipar,irep] <- mse.b.con <- n*(b.CON[,irep]-repmat(B.TRUE,1,dB))%*%Ups%*%t(b.CON[,irep]-repmat(B.TRUE,1,dB))
  b.SEE.MSE[ipar,irep] <- mse.b.see <- n*(b.SEE[,irep]-repmat(B.TRUE,1,dB))%*%Ups%*%t(b.SEE[,irep]-repmat(B.TRUE,1,dB))
  b.IV.MSE[ipar,irep] <- mse.b.iv <- n*(b.IV[,irep]-repmat(B.TRUE,1,dB))%*%Ups%*%t(b.IV[,irep]-repmat(B.TRUE,1,dB))
  b.QR.MSE[ipar,irep] <- mse.b.qr <-n*(b.QR[,irep]-repmat(B.TRUE,1,dB))%*%Ups%*%t(b.QR[,irep]-repmat(B.TRUE,1,dB))
  b.AVG.MMIV.MSE[ipar,irep] <- mse.b.avgmmiv <- n*(b.AVG.MMIV[,irep]-repmat(B.TRUE,1,dB))%*%Ups%*%t(b.AVG.MMIV[,irep]-repmat(B.TRUE,1,dB))
  b.AGG.MMIV.MSE[ipar,irep] <- mse.b.aggmmiv <- n*(b.AGG.MMIV[,irep]-repmat(B.TRUE,1,dB))%*%Ups%*%t(b.AGG.MMIV[,irep]-repmat(B.TRUE,1,dB))
  b.AVG.MMQR.MSE[ipar,irep] <- mse.b.avgmmqr <- n*(b.AVG.MMQR[,irep]-repmat(B.TRUE,1,dB))%*%Ups%*%t(b.AVG.MMQR[,irep]-repmat(B.TRUE,1,dB))
  b.AGG.MMQR.MSE[ipar,irep] <- mse.b.aggmmqr <- n*(b.AGG.MMQR[,irep]-repmat(B.TRUE,1,dB))%*%Ups%*%t(b.AGG.MMQR[,irep]-repmat(B.TRUE,1,dB))
  b.GMMQ.MSE[ipar,irep] <- mse.b.gmmq <- n*(b.GMMQ[,irep]-repmat(B.TRUE,1,dB))%*%Ups%*%t(b.GMMQ[,irep]-repmat(B.TRUE,1,dB))

  if(BS.FLAG){
    b.AVG.BS.MSE.MSE[ipar,irep] <- mse.bs.mse <- n*(b.AVG.BS.MSE[,irep]-repmat(B.TRUE,1,dB))%*%Ups%*%t(b.AVG.BS.MSE[,irep]-repmat(B.TRUE,1,dB))
    b.AVG.BS.RRMSE.MSE[ipar,irep] <- mse.bs.rrmse <- n*(b.AVG.BS.RRMSE[,irep]-repmat(B.TRUE,1,dB))%*%Ups%*%t(b.AVG.BS.RRMSE[,irep]-repmat(B.TRUE,1,dB))
  }

  # slice 
 
       if(!BS.FLAG){
         b.bs.mse <- as.matrix(c("b.bs.mseNA1","b.bs.mseNA2","b.bs.mseNA3","b.bs.mseNA4","b.bs.mseNA5","b.bs.mseNA6","b.bs.mseNA7"), ncol=1, nrow=dB)
         b.bs.rrmse <- as.matrix(c("b.bs.rrmseNA1","b.bs.rrmseNA2","b.bs.rrmseNA3","b.bs.rrmseNA4","b.bs.rrmseNA5","b.bs.rrmseNA6","b.bs.rrmseNA7"), ncol=1, nrow=dB)
         mse.bs.mse <- as.matrix(c("NA.MSE.bs.mse"), ncol=1, nrow=1)
         mse.bs.rrmse <- as.matrix(c("NA.MSE.bs.rrmse"), ncol=1, nrow=1)
       }


       
       SLI<-rbind(as.matrix(c(b.conservative.intslp), ncol=1, nrow=dB), #1-7
                  as.matrix(c(tmp.SEE$b), ncol=1, nrow=dB), #8-14
                  as.matrix(c(b.2sls), ncol=1, nrow=dB), #15-21
                  as.matrix(c(b.qr), ncol=1, nrow=dB), #22-28
                  as.matrix(c(b.avgMMIV.intslp), ncol=1, nrow=dB), #29-35
                  as.matrix(c(b.MMIV.1step.intslp), ncol=1, nrow=dB), #36-42
                  as.matrix(c(b.avgMMQR.intslp), ncol=1, nrow=dB),  #43-49
                  as.matrix(c(b.MMQR.1step.intslp), ncol=1, nrow=dB), #50-56
                  as.matrix(c(b.bs.mse), ncol=1, nrow=dB), #57-63
                  as.matrix(c(b.bs.rrmse), ncol=1, nrow=dB), #71-77   ## NEW NEW: #64-70
                  as.matrix(c(b.gmmq.intslp), ncol=1, nrow=dB), #78-84 ## NEW NEW: #71-77 
                  as.matrix(c(mse.b.con), ncol=1, nrow=1), #90 #new 85 ## NEW NEW: 78
                  as.matrix(c(mse.b.see), ncol=1, nrow=1),  #new 86    ## NEW NEW: 79
                  as.matrix(c(mse.b.iv), ncol=1, nrow=1),   #new 87    ## NEW NEW: 80
                  as.matrix(c(mse.b.qr), ncol=1, nrow=1),   #new 88    ## NEW NEW: 81
                  as.matrix(c(mse.b.avgmmiv), ncol=1, nrow=1), #new 89 ## NEW NEW: 82
                  as.matrix(c(mse.b.aggmmiv), ncol=1, nrow=1), #new 90 ## NEW NEW: 83
                  as.matrix(c(mse.b.avgmmqr), ncol=1, nrow=1), #new 91 ## NEW NEW: 84 
                  as.matrix(c(mse.b.aggmmqr), ncol=1, nrow=1), #new 92 ## NEW NEW: 85
                  as.matrix(c(mse.b.gmmq), ncol=1, nrow=1),  #new 93   ## NEW NEW: 86
                  as.matrix(c(mse.bs.mse), ncol=1, nrow=1),  #new 94   ## NEW NEW: 87
                  as.matrix(c(mse.bs.rrmse), ncol=1, nrow=1)  #new 95  ## NEW NEW: 88
       )
       

   matrix(SLI,ncol=1)
 
  
  
  
  
} # end of irep loop, (same end of irep foreach loop)

########
  b.CON[1,]<-  t(as.numeric(bp.raw[1, ]))   # if non-parallel, comment
  b.CON[2,]<-  t(as.numeric(bp.raw[2, ]))   # if non-parallel, comment
  b.CON[3,]<-  t(as.numeric(bp.raw[3, ]))   # if non-parallel, comment
  b.CON[4,]<-  t(as.numeric(bp.raw[4, ]))   # if non-parallel, comment
  b.CON[5,]<-  t(as.numeric(bp.raw[5, ]))   # if non-parallel, comment
  b.CON[6,]<-  t(as.numeric(bp.raw[6, ]))   # if non-parallel, comment
  b.CON[7,]<-  t(as.numeric(bp.raw[7, ]))   # if non-parallel, comment
  
  b.SEE[1,]<-  t(as.numeric(bp.raw[8, ]))   # if non-parallel, comment
  b.SEE[2,]<-  t(as.numeric(bp.raw[9, ]))   # if non-parallel, comment
  b.SEE[3,]<-  t(as.numeric(bp.raw[10, ]))   # if non-parallel, comment
  b.SEE[4,]<-  t(as.numeric(bp.raw[11, ]))   # if non-parallel, comment
  b.SEE[5,]<-  t(as.numeric(bp.raw[12, ]))   # if non-parallel, comment
  b.SEE[6,]<-  t(as.numeric(bp.raw[13, ]))   # if non-parallel, comment
  b.SEE[7,]<-  t(as.numeric(bp.raw[14, ]))   # if non-parallel, comment
  
  b.IV[1,]<-  t(as.numeric(bp.raw[15, ]))   # if non-parallel, comment
  b.IV[2,]<-  t(as.numeric(bp.raw[16, ]))   # if non-parallel, comment
  b.IV[3,]<-  t(as.numeric(bp.raw[17, ]))   # if non-parallel, comment
  b.IV[4,]<-  t(as.numeric(bp.raw[18, ]))   # if non-parallel, comment
  b.IV[5,]<-  t(as.numeric(bp.raw[19, ]))   # if non-parallel, comment
  b.IV[6,]<-  t(as.numeric(bp.raw[20, ]))   # if non-parallel, comment
  b.IV[7,]<-  t(as.numeric(bp.raw[21, ]))   # if non-parallel, comment
  
  b.QR[1,]<-  t(as.numeric(bp.raw[22, ]))   # if non-parallel, comment
  b.QR[2,]<-  t(as.numeric(bp.raw[23, ]))   # if non-parallel, comment
  b.QR[3,]<-  t(as.numeric(bp.raw[24, ]))   # if non-parallel, comment
  b.QR[4,]<-  t(as.numeric(bp.raw[25, ]))   # if non-parallel, comment
  b.QR[5,]<-  t(as.numeric(bp.raw[26, ]))   # if non-parallel, comment
  b.QR[6,]<-  t(as.numeric(bp.raw[27, ]))   # if non-parallel, comment
  b.QR[7,]<-  t(as.numeric(bp.raw[28, ]))   # if non-parallel, comment
  
  b.AVG.MMIV[1,]<-  t(as.numeric(bp.raw[29, ]))   # if non-parallel, comment
  b.AVG.MMIV[2,]<-  t(as.numeric(bp.raw[30, ]))   # if non-parallel, comment
  b.AVG.MMIV[3,]<-  t(as.numeric(bp.raw[31, ]))   # if non-parallel, comment
  b.AVG.MMIV[4,]<-  t(as.numeric(bp.raw[32, ]))   # if non-parallel, comment
  b.AVG.MMIV[5,]<-  t(as.numeric(bp.raw[33, ]))   # if non-parallel, comment
  b.AVG.MMIV[6,]<-  t(as.numeric(bp.raw[34, ]))   # if non-parallel, comment
  b.AVG.MMIV[7,]<-  t(as.numeric(bp.raw[35, ]))   # if non-parallel, comment
  
  b.AGG.MMIV[1,]<-  t(as.numeric(bp.raw[36, ]))   # if non-parallel, comment
  b.AGG.MMIV[2,]<-  t(as.numeric(bp.raw[37, ]))   # if non-parallel, comment
  b.AGG.MMIV[3,]<-  t(as.numeric(bp.raw[38, ]))   # if non-parallel, comment
  b.AGG.MMIV[4,]<-  t(as.numeric(bp.raw[39, ]))   # if non-parallel, comment
  b.AGG.MMIV[5,]<-  t(as.numeric(bp.raw[40, ]))   # if non-parallel, comment
  b.AGG.MMIV[6,]<-  t(as.numeric(bp.raw[41, ]))   # if non-parallel, comment
  b.AGG.MMIV[7,]<-  t(as.numeric(bp.raw[42, ]))   # if non-parallel, comment
  
  b.AVG.MMQR[1,]<-  t(as.numeric(bp.raw[43, ]))   # if non-parallel, comment
  b.AVG.MMQR[2,]<-  t(as.numeric(bp.raw[44, ]))   # if non-parallel, comment
  b.AVG.MMQR[3,]<-  t(as.numeric(bp.raw[45, ]))   # if non-parallel, comment
  b.AVG.MMQR[4,]<-  t(as.numeric(bp.raw[46, ]))   # if non-parallel, comment
  b.AVG.MMQR[5,]<-  t(as.numeric(bp.raw[47, ]))   # if non-parallel, comment
  b.AVG.MMQR[6,]<-  t(as.numeric(bp.raw[48, ]))   # if non-parallel, comment
  b.AVG.MMQR[7,]<-  t(as.numeric(bp.raw[49, ]))   # if non-parallel, comment
  
  b.AGG.MMQR[1,]<-  t(as.numeric(bp.raw[50, ]))   # if non-parallel, comment
  b.AGG.MMQR[2,]<-  t(as.numeric(bp.raw[51, ]))   # if non-parallel, comment
  b.AGG.MMQR[3,]<-  t(as.numeric(bp.raw[52, ]))   # if non-parallel, comment
  b.AGG.MMQR[4,]<-  t(as.numeric(bp.raw[53, ]))   # if non-parallel, comment
  b.AGG.MMQR[5,]<-  t(as.numeric(bp.raw[54, ]))   # if non-parallel, comment
  b.AGG.MMQR[6,]<-  t(as.numeric(bp.raw[55, ]))   # if non-parallel, comment
  b.AGG.MMQR[7,]<-  t(as.numeric(bp.raw[56, ]))   # if non-parallel, comment
  
  b.AVG.BS.MSE[1,]<-  t(as.numeric(bp.raw[57, ]))   # if non-parallel, comment
  b.AVG.BS.MSE[2,]<-  t(as.numeric(bp.raw[58, ]))   # if non-parallel, comment
  b.AVG.BS.MSE[3,]<-  t(as.numeric(bp.raw[59, ]))   # if non-parallel, comment
  b.AVG.BS.MSE[4,]<-  t(as.numeric(bp.raw[60, ]))   # if non-parallel, comment
  b.AVG.BS.MSE[5,]<-  t(as.numeric(bp.raw[61, ]))   # if non-parallel, comment
  b.AVG.BS.MSE[6,]<-  t(as.numeric(bp.raw[62, ]))   # if non-parallel, comment
  b.AVG.BS.MSE[7,]<-  t(as.numeric(bp.raw[63, ]))   # if non-parallel, comment

  b.AVG.BS.RRMSE[1,]<-  t(as.numeric(bp.raw[64, ]))   # if non-parallel, comment
  b.AVG.BS.RRMSE[2,]<-  t(as.numeric(bp.raw[65, ]))   # if non-parallel, comment
  b.AVG.BS.RRMSE[3,]<-  t(as.numeric(bp.raw[66, ]))   # if non-parallel, comment
  b.AVG.BS.RRMSE[4,]<-  t(as.numeric(bp.raw[67, ]))   # if non-parallel, comment
  b.AVG.BS.RRMSE[5,]<-  t(as.numeric(bp.raw[68, ]))   # if non-parallel, comment
  b.AVG.BS.RRMSE[6,]<-  t(as.numeric(bp.raw[69, ]))   # if non-parallel, comment
  b.AVG.BS.RRMSE[7,]<-  t(as.numeric(bp.raw[70, ]))   # if non-parallel, comment
  
  b.GMMQ[1,]<-  t(as.numeric(bp.raw[71, ]))   # if non-parallel, comment
  b.GMMQ[2,]<-  t(as.numeric(bp.raw[72, ]))   # if non-parallel, comment
  b.GMMQ[3,]<-  t(as.numeric(bp.raw[73, ]))   # if non-parallel, comment
  b.GMMQ[4,]<-  t(as.numeric(bp.raw[74, ]))   # if non-parallel, comment
  b.GMMQ[5,]<-  t(as.numeric(bp.raw[75, ]))   # if non-parallel, comment
  b.GMMQ[6,]<-  t(as.numeric(bp.raw[76, ]))   # if non-parallel, comment
  b.GMMQ[7,]<-  t(as.numeric(bp.raw[77, ]))   # if non-parallel, comment
  

if(!BS.FLAG){  # if non-parallel, comment
  b.AVG.BS.MSE[1,]<- b.AVG.BS.MSE[2,]<- b.AVG.BS.MSE[3,]<- b.AVG.BS.MSE[4,]<- b.AVG.BS.MSE[5,]<- b.AVG.BS.MSE[6,]<-  b.AVG.BS.MSE[7,]<- t(array(NA,dim=c(NREPLIC,1)))  # if non-parallel, comment
  b.AVG.BS.RRMSE[1,]<- b.AVG.BS.RRMSE[2,]<- b.AVG.BS.RRMSE[3,]<- b.AVG.BS.RRMSE[4,]<- b.AVG.BS.RRMSE[5,]<- b.AVG.BS.RRMSE[6,]<- b.AVG.BS.RRMSE[7,]<- t(array(NA,dim=c(NREPLIC,1)))  # if non-parallel, comment
}  # if non-parallel, comment
  
  

  b.CON.MSE[ipar,] <- t(as.numeric(bp.raw[78,]))  # if non-parallel, comment
  b.SEE.MSE[ipar,] <- t(as.numeric(bp.raw[79,]))  # if non-parallel, comment
  b.IV.MSE[ipar,] <- t(as.numeric(bp.raw[80,]))  # if non-parallel, comment
  b.QR.MSE[ipar,] <- t(as.numeric(bp.raw[81,]))  # if non-parallel, comment
  b.AVG.MMIV.MSE[ipar,] <- t(as.numeric(bp.raw[82,]))  # if non-parallel, comment
  b.AGG.MMIV.MSE[ipar,] <- t(as.numeric(bp.raw[83,]))  # if non-parallel, comment
  b.AVG.MMQR.MSE[ipar,] <- t(as.numeric(bp.raw[84,]))  # if non-parallel, comment
  b.AGG.MMQR.MSE[ipar,] <- t(as.numeric(bp.raw[85,]))  # if non-parallel, comment
  b.GMMQ.MSE[ipar,] <- t(as.numeric(bp.raw[86,]))  # if non-parallel, comment
  
  b.AVG.BS.MSE.MSE[ipar,] <- t(as.numeric(bp.raw[87,]))  # if non-parallel, comment
  b.AVG.BS.RRMSE.MSE[ipar,] <- t(as.numeric(bp.raw[88,]))  # if non-parallel, comment
  
if(!BS.FLAG){
  b.AVG.BS.MSE.MSE[ipar,] <- b.AVG.BS.RRMSE.MSE[ipar,] <- t(array(NA,dim=c(NREPLIC,1)))
}

######## SUM LOSS value over replication to compute MSE
b.CON.MSE.AVG[,1] <- rowMeans(b.CON.MSE)
b.SEE.MSE.AVG[,1] <- rowMeans(b.SEE.MSE)
b.IV.MSE.AVG[,1] <- rowMeans(b.IV.MSE)
b.QR.MSE.AVG[,1] <- rowMeans(b.QR.MSE)
b.AVG.MMIV.MSE.AVG[,1] <- rowMeans(b.AVG.MMIV.MSE)
b.AGG.MMIV.MSE.AVG[,1] <- rowMeans(b.AGG.MMIV.MSE)
b.AVG.MMQR.MSE.AVG[,1] <- rowMeans(b.AVG.MMQR.MSE)
b.AGG.MMQR.MSE.AVG[,1] <- rowMeans(b.AGG.MMQR.MSE)
b.GMMQ.MSE.AVG[,1] <- rowMeans(b.GMMQ.MSE)

if(BS.FLAG){
b.AVG.BS.MSE.MSE.AVG[,1] <- rowMeans(b.AVG.BS.MSE.MSE)
b.AVG.BS.RRMSE.MSE.AVG[,1] <- rowMeans(b.AVG.BS.RRMSE.MSE)
}

########## rRMSE=SUM 
b.CON.tr <- t(b.CON[2:7,])
b.CON.median.bias <- apply(X = b.CON.tr, MARGIN = 2, FUN = median) - B.TRUE
b.CON.IQR <- apply(X = b.CON.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.CON.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.CON.rRMSE[ipar] <- sqrt(n*sum(b.CON.median.bias^2 + (b.CON.IQR/1.349)^2))

b.SEE.tr <- t(b.SEE[2:7,])
b.SEE.median.bias <- apply(X = b.SEE.tr, MARGIN = 2, FUN = median) - B.TRUE
b.SEE.IQR <- apply(X = b.SEE.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.SEE.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.SEE.rRMSE[ipar] <- sqrt(n*sum(b.SEE.median.bias^2 + (b.SEE.IQR/1.349)^2))

b.IV.tr <- t(b.IV[2:7,])
b.IV.median.bias <- apply(X = b.IV.tr, MARGIN = 2, FUN = median) - B.TRUE
b.IV.IQR <- apply(X = b.IV.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.IV.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.IV.rRMSE[ipar] <- sqrt(n*sum(b.IV.median.bias^2 + (b.IV.IQR/1.349)^2))

b.QR.tr <- t(b.QR[2:7,])
b.QR.median.bias <- apply(X = b.QR.tr, MARGIN = 2, FUN = median) - B.TRUE
b.QR.IQR <- apply(X = b.QR.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.QR.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.QR.rRMSE[ipar] <- sqrt(n*sum(b.QR.median.bias^2 + (b.QR.IQR/1.349)^2))

b.AVG.MMIV.tr <- t(b.AVG.MMIV[2:7,])
b.AVG.MMIV.median.bias <- apply(X = b.AVG.MMIV.tr, MARGIN = 2, FUN = median) - B.TRUE
b.AVG.MMIV.IQR <- apply(X = b.AVG.MMIV.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.AVG.MMIV.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.AVG.MMIV.rRMSE[ipar] <- sqrt(n*sum(b.AVG.MMIV.median.bias^2 + (b.AVG.MMIV.IQR/1.349)^2))

b.AGG.MMIV.tr <- t(b.AGG.MMIV[2:7,])
b.AGG.MMIV.median.bias <- apply(X = b.AGG.MMIV.tr, MARGIN = 2, FUN = median) - B.TRUE
b.AGG.MMIV.IQR <- apply(X = b.AGG.MMIV.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.AGG.MMIV.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.AGG.MMIV.rRMSE[ipar] <- sqrt(n*sum(b.AGG.MMIV.median.bias^2 + (b.AGG.MMIV.IQR/1.349)^2))

b.AVG.MMQR.tr <- t(b.AVG.MMQR[2:7,])
b.AVG.MMQR.median.bias <- apply(X = b.AVG.MMQR.tr, MARGIN = 2, FUN = median) - B.TRUE
b.AVG.MMQR.IQR <- apply(X = b.AVG.MMQR.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.AVG.MMQR.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.AVG.MMQR.rRMSE[ipar] <- sqrt(n*sum(b.AVG.MMQR.median.bias^2 + (b.AVG.MMQR.IQR/1.349)^2))

b.AGG.MMQR.tr <- t(b.AGG.MMQR[2:7,])
b.AGG.MMQR.median.bias <- apply(X = b.AGG.MMQR.tr, MARGIN = 2, FUN = median) - B.TRUE
b.AGG.MMQR.IQR <- apply(X = b.AGG.MMQR.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.AGG.MMQR.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.AGG.MMQR.rRMSE[ipar] <- sqrt(n*sum(b.AGG.MMQR.median.bias^2 + (b.AGG.MMQR.IQR/1.349)^2))

b.AVG.BS.MSE.tr <- t(b.AVG.BS.MSE[2:7,])
b.AVG.BS.MSE.median.bias <- apply(X = b.AVG.BS.MSE.tr, MARGIN = 2, FUN = median) - B.TRUE
b.AVG.BS.MSE.IQR <- apply(X = b.AVG.BS.MSE.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.AVG.BS.MSE.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.AVG.BS.MSE.rRMSE[ipar] <- sqrt(n*sum(b.AVG.BS.MSE.median.bias^2 + (b.AVG.BS.MSE.IQR/1.349)^2))


b.AVG.BS.RRMSE.tr <- t(b.AVG.BS.RRMSE[2:7,])
b.AVG.BS.RRMSE.median.bias <- apply(X = b.AVG.BS.RRMSE.tr, MARGIN = 2, FUN = median) - B.TRUE
b.AVG.BS.RRMSE.IQR <- apply(X = b.AVG.BS.RRMSE.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.AVG.BS.RRMSE.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.AVG.BS.RRMSE.rRMSE[ipar] <- sqrt(n*sum(b.AVG.BS.RRMSE.median.bias^2 + (b.AVG.BS.RRMSE.IQR/1.349)^2))

b.GMMQ.tr <- t(b.GMMQ[2:7,])
b.GMMQ.median.bias <- apply(X = b.GMMQ.tr, MARGIN = 2, FUN = median) - B.TRUE
b.GMMQ.IQR <- apply(X = b.GMMQ.tr, MARGIN = 2, FUN = quantile, probs=0.75, type=6, na.rm=TRUE) - apply(X = b.GMMQ.tr, MARGIN = 2, FUN = quantile, probs=0.25, type=6, na.rm=TRUE)
b.GMMQ.rRMSE[ipar] <- sqrt(n*sum(b.GMMQ.median.bias^2 + (b.GMMQ.IQR/1.349)^2))


# to report relative results: #relIV=NA, relQR=NA, relAVGGMM1.MMQR=NA, relAVGGMM2.MMIV=NA  ######## rRMSE
b.hats.rRMSE.par.rel[ipar, 'CON' ] <- b.CON.rRMSE[ipar]
b.hats.rRMSE.par.rel[ipar, 'relSEE' ] <- b.SEE.rRMSE[ipar] /b.CON.rRMSE[ipar] 

b.hats.rRMSE.par.rel[ipar, 'relIV' ] <- b.IV.rRMSE[ipar] /b.CON.rRMSE[ipar]
b.hats.rRMSE.par.rel[ipar, 'relQR' ] <- b.QR.rRMSE[ipar] /b.CON.rRMSE[ipar]
b.hats.rRMSE.par.rel[ipar, 'relAVG.MMIV' ] <- b.AVG.MMIV.rRMSE[ipar] /b.CON.rRMSE[ipar]
b.hats.rRMSE.par.rel[ipar, 'relAGG.MMIV' ] <- b.AGG.MMIV.rRMSE[ipar] /b.CON.rRMSE[ipar] 
b.hats.rRMSE.par.rel[ipar, 'relAVG.MMQR' ] <- b.AVG.MMQR.rRMSE[ipar] /b.CON.rRMSE[ipar]
b.hats.rRMSE.par.rel[ipar, 'relAGG.MMQR' ] <- b.AGG.MMQR.rRMSE[ipar] /b.CON.rRMSE[ipar]

b.hats.rRMSE.par.rel[ipar, 'relAVG.BS.MSE' ] <- b.AVG.BS.MSE.rRMSE[ipar] /b.CON.rRMSE[ipar] 
b.hats.rRMSE.par.rel[ipar, 'relAVG.BS.RRMSE' ] <- b.AVG.BS.RRMSE.rRMSE[ipar] /b.CON.rRMSE[ipar] 

b.hats.rRMSE.par.rel[ipar, 'relGMMQ' ] <- b.GMMQ.rRMSE[ipar] /b.CON.rRMSE[ipar] 



# to report relative results: #relIV=NA, relQR=NA, relAVGGMM1.MMQR=NA, relAVGGMM2.MMIV=NA  ######## MSE
b.hats.MSE.par.rel[ipar, 'CON' ] <- b.CON.MSE.AVG[ipar]
b.hats.MSE.par.rel[ipar, 'relSEE' ] <- b.SEE.MSE.AVG[ipar] /b.CON.MSE.AVG[ipar]

b.hats.MSE.par.rel[ipar, 'relIV' ] <- b.IV.MSE.AVG[ipar] /b.CON.MSE.AVG[ipar]
b.hats.MSE.par.rel[ipar, 'relQR' ] <- b.QR.MSE.AVG[ipar] /b.CON.MSE.AVG[ipar]
b.hats.MSE.par.rel[ipar, 'relAVG.MMIV' ] <- b.AVG.MMIV.MSE.AVG[ipar] /b.CON.MSE.AVG[ipar]
b.hats.MSE.par.rel[ipar, 'relAGG.MMIV' ] <- b.AGG.MMIV.MSE.AVG[ipar] /b.CON.MSE.AVG[ipar]
b.hats.MSE.par.rel[ipar, 'relAVG.MMQR' ] <- b.AVG.MMQR.MSE.AVG[ipar] /b.CON.MSE.AVG[ipar]
b.hats.MSE.par.rel[ipar, 'relAGG.MMQR' ] <- b.AGG.MMQR.MSE.AVG[ipar] /b.CON.MSE.AVG[ipar]

b.hats.MSE.par.rel[ipar, 'relAVG.BS.MSE' ] <- b.AVG.BS.MSE.MSE.AVG[ipar] /b.CON.MSE.AVG[ipar]
b.hats.MSE.par.rel[ipar, 'relAVG.BS.RRMSE' ] <- b.AVG.BS.RRMSE.MSE.AVG[ipar] /b.CON.MSE.AVG[ipar]

b.hats.MSE.par.rel[ipar, 'relGMMQ' ] <- b.GMMQ.MSE.AVG[ipar] /b.CON.MSE.AVG[ipar]

### output the b.hats.rRMSE.rel table
if (!Hetero.FLAG){
  hetero.report <- 0
}else{ hetero.report <- hetero
}


rel.rRMSE[ipar,] <- cbind(b.hats.rRMSE.par.rel[ipar,], c0, hetero.report)
rel.MSE[ipar,] <- cbind(b.hats.MSE.par.rel[ipar,], c0, hetero.report)

write.csv(rel.rRMSE, file = "b.hats.rRMSE.par.rel.new.csv")
write.csv(rel.MSE, file = "b.hats.MSE.par.rel.new.csv")

write.csv(rel.rRMSE, file = paste0(sprintf("b.hats.rRMSE.par.rel.new_VH_NREP%d_BREP%d_tau%.2f_n%d_DGP%d_S1S2%d", NREPLIC, BREPLIC, tau, n, nrow(para.space), S1S2), format(Sys.time(),"%Y_%m_%d"),".csv") )
write.csv(rel.MSE, file = paste0(sprintf("b.hats.MSE.par.rel.new_VH_NREP%d_BREP%d_tau%.2f_n%d_DGP%d_S1S2%d",NREPLIC, BREPLIC, tau, n, nrow(para.space), S1S2), format(Sys.time(),"%Y_%m_%d"),".csv"))
####################################################################################################################################################
write.csv(b.CON, file = paste0(sprintf("b.CON_VH_NREP%d_BREP%d_tau%.2f_n%d_DGP%d__ipar%d",NREPLIC, BREPLIC, tau, n, nrow(para.space), ipar), format(Sys.time(),"%Y_%m_%d"),".csv"))
write.csv(b.SEE, file = paste0(sprintf("b.SEE_VH_NREP%d_BREP%d_tau%.2f_n%d_DGP%d__ipar%d",NREPLIC, BREPLIC, tau, n, nrow(para.space), ipar), format(Sys.time(),"%Y_%m_%d"),".csv"))
write.csv(b.IV, file = paste0(sprintf("b.IV_VH_NREP%d_BREP%d_tau%.2f_n%d_DGP%d__ipar%d",NREPLIC, BREPLIC, tau, n, nrow(para.space), ipar), format(Sys.time(),"%Y_%m_%d"),".csv"))
write.csv(b.QR, file = paste0(sprintf("b.QR_VH_NREP%d_BREP%d_tau%.2f_n%d_DGP%d__ipar%d",NREPLIC, BREPLIC, tau, n, nrow(para.space), ipar), format(Sys.time(),"%Y_%m_%d"),".csv"))
write.csv(b.AVG.MMIV, file = paste0(sprintf("b.AVG.MMIV_VH_NREP%d_BREP%d_tau%.2f_n%d_DGP%d__ipar%d",NREPLIC, BREPLIC, tau, n, nrow(para.space), ipar), format(Sys.time(),"%Y_%m_%d"),".csv"))
write.csv(b.AGG.MMIV, file = paste0(sprintf("b.AGG.MMIV_VH_NREP%d_BREP%d_tau%.2f_n%d_DGP%d__ipar%d",NREPLIC, BREPLIC, tau, n, nrow(para.space), ipar), format(Sys.time(),"%Y_%m_%d"),".csv"))
write.csv(b.AVG.MMQR, file = paste0(sprintf("b.AVG.MMQR.new_VH_NREP%d_BREP%d_tau%.2f_n%d_DGP%d__ipar%d",NREPLIC, BREPLIC, tau, n, nrow(para.space), ipar), format(Sys.time(),"%Y_%m_%d"),".csv"))
write.csv(b.AGG.MMQR, file = paste0(sprintf("b.AGG.MMQR_VH_NREP%d_BREP%d_tau%.2f_n%d_DGP%d__ipar%d",NREPLIC, BREPLIC, tau, n, nrow(para.space), ipar), format(Sys.time(),"%Y_%m_%d"),".csv"))
write.csv(b.AVG.BS.MSE, file = paste0(sprintf("b.AVG.BS.MSE_VH_NREP%d_BREP%d_tau%.2f_n%d_DGP%d_ipar%d",NREPLIC, BREPLIC, tau, n, nrow(para.space), ipar), format(Sys.time(),"%Y_%m_%d"),".csv"))
write.csv(b.AVG.BS.RRMSE, file = paste0(sprintf("b.AVG.BS.RRMSE_VH_NREP%d_BREP%d_tau%.2f_n%d_DGP%d__ipar%d",NREPLIC, BREPLIC, tau, n, nrow(para.space), ipar), format(Sys.time(),"%Y_%m_%d"),".csv"))
write.csv(b.GMMQ, file = paste0(sprintf("b.GMMQ_VH_NREP%d_BREP%d_tau%.2f_n%d_DGP%d__ipar%d",NREPLIC, BREPLIC, tau, n, nrow(para.space), ipar), format(Sys.time(),"%Y_%m_%d"),".csv"))
####################################################################################################################################################
###################################




} # end of ipar loop (loop for c)



# ###################################################################################################################################################
# ###################################################################################################################################################
# ###################################################################################################################################################
# ###################################################################################################################################################
# ###                                                                                                                                             ###
# ###                                                                                                                                             ###
# ###                                                                                                                                             ###
# ###                                                                                                                                             ###
# ###                                                                                                                                             ###
# ###                                                                                                                                             ###
# ###                                                                                                                                             ###
# ###################################################################################################################################################
# ###################################################################################################################################################
# ###################################################################################################################################################
# ###################################################################################################################################################
# ############################################################################################################################
# # finally report results into .txt file.
# ############################################################################################################################
# # 
# LaTeX table output


###########################################################################################################################################################
## CLS-VaryHetero (ALL ESTIMATES) TABLE ####################  BEGIN  ################### seperate
for(ibtab in 1:6){
st<-(ibtab-1)*11+1
ed<-ibtab*11
if(ibtab>=4){
  st<-(ibtab-4)*9+1+33
  ed<-(ibtab-3)*9+33
  
}
cat("\\begin{table}[htbp]
    \\centering", file=OUTFILE, sep='', append=TRUE)

cat(sprintf("\\caption{\\label{tab%s:est:S3:tau=%s} Finite sample relative robust RMSE in simulation model M3 at %s.}
            ", 
            ibtab, tau, ifelse(tau==0.5, sprintf("%s quantile (median)",tau), sprintf("%s quantile",tau)) ), file=OUTFILE, sep='', append=TRUE)

cat("\\sisetup{round-precision=2,round-mode=places}
    \\begin{threeparttable}
    \\midrule\n", file=OUTFILE, sep='', append=TRUE)

if(ibtab<=3){
  cat("\\begin{tabular}{cccScccScccSccccc}\n", file=OUTFILE, sep='', append=TRUE)
}else{
  cat("\\begin{tabular}{cScScccScccSccccc}\n", file=OUTFILE, sep='', append=TRUE)
}

cat("\\toprule
    &          &          && \\multicolumn{2}{c}{$\\mathrm{IVQR.2SLS}$}&               && \\multicolumn{2}{c}{$\\mathrm{IVQR.QR} $} &               &&  &  &   \\\\
    \\cmidrule{5-6} \\cmidrule{9-10}
    $\\mathrm{DGP }$  &  $\\mathrm{Endog}$ &  $\\mathrm{Hetero}$  &&  $\\mathrm{AVG} $ & $\\mathrm{AGG}$       & $\\mathrm{IV}$ && $\\mathrm{AVG}$ & $\\mathrm{AGG}$         & $\\mathrm{QR}$ && $\\mathrm{BS} $  & $\\mathrm{MM} $  & $\\mathrm{SEE}$ \\\\
    \\midrule\n", file=OUTFILE, sep='', append=TRUE)

t<-0
for (ipar in st:ed) {
  t<-t+1
  cat(sprintf("%-10s & $%s$ & $%4s$ &&   %s  & %s  && %s  &  %s && %s & %s  \\\\\n",
              t, para.space[ipar,1],   para.space[ipar,2],
              ifelse(b.hats.rRMSE.par.rel[ipar,'relAVG.MMIV']<1, sprintf("$\\num[math-rm=\\mathbf]{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMIV']), sprintf("$\\num{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMIV'])),
              paste0(sprintf("$\\num{%4.6f}$",c(b.hats.rRMSE.par.rel[ipar,'relAGG.MMIV'],b.hats.rRMSE.par.rel[ipar,'relIV'])),collapse=' & '),
              ifelse(b.hats.rRMSE.par.rel[ipar,'relAVG.MMQR']<1, sprintf("$\\num[math-rm=\\mathbf]{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMQR']), sprintf("$\\num{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.MMQR'])),
              paste0(sprintf("$\\num{%4.6f}$",c(b.hats.rRMSE.par.rel[ipar,'relAGG.MMQR'],b.hats.rRMSE.par.rel[ipar,'relQR'])),collapse=' & '),
              ifelse(b.hats.rRMSE.par.rel[ipar,'relAVG.BS.RRMSE']<1, sprintf("$\\num[math-rm=\\mathbf]{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.BS.RRMSE']), sprintf("$\\num{%.6f}$",b.hats.rRMSE.par.rel[ipar,'relAVG.BS.RRMSE'])),
              paste0(sprintf("$\\num{%4.6f}$",c(b.hats.rRMSE.par.rel[ipar,'relGMMQ'],b.hats.rRMSE.par.rel[ipar,'relSEE'])),collapse=' & ')
  ), file=OUTFILE, sep='', append=TRUE)
  
} # end of loop for IVQR-2SLS latex table


cat(sprintf("\\bottomrule
            \\end{tabular}
            %%
            \\begin{tablenotes}
            \\item $\\num{%d}$ replications. $\\num{%d}$ bootstraps. Sample size is %d.\\\\
            \\end{tablenotes}
            \\end{threeparttable}
            \\end{table}\n", NREPLIC, BREPLIC,n), file=OUTFILE, sep='', append=TRUE)

cat(sprintf("%% total time elapsed: %s\n\n\n",format(Sys.time()-start.time)),
    file=OUTFILE, sep='', append=TRUE)

} # end of 1:6 ibtab loop
# CLS-VaryHetero  (ALL ESTIMATES) TABLE ####################  END  ################### sepearate
###########################################################################################################################################################








# ###################################################################################################################################################
# ###################################################################################################################################################
# ###                                                                                                                                             ###
# ###                                                                                                                                             ###
# ###                                                                                                                                             ###
# ###################################################################################################################################################
# ###################################################################################################################################################









# LaTeX table output
###########################################################################################################################################################
## CLS-VaryHeteo (UPPER/LOWER bound) TABLE  ####################  BEGIN  ###################

cat("\\begin{table}[htbp]
    \\centering", file=OUTFILE, sep='', append=TRUE)

cat(sprintf("\\caption{\\label{tab:bound:S3:tau=%s} The lower and upper bounds of finite sample relative robust RMSE in simulation model M3 at %s.}
            ",
            tau, ifelse(tau==0.5, sprintf("%s quantile (median)",tau), sprintf("%s quantile",tau)) ), file=OUTFILE, sep='', append=TRUE)

cat("\\sisetup{round-precision=3,round-mode=places}
    \\begin{threeparttable}
    \\begin{tabular}{cccScccScccSccc}
    \\toprule
    &&          && \\multicolumn{8}{c}{$\\mathrm{Fixed \\ Endog }$}               \\\\
    \\cmidrule{5-12}
    &&          && \\multicolumn{2}{c}{$\\mathrm{No}\\ (c=0)$}              && \\multicolumn{2}{c}{$\\mathrm{Some}\\ (c=0.2) $}     && \\multicolumn{2}{c}{$\\mathrm{Much}\\ (c=0.4) $}    \\\\
     \\cmidrule{5-6} \\cmidrule{8-9} \\cmidrule{11-12}
    &&          && $\\mathrm{Lower}$   & $\\mathrm{Upper}$   &&   $\\mathrm{Lower}$  & $\\mathrm{Upper}$    && $\\mathrm{Lower}$  & $\\mathrm{Upper}$            &&    \\\\
    \\cmidrule{3-12}\n", file=OUTFILE, sep='', append=TRUE)


cat(sprintf("%-6s && $%20s$ &&  %s  & %s  && %s  &  %s && %s & %s  \\\\\n",
            " ", "\\hat{\\theta}_{\\mathrm{AVG.2SLS}}", 
            sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[1:11,'relAVG.MMIV'])),
            sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[1:11,'relAVG.MMIV'])),
            sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[12:22,'relAVG.MMIV'])),
            sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[12:22,'relAVG.MMIV'])),
            sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[23:33,'relAVG.MMIV'])),
            sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[23:33,'relAVG.MMIV']))
), file=OUTFILE, sep='', append=TRUE)
cat(sprintf("%-6s && $%20s$ &&  %s  & %s  && %s  &  %s && %s & %s  \\\\\n",
            sprintf("$\\tau=%s$",tau), 
            "\\hat{\\theta}_{\\mathrm{AVG.QR}}", 
            sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[1:11,'relAVG.MMQR'])),
            sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[1:11,'relAVG.MMQR'])),
            sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[12:22,'relAVG.MMQR'])),
            sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[12:22,'relAVG.MMQR'])),
            sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[23:33,'relAVG.MMQR'])),
            sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[23:33,'relAVG.MMQR']))
), file=OUTFILE, sep='', append=TRUE)
cat(sprintf("%-6s && $%20s$ &&  %s  & %s  && %s  &  %s && %s & %s  \\\\\n",
            " ", "\\hat{\\theta}_{\\mathrm{BS}}", 
            sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[1:11,'relAVG.BS.RRMSE'])),
            sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[1:11,'relAVG.BS.RRMSE'])),
            sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[12:22,'relAVG.BS.RRMSE'])),
            sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[12:22,'relAVG.BS.RRMSE'])),
            sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[23:33,'relAVG.BS.RRMSE'])),
            sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[23:33,'relAVG.BS.RRMSE']))
), file=OUTFILE, sep='', append=TRUE)


cat(sprintf("\\bottomrule
            \\end{tabular}
            %%
            \\begin{tablenotes}
            \\item $\\num{%d}$ replications. $\\num{%d}$ bootstraps. Sample size is %d.\\\\
            \\end{tablenotes}
            \\end{threeparttable}
            \\end{table}\n", NREPLIC, BREPLIC,n), file=OUTFILE, sep='', append=TRUE)

cat(sprintf("%% total time elapsed: %s\n\n\n",format(Sys.time()-start.time)),
    file=OUTFILE, sep='', append=TRUE)
# CLS-VaryHetero (UPPER/LOWER bound) TABLE  ####################  END  ###################
###########################################################################################################################################################


###########################################################################################################################################################
## CLS-VaryEndog (UPPER/LOWER bound) TABLE  ####################  BEGIN  ###################

  
  cat("\\begin{table}[htbp]
      \\centering", file=OUTFILE, sep='', append=TRUE)
  
  cat(sprintf("\\caption{\\label{tab:bound:S3:tau=%s} The lower and upper bounds of finite sample relative robust RMSE in simulation model M3 at %s.}
              ",
              tau, ifelse(tau==0.5, sprintf("%s quantile (median)",tau), sprintf("%s quantile",tau)) ), file=OUTFILE, sep='', append=TRUE)
  
  cat("\\sisetup{round-precision=3,round-mode=places}
      \\begin{threeparttable}
      \\begin{tabular}{cccScccScccSccc}
      \\toprule
      &&          && \\multicolumn{8}{c}{$\\mathrm{Fixed \\ Hetero }$}               \\\\
      \\cmidrule{5-12}
      &&          && \\multicolumn{2}{c}{$\\mathrm{No}\\ (\\mathrm{hetero}=0)$}              && \\multicolumn{2}{c}{$\\mathrm{Some}\\ (\\mathrm{hetero}=0.5) $}     && \\multicolumn{2}{c}{$\\mathrm{Much}\\ (\\mathrm{hetero}=1) $}    \\\\
      \\cmidrule{5-6} \\cmidrule{8-9} \\cmidrule{11-12}
      &&          && $\\mathrm{Lower}$   & $\\mathrm{Upper}$   &&   $\\mathrm{Lower}$  & $\\mathrm{Upper}$    && $\\mathrm{Lower}$  & $\\mathrm{Upper}$            &&    \\\\
      \\cmidrule{3-12}\n", file=OUTFILE, sep='', append=TRUE)
  
  cat(sprintf("%-6s && $%20s$ &&  %s  & %s  && %s  &  %s && %s & %s  \\\\\n",
              " ", "\\hat{\\theta}_{\\mathrm{AVG.2SLS}}", 
              sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[34:42,'relAVG.MMIV'])),
              sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[34:42,'relAVG.MMIV'])),
              sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[43:51,'relAVG.MMIV'])),
              sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[43:51,'relAVG.MMIV'])),
              sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[52:60,'relAVG.MMIV'])),
              sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[52:60,'relAVG.MMIV']))
  ), file=OUTFILE, sep='', append=TRUE)
  cat(sprintf("%-6s && $%20s$ &&  %s  & %s  && %s  &  %s && %s & %s  \\\\\n",
              sprintf("$\\tau=%s$",tau), 
              "\\hat{\\theta}_{\\mathrm{AVG.QR}}", 
              sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[34:42,'relAVG.MMQR'])),
              sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[34:42,'relAVG.MMQR'])),
              sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[43:51,'relAVG.MMQR'])),
              sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[43:51,'relAVG.MMQR'])),
              sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[52:60,'relAVG.MMQR'])),
              sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[52:60,'relAVG.MMQR']))
  ), file=OUTFILE, sep='', append=TRUE)
  cat(sprintf("%-6s && $%20s$ &&  %s  & %s  && %s  &  %s && %s & %s  \\\\\n",
              " ", "\\hat{\\theta}_{\\mathrm{BS}}", 
              sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[34:42,'relAVG.BS.RRMSE'])),
              sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[34:42,'relAVG.BS.RRMSE'])),
              sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[43:51,'relAVG.BS.RRMSE'])),
              sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[43:51,'relAVG.BS.RRMSE'])),
              sprintf("$\\num{%.6f}$",min(b.hats.rRMSE.par.rel[52:60,'relAVG.BS.RRMSE'])),
              sprintf("$\\num{%.6f}$",max(b.hats.rRMSE.par.rel[52:60,'relAVG.BS.RRMSE']))
  ), file=OUTFILE, sep='', append=TRUE)


  cat(sprintf("\\bottomrule
              \\end{tabular}
              %%
              \\begin{tablenotes}
              \\item $\\num{%d}$ replications. $\\num{%d}$ bootstraps. Sample size is %d.\\\\
              \\end{tablenotes}
              \\end{threeparttable}
              \\end{table}\n", NREPLIC, BREPLIC,n), file=OUTFILE, sep='', append=TRUE)
  
  cat(sprintf("%% total time elapsed: %s\n\n\n",format(Sys.time()-start.time)),
      file=OUTFILE, sep='', append=TRUE)
# CLS-VaryEndog (UPPER/LOWER bound) TABLE  ####################  END  ###################
###########################################################################################################################################################



###########################################################################################################################################################
# line for increasing endogeneity level, performance of rel.rRMSE(b.hats)
# if (ENDOG.PIC.FLAG){
  for(ibtab in 1:6){
    st<-(ibtab-1)*11+1
    ed<-ibtab*11
    if(ibtab>=4){
      st<-(ibtab-4)*9+1+33
      ed<-(ibtab-3)*9+33
    }

        if(ibtab<=3){
       a <- 2
      xr <- 1
      xlabl<-"hetero value (heterogeneity level)"
        }else{
          a <- 1
         xr <- 0.4
         xlabl<-"c0 value (endogeneity level)"
    }
  # 3 estimators averaging graphs
  if(ibtab<=3){
    pdf(file=sprintf("Endog_pic_3EST_CLS_VH_atEndog%s_DGPid%s_n%d_NREP%d_BREP%d_tau%s_S1S2%d_%s_.pdf", para.space[st, 1], nrow(para.space), n, NREPLIC, BREPLIC, tau, S1S2,format(Sys.time(),"%Y_%m_%d")),
        pointsize=12, width=7, height=7)  
  }else{
    pdf(file=sprintf("Endog_pic_3EST_CLS_VE_atHetero%s_DGPid%s_n%d_NREP%d_BREP%d_tau%s_S1S2%d_%s_.pdf", para.space[st, 2], nrow(para.space), n, NREPLIC, BREPLIC, tau, S1S2,format(Sys.time(),"%Y_%m_%d")),
        pointsize=12, width=7, height=7)
  }
  par(family="serif",mar=c(5.1, 6, 6, 3))

  plot(para.space[, a], b.hats.rRMSE.par.rel[,5],type="n",
       xlab=xlabl,ylab="(relative) Robust RMSE",
       xlim=range(0,xr),
       ylim=c(0, 1.2), mgp=c(3,1,0),
       cex.main=2, cex.lab=2, cex.axis=2)

  LWD <- 4
  lines(para.space[st:ed, a], b.hats.rRMSE.par.rel[st:ed, 5], col=1, lty=3, lwd=LWD) #ip,tau,n
  lines(para.space[st:ed, a], b.hats.rRMSE.par.rel[st:ed, 7], col="springgreen", lty=1, lwd=LWD) #ip,tau,n
  if(BS.FLAG){
    lines(para.space[st:ed, a], b.hats.rRMSE.par.rel[st:ed, 10], col="chocolate1", lty=2, lwd=LWD) #ip,tau,n
  }
  lines(para.space[st:ed, a], array(1, ed-st+1), col=8, lty=1, lwd=LWD) #ip,tau,n
  grid(nx = NULL, ny = NULL, col = "gray", lty = "dotted", lwd = par("lwd"), equilogs = TRUE)
  
  if(BS.FLAG){
    legend('bottomright',
           c(as.expression(substitute(avg.IVQR.2SLS)), as.expression(substitute(avg.IVQR.QR)),as.expression(substitute(BS)), as.expression(substitute(IVQR)) ),
           lwd=5,col=c(1,"springgreen","chocolate1",8),lty=c(3,1,2,1),inset=0,bty='n',cex=1.8, x.intersp=0.5, y.intersp=1.2)
  }else{
    legend('bottomright',
           c(as.expression(substitute(avg.IVQR.2SLS)), as.expression(substitute(avg.IVQR.QR)), as.expression(substitute(IVQR)) ),
           lwd=5,col=c(1,"springgreen","chocolate1",8),lty=c(3,1,2,1),inset=0,bty='n',cex=1.8, x.intersp=0.5, y.intersp=1.2)
  }

  if (SAVE.FLAG) dev.off()

  # IVQR-QR averaging graphs
  if(ibtab<=3){
    pdf(file=sprintf("Endog_pic_AVGMMQR_CLS_VH_atEndog%s_DGPid%s_n%d_NREP%d_BREP%d_tau%s_S1S2%d_%s_.pdf", para.space[st, 1], nrow(para.space), n, NREPLIC, BREPLIC, tau, S1S2,format(Sys.time(),"%Y_%m_%d")),
                   pointsize=12, width=7, height=7)
  }else{
    pdf(file=sprintf("Endog_pic_AVGMMQR_CLS_VE_atHetero%s_DGPid%s_n%d_NREP%d_BREP%d_tau%s_S1S2%d_%s_.pdf", para.space[st, 2], nrow(para.space), n, NREPLIC, BREPLIC, tau, S1S2,format(Sys.time(),"%Y_%m_%d")),
        pointsize=12, width=7, height=7)
  }
  par(family="serif",mar=c(5.1, 6, 6, 3))
  plot(para.space[st:ed, a], b.hats.rRMSE.par.rel[st:ed,7],type="n",
       xlab=xlabl,ylab="(relative) Robust RMSE",
       xlim=range(0,xr),
       ylim=c(0, 1.2), mgp=c(3,1,0),
       cex.main=2, cex.lab=2, cex.axis=2)
  LWD <- 4
  lines(para.space[st:ed, a], b.hats.rRMSE.par.rel[st:ed, 7], col=1, lty=2, lwd=LWD) #ip,tau,n
  lines(para.space[st:ed, a], array(1, ed-st+1), col=8, lty=1, lwd=LWD) #ip,tau,n
  grid(nx = NULL, ny = NULL, col = "gray", lty = "dotted", lwd = par("lwd"), equilogs = TRUE)
  # topright
  legend('bottomright',
         c(as.expression(substitute(avg.IVQR.QR)), as.expression(substitute(IVQR)) ),
         lwd=5,col=c(1,8,8),lty=c(2,1,1),inset=0,bty='n',cex=1.8, x.intersp=0.5, y.intersp=1.2)
  if (SAVE.FLAG) dev.off()

  
  # IVQR-QR averaging graphs
  if(ibtab<=3){
    pdf(file=sprintf("Endog_pic_AVGMMIV_CLS_VH_atEndog%s_DGPid%s_n%d_NREP%d_BREP%d_tau%s_S1S2%d_%s_.pdf", para.space[st, 1], nrow(para.space), n, NREPLIC, BREPLIC, tau, S1S2,format(Sys.time(),"%Y_%m_%d")),
        pointsize=12, width=7, height=7)    
  }else{
    pdf(file=sprintf("Endog_pic_AVGMMIV_CLS_VE_atHetero%s_DGPid%s_n%d_NREP%d_BREP%d_tau%s_S1S2%d_%s_.pdf", para.space[st, 2], nrow(para.space), n, NREPLIC, BREPLIC, tau, S1S2,format(Sys.time(),"%Y_%m_%d")),
        pointsize=12, width=7, height=7)
  }
  par(family="serif",mar=c(5.1, 6, 6, 3))
  plot(para.space[st:ed, a], b.hats.rRMSE.par.rel[st:ed,5],type="n",
       xlab=xlabl,ylab="(relative) Robust RMSE",
       xlim=range(0,xr),
       ylim=c(0, 1.2), mgp=c(3,1,0),
       cex.main=2, cex.lab=2, cex.axis=2)
  LWD <- 4
  lines(para.space[st:ed, a], b.hats.rRMSE.par.rel[st:ed, 5], col=1, lty=2, lwd=LWD) #ip,tau,n
  lines(para.space[st:ed, a], array(1, ed-st+1), col=8, lty=1, lwd=LWD) #ip,tau,n
  grid(nx = NULL, ny = NULL, col = "gray", lty = "dotted", lwd = par("lwd"), equilogs = TRUE)
  # topright
  legend('bottomright',
         c(as.expression(substitute(avg.IVQR.2SLS)), as.expression(substitute(IVQR)) ),
         lwd=5,col=c(1,8,8),lty=c(2,1,1),inset=0,bty='n',cex=1.8, x.intersp=0.5, y.intersp=1.2)
  if (SAVE.FLAG) dev.off()
  
  # BS graphs
  if(BS.FLAG){
    if(ibtab<=3){
      pdf(file=sprintf("Endog_pic_BS_CLS_VH_atEndog%s_DGPid%s_n%d_NREP%d_BREP%d_tau%s_S1S2%d_%s_.pdf", para.space[st, 1], nrow(para.space), n, NREPLIC, BREPLIC, tau, S1S2,format(Sys.time(),"%Y_%m_%d")),
          pointsize=12, width=7, height=7)
    }else{
      pdf(file=sprintf("Endog_pic_BS_CLS_VE_atHetero%s_DGPid%s_n%d_NREP%d_BREP%d_tau%s_S1S2%d_%s_.pdf", para.space[st, 2], nrow(para.space), n, NREPLIC, BREPLIC, tau, S1S2,format(Sys.time(),"%Y_%m_%d")),
          pointsize=12, width=7, height=7)
    }
    
    par(family="serif",mar=c(5.1, 6, 6, 3))
    plot(para.space[st:ed, a], b.hats.rRMSE.par.rel[st:ed,10],type="n",
         xlab=xlabl,ylab="(relative) Robust RMSE",
         xlim=range(0,xr),
         ylim=c(0, 1.2), mgp=c(3,1,0),
         cex.main=2, cex.lab=2, cex.axis=2)
    LWD <- 4
    lines(para.space[st:ed, a], b.hats.rRMSE.par.rel[st:ed, 10], col=1, lty=2, lwd=LWD) #ip,tau,n
    lines(para.space[st:ed, a], array(1, ed-st+1), col=8, lty=1, lwd=LWD) #ip,tau,n
    grid(nx = NULL, ny = NULL, col = "gray", lty = "dotted", lwd = par("lwd"), equilogs = TRUE)
    # topright
    legend('bottomright',
           c(as.expression(substitute(BS)), as.expression(substitute(IVQR)) ),
           lwd=5,col=c(1,8,8),lty=c(2,1,1),inset=0,bty='n',cex=1.8, x.intersp=0.5, y.intersp=1.2)
    if (SAVE.FLAG) dev.off()
  } # END OF BS.FLAG if
  
} # end of ibtab loop
  
###########################################################################################################################################################

##########################################################################################################################################################
##########################################################################################################################################################






st1 <- Sys.time()
st1-st0


############################### if using parallel, then close parallel in the end
if (PARALLEL.FLAG){
  if (num_cores>1) stopCluster(cl)
}
#############################################################################################









#EOF



