# by Xin Liu
# xl6f6@mail.missouri.edu or xliu@mail.missouri.edu


# modified from ivqr_gmm. Use GenSA on the aggressive moments for aggressive estimator. Here "aggressive" moments= "conservative" moments + IV moments
IDENTITY.FLAG <- FALSE
ZZ.FLAG <- FALSE


source("ivqr_onestep.R")
source("gmmq.R")

library("GenSA")

ivqr.iv.aggressive <- function(Y, X.excl, Z.excl, D, tau, h=0, dB, max.time, upper, lower, weight.mtx, 
                     structure=c('iid','ts'), LRV.kernel=c('QS','Bartlett','uniform'), 
                     Lambda=function(y, x, b){y-x%*%b }, Lambda.derivative=function(y,x,b){-x}, 
                     Lambda.gmmq=function(y,x,b) {y[,1]-cbind(y[,-1],x)%*%b}, 
                     Lambda.derivative.gmmq=function(y,x,b) {-cbind(y[,-1],x)}, 
                     b.init=0 ) {

    X <- cbind(D, X.excl) # 
    Z <- cbind(Z.excl, X.excl)  # bc the weighting mtx is (addmoments, Z.excl, X.excl). HERE Correspondingly should be (demean.Z.excl, Z.excl, X.excl)
  
    Y.gmmq <- cbind(Y,D)
    demean.Z.excl<- Z.excl-colMeans(Z.excl)

    n <- length(Y)
    dZ <- ncol(Z) #number of instruments
    dX <- ncol(X) #number of regressors

    Itilde <- Itilde.KS17
    Itilde.deriv <- Itilde.deriv.KS17
    
    # Add bandwidth as second argument to G() and G'() functions.
    Gfn <- function(v,h) {      Itilde.KS17(v/h)    }
    Gpfn <- function(v,h) {      Itilde.deriv.KS17(v/h)    }
    
    # 
    if (missing(b.init)){
    ret <- ivqr.onestep(Y=Y, X.excl=X.excl, Z.excl=Z.excl, D=D, tau=tau, h=h, dB=dB,
                      structure=structure, LRV.kernel=LRV.kernel, 
                      Lambda=Lambda, Lambda.derivative=Lambda.derivative,
                      Lambda.gmmq=Lambda.gmmq, Lambda.derivative.gmmq=Lambda.derivative.gmmq, 
                      b.init=b.init ) 
    
    
    
    b.initial <- ret$b.onestep # b.onestep is (slope, intercetp); 
    if(missing(h)) {h.initial <- ret$h
    } else{
        h.initial <- h
      }
    }else{
      b.initial <- b.init 
      h.initial <- h
      
    }
    
    if(length(b.initial)!=length(lower)){stop("lower bound doesn't have right dimension.")}
    if(length(b.initial)!=length(upper)){stop("upper bound doesn't have right dimension.")}
    

    
    W.hat<-weight.mtx
    
      ivqr.obj <- function(b, h, weight.mtx) {
          L <- matrix(Lambda(y=Y, x=X, b=b), ncol=1) # here bc both X and b are (slope, intercept), thus can use Lambda fn

          gni <- cbind(demean.Z.excl*repmat(L,1,ncol(Z.excl)), cbind(Z.excl, X.excl)*repmat((Gfn(-L,h)-tau),1,dZ))
          g.bar <- colMeans(gni) 
          
          obj.fn <- t(g.bar)%*% weight.mtx %*%g.bar
          return(obj.fn)
      }

      obj.fn <- function(b) n*ivqr.obj(b=b,  h=h.initial, weight.mtx=W.hat)
 
     
     ivqr.gmm1 <- GenSA(par=b.initial, fn=obj.fn,
                        lower=lower, upper=upper,
                  control=list(max.time=max.time))
     b <- ivqr.gmm1$par
     
     J.stat <- obj.fn(b)
 
     return(list(b=b, obj.fn=obj.fn, h=h.initial, W.hat=W.hat, J.stat=J.stat))
  }








#EOF